/**********************************************************************************************************************************/
/* Project      : TAPER                                                                                                           */
/* Module       : TPDISASS.C                                                                                                      */
/* Description  : Block disassembly                                                                                               */
/* Version type : Program module                                                                                                  */
/* Last changed : 22-06-1998  5:45                                                                                                */
/* Update count : 13                                                                                                              */
/* OS type      : PC (DOS)                                                                                                        */
/* Compiler     : Watcom C32 Optimizing Compiler Version 10.x                                                                     */
/* Linker       : Watcom C Linker                                                                                                 */
/*                                                                                                                                */
/*                Copyleft (C) 1997-2001 ThunderWare Research Center, written by Martijn van der Heide.                           */
/*                                                                                                                                */
/*                This program is free software; you can redistribute it and/or                                                   */
/*                modify it under the terms of the GNU General Public License                                                     */
/*                as published by the Free Software Foundation; either version 2                                                  */
/*                of the License, or (at your option) any later version.                                                          */
/*                                                                                                                                */
/*                This program is distributed in the hope that it will be useful,                                                 */
/*                but WITHOUT ANY WARRANTY; without even the implied warranty of                                                  */
/*                MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                                                   */
/*                GNU General Public License for more details.                                                                    */
/*                                                                                                                                */
/*                You should have received a copy of the GNU General Public License                                               */
/*                along with this program; if not, write to the Free Software                                                     */
/*                Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                                     */
/*                                                                                                                                */
/**********************************************************************************************************************************/

#define __TPDISASS_MAIN__

#include "wbench32/wbench32.h"
#include <io.h>
#include <fcntl.h>
#include <math.h>
#include "tpfile.h"
#include "tpconfig.h"
#include "tpmain.h"
#include "tpaudio.h"
#include "tptape.h"
#include "tpscreen.h"
#include "tpdisass.h"

#pragma pack          ()

/**********************************************************************************************************************************/
/* Define the various window widths.                                                                                              */
/**********************************************************************************************************************************/

#define MINVWLINES   15                                                          /* Max no. Spectrum lines that fit on the screen */
#define DEFVWLINES   25
#define MAXVWLINES   38

/**********************************************************************************************************************************/
/* Define all (static) global variables used by the disassembler.                                                                 */
/**********************************************************************************************************************************/

#define DISASSWIDTH  60                                                                           /* Width of a disassembled line */

static struct ComboInfo  _SymbolWindow;

static struct _Labels_s                                                                                           /* Symbol table */
{
  word                _Address;
  char                _Name[10];
};

static struct _Labels_s **_Labels       = NULL;                                                      /* User defined symbol table */
static int       _NumLabels             = 0;
static int       _AllocLabels           = 0;
static bool      _HaveLabels            = FALSE;                                   /* If TRUE, we have a file with address labels */
static bool      _LabelsChanged         = FALSE;                    /* TRUE if the labels have been changed using EditSymbolTable */

static struct _Labels_s **_SpecLabels   = NULL;                                                  /* 48K Spectrum ROM symbol table */
static int       _NumSpecLabels         = 0;
static bool      _HaveSpecLabels        = FALSE;
static bool      _SpecLabelsLoaded      = FALSE;                                     /* Load internal symbol table file only once */

static byte      _SpectrumROMWatermark[8] = { 0xF3, 0xAF, 0x11, 0xFF, 0xFF, 0xC3, 0xCB, 0x11 };
static bool      _IsSpectrumROM         = FALSE;

static char      _LabelFileName[13]     = "\0";
static char      _LabelPath[MAXDIR + 1] = "\0";
static char      _Path[MAXDIR + 1]      = "\0";
static char      _File[13]              = "\0";
static char      _Mask[13]              = "*.sym";
static char     *_SpectrumLabelFile     = "sprom48.sym";
static short     _DisassemblerWidth     = DISASSWIDTH;                             /* Number of characters in a disassembled line */

static bool      _ExpandSystemVariables = TRUE;                            /* Prints system variables by name is stead of address */
static bool      _ExpandSpectrumROM     = TRUE;            /* Expands RST 0008 and 0028 literals to literals in stead of Z80 code */
static bool      _NumbersAreHex         = TRUE;                                            /* Prints numbers either as hex or dec */

static struct _InstructionMemory_s                                          /* Storage for the start of one instruction in memory */
{
  word               _Address;
  byte               _RSTValue;                                             /* Used only when handling RST 08 and RST 28 literals */
};

static struct _SpectrumFP_s                                                      /* The Spectrum-style 5-byte floating point form */
{                                                                                             /* (Small integers are catered for) */
  byte               _Exponent;
  byte               _Mantissa[4];
};

/**********************************************************************************************************************************/
/* These are the variables needed for backward scrolling in the disassembler.                                                     */
/**********************************************************************************************************************************/

static bool      _CountBytesOnly        = FALSE;                            /* If TRUE, do not print, simply advance instructions */
static short     _CurrentHeight         = 20;                                        /* Number of instructions per page on screen */
static struct _InstructionMemory_s _BackScrollArray[MAXVWLINES];

/**********************************************************************************************************************************/
/* Define the scroller gadgets at the right of the window.                                                                        */
/**********************************************************************************************************************************/

static byte      _GadUpBt[] = { 8, 12,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 1, 0, 0, 0,
  0, 0, 0, 1, 0, 1, 0, 0,
  0, 0, 1, 0, 0, 0, 1, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0 };
static byte      _GadUpLn[] = { 8, 12,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 1, 0, 0, 0,
  0, 0, 0, 1, 1, 1, 0, 0,
  0, 0, 1, 1, 1, 1, 1, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0 };
static byte      _GadUpPg[] = { 8, 12,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 1, 0, 0, 0,
  0, 0, 0, 1, 1, 1, 0, 0,
  0, 0, 1, 1, 1, 1, 1, 0,
  0, 1, 1, 1, 1, 1, 1, 1,
  0, 0, 0, 1, 1, 1, 0, 0,
  0, 0, 0, 1, 1, 1, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0 };
static byte      _GadDnPg[] = { 8, 12,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 1, 1, 1, 0, 0,
  0, 0, 0, 1, 1, 1, 0, 0,
  0, 1, 1, 1, 1, 1, 1, 1,
  0, 0, 1, 1, 1, 1, 1, 0,
  0, 0, 0, 1, 1, 1, 0, 0,
  0, 0, 0, 0, 1, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0 };
static byte      _GadDnLn[] = { 8, 12,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 1, 1, 1, 1, 1, 0,
  0, 0, 0, 1, 1, 1, 0, 0,
  0, 0, 0, 0, 1, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0 };
static byte      _GadDnBt[] = { 8, 12,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 1, 0, 0, 0, 1, 0,
  0, 0, 0, 1, 0, 1, 0, 0,
  0, 0, 0, 0, 1, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0 };
static bool _GadgetsConverted = FALSE;                                        /* They're static, so we need to convert only once! */

/**********************************************************************************************************************************/
/* The list of opcodes. Used just like that.                                                                                      */
/**********************************************************************************************************************************/

static char      *_OpCodes[] = {
  "ADC",  /* 00 */     "ADD",  /* 01 */     "AND",  /* 02 */     "BIT",  /* 03 */
  "CALL", /* 04 */     "CCF",  /* 05 */     "CP",   /* 06 */     "CPD",  /* 07 */
  "CPDR", /* 08 */     "CPI",  /* 09 */     "CPIR", /* 0A */     "CPL",  /* 0B */
  "DAA",  /* 0C */     "DEC",  /* 0D */     "DI",   /* 0E */     "DJNZ", /* 0F */
  "EI",   /* 10 */     "EX",   /* 11 */     "EXX",  /* 12 */     "HALT", /* 13 */
  "IM",   /* 14 */     "IN",   /* 15 */     "INC",  /* 16 */     "IND",  /* 17 */
  "INDR", /* 18 */     "INI",  /* 19 */     "INIR", /* 1A */     "JP",   /* 1B */
  "JR",   /* 1C */     "LD",   /* 1D */     "LDD",  /* 1E */     "LDDR", /* 1F */
  "LDI",  /* 20 */     "LDIR", /* 21 */     "NEG",  /* 22 */     "NOP",  /* 23 */
  "OR",   /* 24 */     "OTDR", /* 25 */     "OTIR", /* 26 */     "OUT",  /* 27 */
  "OUTD", /* 28 */     "OUTI", /* 29 */     "POP",  /* 2A */     "PUSH", /* 2B */
  "RES",  /* 2C */     "RET",  /* 2D */     "RETI", /* 2E */     "RETN", /* 2F */
  "RL",   /* 30 */     "RLA",  /* 31 */     "RLC",  /* 32 */     "RLCA", /* 33 */
  "RLD",  /* 34 */     "RR",   /* 35 */     "RRA",  /* 36 */     "RRC",  /* 37 */
  "RRCA", /* 38 */     "RRD",  /* 39 */     "RST",  /* 3A */     "SBC",  /* 3B */
  "SCF",  /* 3C */     "SET",  /* 3D */     "SLA",  /* 3E */     "SLL",  /* 3F */
  "SRA",  /* 40 */     "SRL",  /* 41 */     "SUB",  /* 42 */     "XOR"   /* 43 */
};

/**********************************************************************************************************************************/
/* The list of registers. Special: in DD/FD expansions, HL denotes IX/IY, H and L denote IXh/IYh and IXl/IYl.                     */
/**********************************************************************************************************************************/

static char      *_Registers[] = {
  "A",    /* 00 */     "B",    /* 01 */     "C",    /* 02 */     "D",    /* 03 */
  "E",    /* 04 */     "H",    /* 05 */     "L",    /* 06 */     "I",    /* 07 */
  "R",    /* 08 */     "AF",   /* 09 */     "BC",   /* 0A */     "DE",   /* 0B */
  "HL",   /* 0C */     "SP"    /* 0D */
};

/**********************************************************************************************************************************/
/* The list of flags. Used just like that.                                                                                        */
/**********************************************************************************************************************************/

static char      *_Flags[] = {
  "C",    /* 00 */     "M",    /* 01 */     "NC",   /* 02 */     "NZ",   /* 03 */
  "P",    /* 04 */     "PE",   /* 05 */     "PO",   /* 06 */     "Z"     /* 07 */
};

/**********************************************************************************************************************************/
/* Define the operand types of an instruction.                                                                                    */
/**********************************************************************************************************************************/

#define OPER_BRACKETS 0x10    /* 0001 - Operand in brackets */

#define OPER_DIRECT   0x00    /* 000 - For RST. Lower 4 bits tell the value (/8) */
#define OPER_REGISTER 0x20    /* 001 - Lower 4 bits are the index into the _Registers table */
#define OPER_FLAG     0x40    /* 010 - Lower 4 bits are the index into the _Flags table */
#define OPER_8BIT     0x60    /* 011 - Value always in last byte of the instruction */
#define OPER_16BIT    0x80    /* 100 - Value always in the last word of the instuction */
#define OPER_ADDRESS  0xA0    /* 101 - Same as 16BIT, used to seperate numbers and addresses */
#define OPER_RELJUMP  0xC0    /* 110 - For JR and DJNZ */
#define OPER_BITVALUE 0xE0    /* 111 - Lower 4 bits tell the bit value. Also used for IM values (hex = dec) */

/**********************************************************************************************************************************/
/* Define the storage format of each instruction.                                                                                 */
/**********************************************************************************************************************************/

static struct _Z80Assem_s
{
  byte                _NumBytes;                                        /* Number of bytes that the instructions takes altogether */
  byte                _NumOperands;                                                                   /* Number of operands - 0-2 */
  byte                _OpCode;                                                                   /* Index into the _OpCodes array */
  byte                _Operand1;                                                                  /* Combined operand information */
  byte                _Operand2;
};

/**********************************************************************************************************************************/
/* Here goes: Let the instructions begin!                                                                                         */
/**********************************************************************************************************************************/

static struct _Z80Assem_s _SingleCodes[256] = {
  { 1, 0, 0x23, 0, 0 },                                                         /* 00 = NOP           */
  { 3, 2, 0x1D, OPER_REGISTER | 0x0A, OPER_16BIT },                             /* 01 = LD   BC, nn   */
  { 1, 2, 0x1D, OPER_REGISTER | OPER_BRACKETS | 0x0A, OPER_REGISTER | 0x00 },   /* 02 = LD   (BC), A  */
  { 1, 1, 0x16, OPER_REGISTER | 0x0A, 0 },                                      /* 03 = INC  BC       */
  { 1, 1, 0x16, OPER_REGISTER | 0x01, 0 },                                      /* 04 = INC  B        */
  { 1, 1, 0x0D, OPER_REGISTER | 0x01, 0 },                                      /* 05 = DEC  B        */
  { 2, 2, 0x1D, OPER_REGISTER | 0x01, OPER_8BIT },                              /* 06 = LD   B,n      */
  { 1, 0, 0x33, 0, 0 },                                                         /* 07 = RLCA          */
  { 1, 2, 0x11, OPER_REGISTER | 0x09, OPER_REGISTER | 0x09 },                   /* 08 = EX   AF, AF'  ---> SPECIAL AF' */
  { 1, 2, 0x01, OPER_REGISTER | 0x0C, OPER_REGISTER | 0x0A },                   /* 09 = ADD  HL, BC   */
  { 1, 2, 0x1D, OPER_REGISTER | 0x00, OPER_REGISTER | OPER_BRACKETS | 0x0A },   /* 0A = LD   A, (BC)  */
  { 1, 1, 0x0D, OPER_REGISTER | 0x0A, 0 },                                      /* 0B = DEC  BC       */
  { 1, 1, 0x16, OPER_REGISTER | 0x02, 0 },                                      /* 0C = INC  C        */
  { 1, 1, 0x0D, OPER_REGISTER | 0x02, 0 },                                      /* 0D = DEC  C        */
  { 2, 2, 0x1D, OPER_REGISTER | 0x02, OPER_8BIT },                              /* 0E = LD   C, n     */
  { 1, 0, 0x38, 0, 0 },                                                         /* 0F = RRCA          */
  { 2, 1, 0x0F, OPER_RELJUMP, 0 },                                              /* 10 = DJNZ e        */
  { 3, 2, 0x1D, OPER_REGISTER | 0x0B, OPER_16BIT },                             /* 11 = LD   DE, nn   */
  { 1, 2, 0x1D, OPER_REGISTER | OPER_BRACKETS | 0x0B, OPER_REGISTER | 0x00 },   /* 12 = LD   (DE), A  */
  { 1, 1, 0x16, OPER_REGISTER | 0x0B, 0 },                                      /* 13 = INC  DE       */
  { 1, 1, 0x16, OPER_REGISTER | 0x03, 0 },                                      /* 14 = INC  D        */
  { 1, 1, 0x0D, OPER_REGISTER | 0x03, 0 },                                      /* 15 = DEC  D        */
  { 2, 2, 0x1D, OPER_REGISTER | 0x03, OPER_8BIT },                              /* 16 = LD   D, n     */
  { 1, 0, 0x31, 0, 0 },                                                         /* 17 = RLA           */
  { 2, 1, 0x1C, OPER_RELJUMP, 0 },                                              /* 18 = JR   e        */
  { 1, 2, 0x01, OPER_REGISTER | 0x0C, OPER_REGISTER | 0x0B },                   /* 19 = ADD  HL, DE   */
  { 1, 2, 0x1D, OPER_REGISTER | 0x00, OPER_REGISTER | OPER_BRACKETS | 0x0B },   /* 1A = LD   A, (DE)  */
  { 1, 1, 0x0D, OPER_REGISTER | 0x0B, 0 },                                      /* 1B = DEC  DE       */
  { 1, 1, 0x16, OPER_REGISTER | 0x04, 0 },                                      /* 1C = INC  E        */
  { 1, 1, 0x0D, OPER_REGISTER | 0x04, 0 },                                      /* 1D = DEC  E        */
  { 2, 2, 0x1D, OPER_REGISTER | 0x04, OPER_8BIT },                              /* 1E = LD   E, n     */
  { 1, 0, 0x36, 0, 0 },                                                         /* 1F = RRA           */
  { 2, 2, 0x1C, OPER_FLAG | 0x03, OPER_RELJUMP },                               /* 20 = JR   NZ, e    */
  { 3, 2, 0x1D, OPER_REGISTER | 0x0C, OPER_16BIT },                             /* 21 = LD   HL, nn   */
  { 3, 2, 0x1D, OPER_16BIT | OPER_BRACKETS, OPER_REGISTER | 0x0C },             /* 22 = LD   (nn), HL */
  { 1, 1, 0x16, OPER_REGISTER | 0x0C, 0 },                                      /* 23 = INC  HL       */
  { 1, 1, 0x16, OPER_REGISTER | 0x05, 0 },                                      /* 24 = INC  H        */
  { 1, 1, 0x0D, OPER_REGISTER | 0x05, 0 },                                      /* 25 = DEC  H        */
  { 2, 2, 0x1D, OPER_REGISTER | 0x05, OPER_8BIT },                              /* 26 = LD   H, n     */
  { 1, 0, 0x0C, 0, 0 },                                                         /* 27 = DAA           */
  { 2, 2, 0x1C, OPER_FLAG | 0x07, OPER_RELJUMP },                               /* 28 = JR   Z, e     */
  { 1, 2, 0x01, OPER_REGISTER | 0x0C, OPER_REGISTER | 0x0C },                   /* 29 = ADD  HL, HL   */
  { 3, 2, 0x1D, OPER_REGISTER | 0x0C, OPER_16BIT | OPER_BRACKETS },             /* 2A = LD   HL, (nn) */
  { 1, 1, 0x0D, OPER_REGISTER | 0x0C, 0 },                                      /* 2B = DEC  HL       */
  { 1, 1, 0x16, OPER_REGISTER | 0x06, 0 },                                      /* 2C = INC  L        */
  { 1, 1, 0x0D, OPER_REGISTER | 0x06, 0 },                                      /* 2D = DEC  L        */
  { 2, 2, 0x1D, OPER_REGISTER | 0x06, OPER_8BIT },                              /* 2E = LD   L, n     */
  { 1, 0, 0x0B, 0, 0 },                                                         /* 2F = CPL           */
  { 2, 2, 0x1C, OPER_FLAG | 0x02, OPER_RELJUMP },                               /* 30 = JR   NC, e    */
  { 3, 2, 0x1D, OPER_REGISTER | 0x0D, OPER_16BIT },                             /* 31 = LD   SP, nn   */
  { 3, 2, 0x1D, OPER_16BIT | OPER_BRACKETS, OPER_REGISTER | 0x00 },             /* 32 = LD   (nn), A  */
  { 1, 1, 0x16, OPER_REGISTER | 0x0D, 0 },                                      /* 33 = INC  SP       */
  { 1, 1, 0x16, OPER_REGISTER | OPER_BRACKETS | 0x0C, 0 },                      /* 34 = INC  (HL)     */
  { 1, 1, 0x0D, OPER_REGISTER | OPER_BRACKETS | 0x0C, 0 },                      /* 35 = DEC  (HL)     */
  { 2, 2, 0x1D, OPER_REGISTER | OPER_BRACKETS | 0x0C, OPER_8BIT },              /* 36 = LD   (HL), n  */
  { 1, 0, 0x3C, 0, 0 },                                                         /* 37 = SCF           */
  { 2, 2, 0x1C, OPER_FLAG | 0x00, OPER_RELJUMP },                               /* 38 = JR   C, e     */
  { 1, 2, 0x01, OPER_REGISTER | 0x0C, OPER_REGISTER | 0x0D },                   /* 39 = ADD  HL, SP   */
  { 3, 2, 0x1D, OPER_REGISTER | 0x00, OPER_16BIT | OPER_BRACKETS },             /* 3A = LD   A, (nn)  */
  { 1, 1, 0x0D, OPER_REGISTER | 0x0D, 0 },                                      /* 3B = DEC  SP       */
  { 1, 1, 0x16, OPER_REGISTER | 0x00, 0 },                                      /* 3C = INC  A        */
  { 1, 1, 0x0D, OPER_REGISTER | 0x00, 0 },                                      /* 3D = DEC  A        */
  { 2, 2, 0x1D, OPER_REGISTER | 0x00, OPER_8BIT },                              /* 3E = LD   A, n     */
  { 1, 0, 0x05, 0, 0 },                                                         /* 3F = CCF           */
  { 1, 2, 0x1D, OPER_REGISTER | 0x01, OPER_REGISTER | 0x01 },                   /* 40 = LD   B, B     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x01, OPER_REGISTER | 0x02 },                   /* 41 = LD   B, C     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x01, OPER_REGISTER | 0x03 },                   /* 42 = LD   B, D     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x01, OPER_REGISTER | 0x04 },                   /* 43 = LD   B, E     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x01, OPER_REGISTER | 0x05 },                   /* 44 = LD   B, H     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x01, OPER_REGISTER | 0x06 },                   /* 45 = LD   B, L     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x01, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* 46 = LD   B, (HL)  */
  { 1, 2, 0x1D, OPER_REGISTER | 0x01, OPER_REGISTER | 0x00 },                   /* 47 = LD   B, A     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x02, OPER_REGISTER | 0x01 },                   /* 48 = LD   C, B     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x02, OPER_REGISTER | 0x02 },                   /* 49 = LD   C, C     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x02, OPER_REGISTER | 0x03 },                   /* 4A = LD   C, D     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x02, OPER_REGISTER | 0x04 },                   /* 4B = LD   C, E     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x02, OPER_REGISTER | 0x05 },                   /* 4C = LD   C, H     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x02, OPER_REGISTER | 0x06 },                   /* 4D = LD   C, L     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x02, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* 4E = LD   C, (HL)  */
  { 1, 2, 0x1D, OPER_REGISTER | 0x02, OPER_REGISTER | 0x00 },                   /* 4F = LD   C, A     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x03, OPER_REGISTER | 0x01 },                   /* 50 = LD   D, B     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x03, OPER_REGISTER | 0x02 },                   /* 51 = LD   D, C     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x03, OPER_REGISTER | 0x03 },                   /* 52 = LD   D, D     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x03, OPER_REGISTER | 0x04 },                   /* 53 = LD   D, E     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x03, OPER_REGISTER | 0x05 },                   /* 54 = LD   D, H     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x03, OPER_REGISTER | 0x06 },                   /* 55 = LD   D, L     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x03, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* 56 = LD   D, (HL)  */
  { 1, 2, 0x1D, OPER_REGISTER | 0x03, OPER_REGISTER | 0x00 },                   /* 57 = LD   D, A     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x04, OPER_REGISTER | 0x01 },                   /* 58 = LD   E, B     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x04, OPER_REGISTER | 0x02 },                   /* 59 = LD   E, C     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x04, OPER_REGISTER | 0x03 },                   /* 5A = LD   E, D     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x04, OPER_REGISTER | 0x04 },                   /* 5B = LD   E, E     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x04, OPER_REGISTER | 0x05 },                   /* 5C = LD   E, H     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x04, OPER_REGISTER | 0x06 },                   /* 5D = LD   E, L     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x04, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* 5E = LD   E, (HL)  */
  { 1, 2, 0x1D, OPER_REGISTER | 0x04, OPER_REGISTER | 0x00 },                   /* 5F = LD   E, A     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x05, OPER_REGISTER | 0x01 },                   /* 60 = LD   H, B     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x05, OPER_REGISTER | 0x02 },                   /* 61 = LD   H, C     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x05, OPER_REGISTER | 0x03 },                   /* 62 = LD   H, D     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x05, OPER_REGISTER | 0x04 },                   /* 63 = LD   H, E     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x05, OPER_REGISTER | 0x05 },                   /* 64 = LD   H, H     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x05, OPER_REGISTER | 0x06 },                   /* 65 = LD   H, L     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x05, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* 66 = LD   H, (HL)  */
  { 1, 2, 0x1D, OPER_REGISTER | 0x05, OPER_REGISTER | 0x00 },                   /* 67 = LD   H, A     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x06, OPER_REGISTER | 0x01 },                   /* 68 = LD   L, B     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x06, OPER_REGISTER | 0x02 },                   /* 69 = LD   L, C     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x06, OPER_REGISTER | 0x03 },                   /* 6A = LD   L, D     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x06, OPER_REGISTER | 0x04 },                   /* 6B = LD   L, E     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x06, OPER_REGISTER | 0x05 },                   /* 6C = LD   L, H     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x06, OPER_REGISTER | 0x06 },                   /* 6D = LD   L, L     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x06, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* 6E = LD   L, (HL)  */
  { 1, 2, 0x1D, OPER_REGISTER | 0x06, OPER_REGISTER | 0x00 },                   /* 6F = LD   L, A     */
  { 1, 2, 0x1D, OPER_REGISTER | OPER_BRACKETS | 0x0C, OPER_REGISTER | 0x01 },   /* 70 = LD   (HL), B  */
  { 1, 2, 0x1D, OPER_REGISTER | OPER_BRACKETS | 0x0C, OPER_REGISTER | 0x02 },   /* 71 = LD   (HL), C  */
  { 1, 2, 0x1D, OPER_REGISTER | OPER_BRACKETS | 0x0C, OPER_REGISTER | 0x03 },   /* 72 = LD   (HL), D  */
  { 1, 2, 0x1D, OPER_REGISTER | OPER_BRACKETS | 0x0C, OPER_REGISTER | 0x04 },   /* 73 = LD   (HL), E  */
  { 1, 2, 0x1D, OPER_REGISTER | OPER_BRACKETS | 0x0C, OPER_REGISTER | 0x05 },   /* 74 = LD   (HL), H  */
  { 1, 2, 0x1D, OPER_REGISTER | OPER_BRACKETS | 0x0C, OPER_REGISTER | 0x06 },   /* 75 = LD   (HL), L  */
  { 1, 0, 0x13, 0, 0 },                                                         /* 76 = HALT          */
  { 1, 2, 0x1D, OPER_REGISTER | OPER_BRACKETS | 0x0C, OPER_REGISTER | 0x00 },   /* 77 = LD   (HL), A  */
  { 1, 2, 0x1D, OPER_REGISTER | 0x00, OPER_REGISTER | 0x01 },                   /* 78 = LD   A, B     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x00, OPER_REGISTER | 0x02 },                   /* 79 = LD   A, C     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x00, OPER_REGISTER | 0x03 },                   /* 7A = LD   A, D     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x00, OPER_REGISTER | 0x04 },                   /* 7B = LD   A, E     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x00, OPER_REGISTER | 0x05 },                   /* 7C = LD   A, H     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x00, OPER_REGISTER | 0x06 },                   /* 7D = LD   A, L     */
  { 1, 2, 0x1D, OPER_REGISTER | 0x00, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* 7E = LD   A, (HL)  */
  { 1, 2, 0x1D, OPER_REGISTER | 0x00, OPER_REGISTER | 0x00 },                   /* 7F = LD   A, A     */
  { 1, 2, 0x01, OPER_REGISTER | 0x00, OPER_REGISTER | 0x01 },                   /* 80 = ADD  A, B     */
  { 1, 2, 0x01, OPER_REGISTER | 0x00, OPER_REGISTER | 0x02 },                   /* 81 = ADD  A, C     */
  { 1, 2, 0x01, OPER_REGISTER | 0x00, OPER_REGISTER | 0x03 },                   /* 82 = ADD  A, D     */
  { 1, 2, 0x01, OPER_REGISTER | 0x00, OPER_REGISTER | 0x04 },                   /* 83 = ADD  A, E     */
  { 1, 2, 0x01, OPER_REGISTER | 0x00, OPER_REGISTER | 0x05 },                   /* 84 = ADD  A, H     */
  { 1, 2, 0x01, OPER_REGISTER | 0x00, OPER_REGISTER | 0x06 },                   /* 85 = ADD  A, L     */
  { 1, 2, 0x01, OPER_REGISTER | 0x00, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* 86 = ADD  A, (HL)  */
  { 1, 2, 0x01, OPER_REGISTER | 0x00, OPER_REGISTER | 0x00 },                   /* 87 = ADD  A, A     */
  { 1, 2, 0x00, OPER_REGISTER | 0x00, OPER_REGISTER | 0x01 },                   /* 88 = ADC  A, B     */
  { 1, 2, 0x00, OPER_REGISTER | 0x00, OPER_REGISTER | 0x02 },                   /* 89 = ADC  A, C     */
  { 1, 2, 0x00, OPER_REGISTER | 0x00, OPER_REGISTER | 0x03 },                   /* 8A = ADC  A, D     */
  { 1, 2, 0x00, OPER_REGISTER | 0x00, OPER_REGISTER | 0x04 },                   /* 8B = ADC  A, E     */
  { 1, 2, 0x00, OPER_REGISTER | 0x00, OPER_REGISTER | 0x05 },                   /* 8C = ADC  A, H     */
  { 1, 2, 0x00, OPER_REGISTER | 0x00, OPER_REGISTER | 0x06 },                   /* 8D = ADC  A, L     */
  { 1, 2, 0x00, OPER_REGISTER | 0x00, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* 8E = ADC  A, (HL)  */
  { 1, 2, 0x00, OPER_REGISTER | 0x00, OPER_REGISTER | 0x00 },                   /* 8F = ADC  A, A     */
  { 1, 1, 0x42, OPER_REGISTER | 0x01, 0 },                                      /* 90 = SUB  B        */
  { 1, 1, 0x42, OPER_REGISTER | 0x02, 0 },                                      /* 91 = SUB  C        */
  { 1, 1, 0x42, OPER_REGISTER | 0x03, 0 },                                      /* 92 = SUB  D        */
  { 1, 1, 0x42, OPER_REGISTER | 0x04, 0 },                                      /* 93 = SUB  E        */
  { 1, 1, 0x42, OPER_REGISTER | 0x05, 0 },                                      /* 94 = SUB  H        */
  { 1, 1, 0x42, OPER_REGISTER | 0x06, 0 },                                      /* 95 = SUB  L        */
  { 1, 1, 0x42, OPER_REGISTER | OPER_BRACKETS | 0x0C, 0 },                      /* 96 = SUB  (HL)     */
  { 1, 1, 0x42, OPER_REGISTER | 0x00, 0 },                                      /* 97 = SUB  A        */
  { 1, 2, 0x3B, OPER_REGISTER | 0x00, OPER_REGISTER | 0x01 },                   /* 98 = SBC  A, B     */
  { 1, 2, 0x3B, OPER_REGISTER | 0x00, OPER_REGISTER | 0x02 },                   /* 99 = SBC  A, C     */
  { 1, 2, 0x3B, OPER_REGISTER | 0x00, OPER_REGISTER | 0x03 },                   /* 9A = SBC  A, D     */
  { 1, 2, 0x3B, OPER_REGISTER | 0x00, OPER_REGISTER | 0x04 },                   /* 9B = SBC  A, E     */
  { 1, 2, 0x3B, OPER_REGISTER | 0x00, OPER_REGISTER | 0x05 },                   /* 9C = SBC  A, H     */
  { 1, 2, 0x3B, OPER_REGISTER | 0x00, OPER_REGISTER | 0x06 },                   /* 9D = SBC  A, L     */
  { 1, 2, 0x3B, OPER_REGISTER | 0x00, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* 9E = SBC  A, (HL)  */
  { 1, 2, 0x3B, OPER_REGISTER | 0x00, OPER_REGISTER | 0x00 },                   /* 9F = SBC  A, A     */
  { 1, 1, 0x02, OPER_REGISTER | 0x01, 0 },                                      /* A0 = AND  B        */
  { 1, 1, 0x02, OPER_REGISTER | 0x02, 0 },                                      /* A1 = AND  C        */
  { 1, 1, 0x02, OPER_REGISTER | 0x03, 0 },                                      /* A2 = AND  D        */
  { 1, 1, 0x02, OPER_REGISTER | 0x04, 0 },                                      /* A3 = AND  E        */
  { 1, 1, 0x02, OPER_REGISTER | 0x05, 0 },                                      /* A4 = AND  H        */
  { 1, 1, 0x02, OPER_REGISTER | 0x06, 0 },                                      /* A5 = AND  L        */
  { 1, 1, 0x02, OPER_REGISTER | OPER_BRACKETS | 0x0C, 0 },                      /* A6 = AND  (HL)     */
  { 1, 1, 0x02, OPER_REGISTER | 0x00, 0 },                                      /* A7 = AND  A        */
  { 1, 1, 0x43, OPER_REGISTER | 0x01, 0 },                                      /* A8 = XOR  B        */
  { 1, 1, 0x43, OPER_REGISTER | 0x02, 0 },                                      /* A9 = XOR  C        */
  { 1, 1, 0x43, OPER_REGISTER | 0x03, 0 },                                      /* AA = XOR  D        */
  { 1, 1, 0x43, OPER_REGISTER | 0x04, 0 },                                      /* AB = XOR  E        */
  { 1, 1, 0x43, OPER_REGISTER | 0x05, 0 },                                      /* AC = XOR  H        */
  { 1, 1, 0x43, OPER_REGISTER | 0x06, 0 },                                      /* AD = XOR  L        */
  { 1, 1, 0x43, OPER_REGISTER | OPER_BRACKETS | 0x0C, 0 },                      /* AE = XOR  (HL)     */
  { 1, 1, 0x43, OPER_REGISTER | 0x00, 0 },                                      /* AF = XOR  A        */
  { 1, 1, 0x24, OPER_REGISTER | 0x01, 0 },                                      /* B0 = OR   B        */
  { 1, 1, 0x24, OPER_REGISTER | 0x02, 0 },                                      /* B1 = OR   C        */
  { 1, 1, 0x24, OPER_REGISTER | 0x03, 0 },                                      /* B2 = OR   D        */
  { 1, 1, 0x24, OPER_REGISTER | 0x04, 0 },                                      /* B3 = OR   E        */
  { 1, 1, 0x24, OPER_REGISTER | 0x05, 0 },                                      /* B4 = OR   H        */
  { 1, 1, 0x24, OPER_REGISTER | 0x06, 0 },                                      /* B5 = OR   L        */
  { 1, 1, 0x24, OPER_REGISTER | OPER_BRACKETS | 0x0C, 0 },                      /* B6 = OR   (HL)     */
  { 1, 1, 0x24, OPER_REGISTER | 0x00, 0 },                                      /* B7 = OR   A        */
  { 1, 1, 0x06, OPER_REGISTER | 0x01, 0 },                                      /* B8 = CP   B        */
  { 1, 1, 0x06, OPER_REGISTER | 0x02, 0 },                                      /* B9 = CP   C        */
  { 1, 1, 0x06, OPER_REGISTER | 0x03, 0 },                                      /* BA = CP   D        */
  { 1, 1, 0x06, OPER_REGISTER | 0x04, 0 },                                      /* BB = CP   E        */
  { 1, 1, 0x06, OPER_REGISTER | 0x05, 0 },                                      /* BC = CP   H        */
  { 1, 1, 0x06, OPER_REGISTER | 0x06, 0 },                                      /* BD = CP   L        */
  { 1, 1, 0x06, OPER_REGISTER | OPER_BRACKETS | 0x0C, 0 },                      /* BE = CP   (HL)     */
  { 1, 1, 0x06, OPER_REGISTER | 0x00, 0 },                                      /* BF = CP   A        */
  { 1, 1, 0x2D, OPER_FLAG | 0x03, 0 },                                          /* C0 = RET  NZ       */
  { 1, 1, 0x2A, OPER_REGISTER | 0x0A, 0 },                                      /* C1 = POP  BC       */
  { 3, 2, 0x1B, OPER_FLAG | 0x03, OPER_ADDRESS },                               /* C2 = JP   NZ, nn   */
  { 3, 1, 0x1B, OPER_ADDRESS, 0 },                                              /* C3 = JP   nn       */
  { 3, 2, 0x04, OPER_FLAG | 0x03, OPER_ADDRESS },                               /* C4 = CALL NZ, nn   */
  { 1, 1, 0x2B, OPER_REGISTER | 0x0A, 0 },                                      /* C5 = PUSH BC       */
  { 2, 2, 0x01, OPER_REGISTER | 0x00, OPER_8BIT },                              /* C6 = ADD  A, n     */
  { 1, 1, 0x3A, OPER_DIRECT | 0x00, 0 },                                        /* C7 = RST  00       */
  { 1, 1, 0x2D, OPER_FLAG | 0x07, 0 },                                          /* C8 = RET  Z        */
  { 1, 0, 0x2D, 0, 0 },                                                         /* C9 = RET           */
  { 3, 2, 0x1B, OPER_FLAG | 0x07, OPER_ADDRESS },                               /* CA = JP   Z, nn    */
  { 0, 0, 0, 0, 0 },                                                            /* CB -----> EXTENDED */
  { 3, 2, 0x04, OPER_FLAG | 0x07, OPER_ADDRESS },                               /* CC = CALL Z, nn    */
  { 3, 1, 0x04, OPER_ADDRESS, 0 },                                              /* CD = CALL nn       */
  { 2, 2, 0x00, OPER_REGISTER | 0x00, OPER_8BIT },                              /* CE = ADC  A, n     */
  { 1, 1, 0x3A, OPER_DIRECT | 0x01, 0 },                                        /* CF = RST  08       */
  { 1, 1, 0x2D, OPER_FLAG | 0x02, 0 },                                          /* D0 = RET  NC       */
  { 1, 1, 0x2A, OPER_REGISTER | 0x0B, 0 },                                      /* D1 = POP  DE       */
  { 3, 2, 0x1B, OPER_FLAG | 0x02, OPER_ADDRESS },                               /* D2 = JP   NC, nn   */
  { 2, 2, 0x27, OPER_8BIT | OPER_BRACKETS, OPER_REGISTER | 0x00 },              /* D3 = OUT  (n), A   */
  { 3, 2, 0x04, OPER_FLAG | 0x02, OPER_ADDRESS },                               /* D4 = CALL NC, nn   */
  { 1, 1, 0x2B, OPER_REGISTER | 0x0B, 0 },                                      /* D5 = PUSH DE       */
  { 2, 1, 0x42, OPER_8BIT, 0 },                                                 /* D6 = SUB  n        */
  { 1, 1, 0x3A, OPER_DIRECT | 0x02, 0 },                                        /* D7 = RST  10       */
  { 1, 1, 0x2D, OPER_FLAG | 0x00, 0 },                                          /* D8 = RET  C        */
  { 1, 0, 0x12, 0, 0 },                                                         /* D9 = EXX           */
  { 3, 2, 0x1B, OPER_FLAG | 0x00, OPER_ADDRESS },                               /* DA = JP   C, nn    */
  { 2, 2, 0x15, OPER_REGISTER | 0x00, OPER_8BIT | OPER_BRACKETS },              /* DB = IN   A, (n)   */
  { 3, 2, 0x04, OPER_FLAG | 0x00, OPER_ADDRESS },                               /* DC = CALL C, nn    */
  { 0, 0, 0, 0, 0 },                                                            /* DD -----> EXTENDED */
  { 2, 2, 0x3B, OPER_REGISTER | 0x00, OPER_8BIT },                              /* DE = SBC  A, n     */
  { 1, 1, 0x3A, OPER_DIRECT | 0x03, 0 },                                        /* DF = RST  18       */
  { 1, 1, 0x2D, OPER_FLAG | 0x06, 0 },                                          /* E0 = RET  PO       */
  { 1, 1, 0x2A, OPER_REGISTER | 0x0C, 0 },                                      /* E1 = POP  HL       */
  { 3, 2, 0x1B, OPER_FLAG | 0x06, OPER_ADDRESS },                               /* E2 = JP   PO, nn   */
  { 1, 2, 0x11, OPER_REGISTER | OPER_BRACKETS | 0x0D, OPER_REGISTER | 0x0C },   /* E3 = EX   (SP), HL */
  { 3, 2, 0x04, OPER_FLAG | 0x06, OPER_ADDRESS },                               /* E4 = CALL PO, nn   */
  { 1, 1, 0x2B, OPER_REGISTER | 0x0C, 0 },                                      /* E5 = PUSH HL       */
  { 2, 1, 0x02, OPER_8BIT, 0 },                                                 /* E6 = AND  n        */
  { 1, 1, 0x3A, OPER_DIRECT | 0x04, 0 },                                        /* E7 = RST  20       */
  { 1, 1, 0x2D, OPER_FLAG | 0x05, 0 },                                          /* E8 = RET  PE       */
  { 1, 1, 0x1B, OPER_REGISTER | OPER_BRACKETS | 0x0C, 0 },                      /* E9 = JP   (HL)     ---> BAD SYNTAX! */
  { 3, 2, 0x1B, OPER_FLAG | 0x05, OPER_ADDRESS },                               /* EA = JP   PE, nn   */
  { 1, 2, 0x11, OPER_REGISTER | 0x0C, OPER_REGISTER | 0x0B },                   /* EB = EX   HL, DE   ---> NO IX EXPANSION! */
  { 3, 2, 0x04, OPER_FLAG | 0x05, OPER_ADDRESS },                               /* EC = CALL PE, nn   */
  { 0, 0, 0, 0, 0 },                                                            /* ED -----> EXTENDED */
  { 2, 1, 0x43, OPER_8BIT, 0 },                                                 /* EE = XOR  n        */
  { 1, 1, 0x3A, OPER_DIRECT | 0x05, 0 },                                        /* EF = RST  28       */
  { 1, 1, 0x2D, OPER_FLAG | 0x04, 0 },                                          /* F0 = RET  P        */
  { 1, 1, 0x2A, OPER_REGISTER | 0x09, 0 },                                      /* F1 = POP  AF       */
  { 3, 2, 0x1B, OPER_FLAG | 0x04, OPER_ADDRESS },                               /* F2 = JP   P, nn    */
  { 1, 0, 0x0E, 0, 0 },                                                         /* F3 = DI            */
  { 3, 2, 0x04, OPER_FLAG | 0x04, OPER_ADDRESS },                               /* F4 = CALL P, nn    */
  { 1, 1, 0x2B, OPER_REGISTER | 0x09, 0 },                                      /* F5 = PUSH AF       */
  { 2, 1, 0x24, OPER_8BIT, 0 },                                                 /* F6 = OR   n        */
  { 1, 1, 0x3A, OPER_DIRECT | 0x06, 0 },                                        /* F7 = RST  30       */
  { 1, 1, 0x2D, OPER_FLAG | 0x01, 0 },                                          /* F8 = RET  M        */
  { 1, 2, 0x1D, OPER_REGISTER | 0x0D, OPER_REGISTER | 0x0C },                   /* F9 = LD   SP, HL   */
  { 3, 2, 0x1B, OPER_FLAG | 0x01, OPER_ADDRESS },                               /* FA = JP   M, nn    */
  { 1, 0, 0x10, 0, 0 },                                                         /* FB = EI            */
  { 3, 2, 0x04, OPER_FLAG | 0x01, OPER_ADDRESS },                               /* FC = CALL M, nn    */
  { 0, 0, 0, 0, 0 },                                                            /* FD -----> EXTENDED */
  { 2, 1, 0x06, OPER_8BIT, 0 },                                                 /* FE = CP   n        */
  { 1, 1, 0x3A, OPER_DIRECT | 0x07, 0 },                                        /* FF = RST  38       */
};

static struct _Z80Assem_s _CBCodes[256] = {
  { 2, 1, 0x32, OPER_REGISTER | 0x01, 0 },                                      /* CB00 = RLC  B        */
  { 2, 1, 0x32, OPER_REGISTER | 0x02, 0 },                                      /* CB01 = RLC  C        */
  { 2, 1, 0x32, OPER_REGISTER | 0x03, 0 },                                      /* CB02 = RLC  D        */
  { 2, 1, 0x32, OPER_REGISTER | 0x04, 0 },                                      /* CB03 = RLC  E        */
  { 2, 1, 0x32, OPER_REGISTER | 0x05, 0 },                                      /* CB04 = RLC  H        */
  { 2, 1, 0x32, OPER_REGISTER | 0x06, 0 },                                      /* CB05 = RLC  L        */
  { 2, 1, 0x32, OPER_REGISTER | OPER_BRACKETS | 0x0C, 0 },                      /* CB06 = RLC  (HL)     */
  { 2, 1, 0x32, OPER_REGISTER | 0x00, 0 },                                      /* CB07 = RLC  A        */
  { 2, 1, 0x37, OPER_REGISTER | 0x01, 0 },                                      /* CB08 = RRC  B        */
  { 2, 1, 0x37, OPER_REGISTER | 0x02, 0 },                                      /* CB09 = RRC  C        */
  { 2, 1, 0x37, OPER_REGISTER | 0x03, 0 },                                      /* CB0A = RRC  D        */
  { 2, 1, 0x37, OPER_REGISTER | 0x04, 0 },                                      /* CB0B = RRC  E        */
  { 2, 1, 0x37, OPER_REGISTER | 0x05, 0 },                                      /* CB0C = RRC  H        */
  { 2, 1, 0x37, OPER_REGISTER | 0x06, 0 },                                      /* CB0D = RRC  L        */
  { 2, 1, 0x37, OPER_REGISTER | OPER_BRACKETS | 0x0C, 0 },                      /* CB0E = RRC  (HL)     */
  { 2, 1, 0x37, OPER_REGISTER | 0x00, 0 },                                      /* CB0F = RRC  A        */
  { 2, 1, 0x30, OPER_REGISTER | 0x01, 0 },                                      /* CB10 = RL   B        */
  { 2, 1, 0x30, OPER_REGISTER | 0x02, 0 },                                      /* CB11 = RL   C        */
  { 2, 1, 0x30, OPER_REGISTER | 0x03, 0 },                                      /* CB12 = RL   D        */
  { 2, 1, 0x30, OPER_REGISTER | 0x04, 0 },                                      /* CB13 = RL   E        */
  { 2, 1, 0x30, OPER_REGISTER | 0x05, 0 },                                      /* CB14 = RL   H        */
  { 2, 1, 0x30, OPER_REGISTER | 0x06, 0 },                                      /* CB15 = RL   L        */
  { 2, 1, 0x30, OPER_REGISTER | OPER_BRACKETS | 0x0C, 0 },                      /* CB16 = RL   (HL)     */
  { 2, 1, 0x30, OPER_REGISTER | 0x00, 0 },                                      /* CB17 = RL   A        */
  { 2, 1, 0x35, OPER_REGISTER | 0x01, 0 },                                      /* CB18 = RR   B        */
  { 2, 1, 0x35, OPER_REGISTER | 0x02, 0 },                                      /* CB19 = RR   C        */
  { 2, 1, 0x35, OPER_REGISTER | 0x03, 0 },                                      /* CB1A = RR   D        */
  { 2, 1, 0x35, OPER_REGISTER | 0x04, 0 },                                      /* CB1B = RR   E        */
  { 2, 1, 0x35, OPER_REGISTER | 0x05, 0 },                                      /* CB1C = RR   H        */
  { 2, 1, 0x35, OPER_REGISTER | 0x06, 0 },                                      /* CB1D = RR   L        */
  { 2, 1, 0x35, OPER_REGISTER | OPER_BRACKETS | 0x0C, 0 },                      /* CB1E = RR   (HL)     */
  { 2, 1, 0x35, OPER_REGISTER | 0x00, 0 },                                      /* CB1F = RR   A        */
  { 2, 1, 0x3E, OPER_REGISTER | 0x01, 0 },                                      /* CB20 = SLA  B        */
  { 2, 1, 0x3E, OPER_REGISTER | 0x02, 0 },                                      /* CB21 = SLA  C        */
  { 2, 1, 0x3E, OPER_REGISTER | 0x03, 0 },                                      /* CB22 = SLA  D        */
  { 2, 1, 0x3E, OPER_REGISTER | 0x04, 0 },                                      /* CB23 = SLA  E        */
  { 2, 1, 0x3E, OPER_REGISTER | 0x05, 0 },                                      /* CB24 = SLA  H        */
  { 2, 1, 0x3E, OPER_REGISTER | 0x06, 0 },                                      /* CB25 = SLA  L        */
  { 2, 1, 0x3E, OPER_REGISTER | OPER_BRACKETS | 0x0C, 0 },                      /* CB26 = SLA  (HL)     */
  { 2, 1, 0x3E, OPER_REGISTER | 0x00, 0 },                                      /* CB27 = SLA  A        */
  { 2, 1, 0x40, OPER_REGISTER | 0x01, 0 },                                      /* CB28 = SRA  B        */
  { 2, 1, 0x40, OPER_REGISTER | 0x02, 0 },                                      /* CB29 = SRA  C        */
  { 2, 1, 0x40, OPER_REGISTER | 0x03, 0 },                                      /* CB2A = SRA  D        */
  { 2, 1, 0x40, OPER_REGISTER | 0x04, 0 },                                      /* CB2B = SRA  E        */
  { 2, 1, 0x40, OPER_REGISTER | 0x05, 0 },                                      /* CB2C = SRA  H        */
  { 2, 1, 0x40, OPER_REGISTER | 0x06, 0 },                                      /* CB2D = SRA  L        */
  { 2, 1, 0x40, OPER_REGISTER | OPER_BRACKETS | 0x0C, 0 },                      /* CB2E = SRA  (HL)     */
  { 2, 1, 0x40, OPER_REGISTER | 0x00, 0 },                                      /* CB2F = SRA  A        */
  { 2, 1, 0x3F, OPER_REGISTER | 0x01, 0 },                                      /* CB30 = SLL  B        */
  { 2, 1, 0x3F, OPER_REGISTER | 0x02, 0 },                                      /* CB31 = SLL  C        */
  { 2, 1, 0x3F, OPER_REGISTER | 0x03, 0 },                                      /* CB32 = SLL  D        */
  { 2, 1, 0x3F, OPER_REGISTER | 0x04, 0 },                                      /* CB33 = SLL  E        */
  { 2, 1, 0x3F, OPER_REGISTER | 0x05, 0 },                                      /* CB34 = SLL  H        */
  { 2, 1, 0x3F, OPER_REGISTER | 0x06, 0 },                                      /* CB35 = SLL  L        */
  { 2, 1, 0x3F, OPER_REGISTER | OPER_BRACKETS | 0x0C, 0 },                      /* CB36 = SLL  (HL)     */
  { 2, 1, 0x3F, OPER_REGISTER | 0x00, 0 },                                      /* CB37 = SLL  A        */
  { 2, 1, 0x41, OPER_REGISTER | 0x01, 0 },                                      /* CB38 = SRL  B        */
  { 2, 1, 0x41, OPER_REGISTER | 0x02, 0 },                                      /* CB39 = SRL  C        */
  { 2, 1, 0x41, OPER_REGISTER | 0x03, 0 },                                      /* CB3A = SRL  D        */
  { 2, 1, 0x41, OPER_REGISTER | 0x04, 0 },                                      /* CB3B = SRL  E        */
  { 2, 1, 0x41, OPER_REGISTER | 0x05, 0 },                                      /* CB3C = SRL  H        */
  { 2, 1, 0x41, OPER_REGISTER | 0x06, 0 },                                      /* CB3D = SRL  L        */
  { 2, 1, 0x41, OPER_REGISTER | OPER_BRACKETS | 0x0C, 0 },                      /* CB3E = SRL  (HL)     */
  { 2, 1, 0x41, OPER_REGISTER | 0x00, 0 },                                      /* CB3F = SRL  A        */
  { 2, 2, 0x03, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x01 },                   /* CB40 = BIT  0, B     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x02 },                   /* CB41 = BIT  0, C     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x03 },                   /* CB42 = BIT  0, D     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x04 },                   /* CB43 = BIT  0, E     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x05 },                   /* CB44 = BIT  0, H     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x06 },                   /* CB45 = BIT  0, L     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x00, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CB46 = BIT  0, (HL)  */
  { 2, 2, 0x03, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x00 },                   /* CB47 = BIT  0, A     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x01 },                   /* CB48 = BIT  1, B     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x02 },                   /* CB49 = BIT  1, C     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x03 },                   /* CB4A = BIT  1, D     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x04 },                   /* CB4B = BIT  1, E     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x05 },                   /* CB4C = BIT  1, H     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x06 },                   /* CB4D = BIT  1, L     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x01, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CB4E = BIT  1, (HL)  */
  { 2, 2, 0x03, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x00 },                   /* CB4F = BIT  1, A     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x01 },                   /* CB50 = BIT  2, B     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x02 },                   /* CB51 = BIT  2, C     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x03 },                   /* CB52 = BIT  2, D     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x04 },                   /* CB53 = BIT  2, E     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x05 },                   /* CB54 = BIT  2, H     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x06 },                   /* CB55 = BIT  2, L     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x02, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CB56 = BIT  2, (HL)  */
  { 2, 2, 0x03, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x00 },                   /* CB57 = BIT  2, A     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x01 },                   /* CB58 = BIT  3, B     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x02 },                   /* CB59 = BIT  3, C     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x03 },                   /* CB5A = BIT  3, D     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x04 },                   /* CB5B = BIT  3, E     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x05 },                   /* CB5C = BIT  3, H     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x06 },                   /* CB5D = BIT  3, L     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x03, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CB5E = BIT  3, (HL)  */
  { 2, 2, 0x03, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x00 },                   /* CB5F = BIT  3, A     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x01 },                   /* CB60 = BIT  4, B     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x02 },                   /* CB61 = BIT  4, C     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x03 },                   /* CB62 = BIT  4, D     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x04 },                   /* CB63 = BIT  4, E     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x05 },                   /* CB64 = BIT  4, H     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x06 },                   /* CB65 = BIT  4, L     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x04, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CB66 = BIT  4, (HL)  */
  { 2, 2, 0x03, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x00 },                   /* CB67 = BIT  4, A     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x01 },                   /* CB68 = BIT  5, B     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x02 },                   /* CB69 = BIT  5, C     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x03 },                   /* CB6A = BIT  5, D     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x04 },                   /* CB6B = BIT  5, E     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x05 },                   /* CB6C = BIT  5, H     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x06 },                   /* CB6D = BIT  5, L     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x05, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CB6E = BIT  5, (HL)  */
  { 2, 2, 0x03, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x00 },                   /* CB6F = BIT  5, A     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x01 },                   /* CB70 = BIT  6, B     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x02 },                   /* CB71 = BIT  6, C     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x03 },                   /* CB72 = BIT  6, D     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x04 },                   /* CB73 = BIT  6, E     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x05 },                   /* CB74 = BIT  6, H     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x06 },                   /* CB75 = BIT  6, L     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x06, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CB76 = BIT  6, (HL)  */
  { 2, 2, 0x03, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x00 },                   /* CB77 = BIT  6, A     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x01 },                   /* CB78 = BIT  7, B     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x02 },                   /* CB79 = BIT  7, C     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x03 },                   /* CB7A = BIT  7, D     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x04 },                   /* CB7B = BIT  7, E     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x05 },                   /* CB7C = BIT  7, H     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x06 },                   /* CB7D = BIT  7, L     */
  { 2, 2, 0x03, OPER_BITVALUE | 0x07, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CB7E = BIT  7, (HL)  */
  { 2, 2, 0x03, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x00 },                   /* CB7F = BIT  7, A     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x01 },                   /* CB80 = RES  0, B     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x02 },                   /* CB81 = RES  0, C     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x03 },                   /* CB82 = RES  0, D     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x04 },                   /* CB83 = RES  0, E     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x05 },                   /* CB84 = RES  0, H     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x06 },                   /* CB85 = RES  0, L     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x00, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CB86 = RES  0, (HL)  */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x00 },                   /* CB87 = RES  0, A     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x01 },                   /* CB88 = RES  1, B     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x02 },                   /* CB89 = RES  1, C     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x03 },                   /* CB8A = RES  1, D     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x04 },                   /* CB8B = RES  1, E     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x05 },                   /* CB8C = RES  1, H     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x06 },                   /* CB8D = RES  1, L     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x01, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CB8E = RES  1, (HL)  */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x00 },                   /* CB8F = RES  1, A     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x01 },                   /* CB90 = RES  2, B     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x02 },                   /* CB91 = RES  2, C     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x03 },                   /* CB92 = RES  2, D     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x04 },                   /* CB93 = RES  2, E     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x05 },                   /* CB94 = RES  2, H     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x06 },                   /* CB95 = RES  2, L     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x02, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CB96 = RES  2, (HL)  */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x00 },                   /* CB97 = RES  2, A     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x01 },                   /* CB98 = RES  3, B     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x02 },                   /* CB99 = RES  3, C     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x03 },                   /* CB9A = RES  3, D     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x04 },                   /* CB9B = RES  3, E     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x05 },                   /* CB9C = RES  3, H     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x06 },                   /* CB9D = RES  3, L     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x03, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CB9E = RES  3, (HL)  */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x00 },                   /* CB9F = RES  3, A     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x01 },                   /* CBA0 = RES  4, B     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x02 },                   /* CBA1 = RES  4, C     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x03 },                   /* CBA2 = RES  4, D     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x04 },                   /* CBA3 = RES  4, E     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x05 },                   /* CBA4 = RES  4, H     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x06 },                   /* CBA5 = RES  4, L     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x04, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CBA6 = RES  4, (HL)  */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x00 },                   /* CBA7 = RES  4, A     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x01 },                   /* CBA8 = RES  5, B     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x02 },                   /* CBA9 = RES  5, C     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x03 },                   /* CBAA = RES  5, D     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x04 },                   /* CBAB = RES  5, E     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x05 },                   /* CBAC = RES  5, H     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x06 },                   /* CBAD = RES  5, L     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x05, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CBAE = RES  5, (HL)  */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x00 },                   /* CBAF = RES  5, A     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x01 },                   /* CBB0 = RES  6, B     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x02 },                   /* CBB1 = RES  6, C     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x03 },                   /* CBB2 = RES  6, D     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x04 },                   /* CBB3 = RES  6, E     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x05 },                   /* CBB4 = RES  6, H     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x06 },                   /* CBB5 = RES  6, L     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x06, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CBB6 = RES  6, (HL)  */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x00 },                   /* CBB7 = RES  6, A     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x01 },                   /* CBB8 = RES  7, B     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x02 },                   /* CBB9 = RES  7, C     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x03 },                   /* CBBA = RES  7, D     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x04 },                   /* CBBB = RES  7, E     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x05 },                   /* CBBC = RES  7, H     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x06 },                   /* CBBD = RES  7, L     */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x07, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CBBE = RES  7, (HL)  */
  { 2, 2, 0x2C, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x00 },                   /* CBBF = RES  7, A     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x01 },                   /* CBC0 = SET  0, B     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x02 },                   /* CBC1 = SET  0, C     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x03 },                   /* CBC2 = SET  0, D     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x04 },                   /* CBC3 = SET  0, E     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x05 },                   /* CBC4 = SET  0, H     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x06 },                   /* CBC5 = SET  0, L     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x00, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CBC6 = SET  0, (HL)  */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x00, OPER_REGISTER | 0x00 },                   /* CBC7 = SET  0, A     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x01 },                   /* CBC8 = SET  1, B     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x02 },                   /* CBC9 = SET  1, C     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x03 },                   /* CBCA = SET  1, D     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x04 },                   /* CBCB = SET  1, E     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x05 },                   /* CBCC = SET  1, H     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x06 },                   /* CBCD = SET  1, L     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x01, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CBCE = SET  1, (HL)  */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x01, OPER_REGISTER | 0x00 },                   /* CBCF = SET  1, A     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x01 },                   /* CBD0 = SET  2, B     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x02 },                   /* CBD1 = SET  2, C     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x03 },                   /* CBD2 = SET  2, D     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x04 },                   /* CBD3 = SET  2, E     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x05 },                   /* CBD4 = SET  2, H     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x06 },                   /* CBD5 = SET  2, L     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x02, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CBD6 = SET  2, (HL)  */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x02, OPER_REGISTER | 0x00 },                   /* CBD7 = SET  2, A     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x01 },                   /* CBD8 = SET  3, B     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x02 },                   /* CBD9 = SET  3, C     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x03 },                   /* CBDA = SET  3, D     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x04 },                   /* CBDB = SET  3, E     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x05 },                   /* CBDC = SET  3, H     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x06 },                   /* CBDD = SET  3, L     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x03, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CBDE = SET  3, (HL)  */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x03, OPER_REGISTER | 0x00 },                   /* CBDF = SET  3, A     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x01 },                   /* CBE0 = SET  4, B     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x02 },                   /* CBE1 = SET  4, C     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x03 },                   /* CBE2 = SET  4, D     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x04 },                   /* CBE3 = SET  4, E     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x05 },                   /* CBE4 = SET  4, H     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x06 },                   /* CBE5 = SET  4, L     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x04, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CBE6 = SET  4, (HL)  */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x04, OPER_REGISTER | 0x00 },                   /* CBE7 = SET  4, A     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x01 },                   /* CBE8 = SET  5, B     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x02 },                   /* CBE9 = SET  5, C     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x03 },                   /* CBEA = SET  5, D     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x04 },                   /* CBEB = SET  5, E     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x05 },                   /* CBEC = SET  5, H     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x06 },                   /* CBED = SET  5, L     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x05, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CBEE = SET  5, (HL)  */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x05, OPER_REGISTER | 0x00 },                   /* CBEF = SET  5, A     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x01 },                   /* CBF0 = SET  6, B     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x02 },                   /* CBF1 = SET  6, C     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x03 },                   /* CBF2 = SET  6, D     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x04 },                   /* CBF3 = SET  6, E     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x05 },                   /* CBF4 = SET  6, H     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x06 },                   /* CBF5 = SET  6, L     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x06, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CBF6 = SET  6, (HL)  */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x06, OPER_REGISTER | 0x00 },                   /* CBF7 = SET  6, A     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x01 },                   /* CBF8 = SET  7, B     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x02 },                   /* CBF9 = SET  7, C     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x03 },                   /* CBFA = SET  7, D     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x04 },                   /* CBFB = SET  7, E     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x05 },                   /* CBFC = SET  7, H     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x06 },                   /* CBFD = SET  7, L     */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x07, OPER_REGISTER | OPER_BRACKETS | 0x0C },   /* CBFE = SET  7, (HL)  */
  { 2, 2, 0x3D, OPER_BITVALUE | 0x07, OPER_REGISTER | 0x00 },                   /* CBFF = SET  7, A     */
};

static struct _Z80Assem_s _EDCodes[256] = {
  { 2, 0, 0x23, 0, 0 },                                                         /* ED00 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED01 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED02 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED03 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED04 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED05 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED06 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED07 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED08 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED09 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED0A = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED0B = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED0C = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED0D = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED0E = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED0F = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED10 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED11 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED12 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED13 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED14 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED15 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED16 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED17 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED18 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED19 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED1A = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED1B = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED1C = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED1D = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED1E = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED1F = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED20 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED21 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED22 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED23 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED24 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED25 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED26 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED27 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED28 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED29 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED2A = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED2B = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED2C = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED2D = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED2E = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED2F = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED30 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED31 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED32 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED33 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED34 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED35 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED36 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED37 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED38 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED39 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED3A = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED3B = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED3C = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED3D = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED3E = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED3F = NOP          -*/
  { 2, 2, 0x15, OPER_REGISTER | 0x01, OPER_REGISTER | OPER_BRACKETS | 0x02 },   /* ED40 = IN   B, (C)   */
  { 2, 2, 0x27, OPER_REGISTER | OPER_BRACKETS | 0x02, OPER_REGISTER | 0x01 },   /* ED41 = OUT  (C), B   */
  { 2, 2, 0x3B, OPER_REGISTER | 0x0C, OPER_REGISTER | 0x0A },                   /* ED42 = SBC  HL, BC   */
  { 4, 2, 0x1D, OPER_16BIT | OPER_BRACKETS, OPER_REGISTER | 0x0A },             /* ED43 = LD   (nn), BC */
  { 2, 0, 0x22, 0, 0 },                                                         /* ED44 = NEG           */
  { 2, 0, 0x2F, 0, 0 },                                                         /* ED45 = RETN          */
  { 2, 1, 0x14, OPER_BITVALUE | 0x00, 0 },                                      /* ED46 = IM   0        */
  { 2, 2, 0x1D, OPER_REGISTER | 0x07, OPER_REGISTER | 0x00 },                   /* ED47 = LD   I, A     */
  { 2, 2, 0x15, OPER_REGISTER | 0x02, OPER_REGISTER | OPER_BRACKETS | 0x02 },   /* ED48 = IN   C, (C)   */
  { 2, 2, 0x27, OPER_REGISTER | OPER_BRACKETS | 0x02, OPER_REGISTER | 0x02 },   /* ED49 = OUT  (C), C   */
  { 2, 2, 0x00, OPER_REGISTER | 0x0C, OPER_REGISTER | 0x0A },                   /* ED4A = ADC  HL, BC   */
  { 4, 2, 0x1D, OPER_REGISTER | 0x0A, OPER_16BIT | OPER_BRACKETS },             /* ED4B = LD   BC, (nn) */
  { 2, 0, 0x22, 0, 0 },                                                         /* ED4C = NEG          -*/
  { 2, 0, 0x2E, 0, 0 },                                                         /* ED4D = RETI          */
  { 2, 1, 0x14, OPER_BITVALUE | 0x00, 0 },                                      /* ED4E = IM   0       -*/
  { 2, 2, 0x1D, OPER_REGISTER | 0x08, OPER_REGISTER | 0x00 },                   /* ED4F = LD   R, A     */
  { 2, 2, 0x15, OPER_REGISTER | 0x03, OPER_REGISTER | OPER_BRACKETS | 0x02 },   /* ED50 = IN   D, (C)   */
  { 2, 2, 0x27, OPER_REGISTER | OPER_BRACKETS | 0x02, OPER_REGISTER | 0x03 },   /* ED51 = OUT  (C), D   */
  { 2, 2, 0x3B, OPER_REGISTER | 0x0C, OPER_REGISTER | 0x0B },                   /* ED52 = SBC  HL, DE   */
  { 4, 2, 0x1D, OPER_16BIT | OPER_BRACKETS, OPER_REGISTER | 0x0B },             /* ED53 = LD   (nn), DE */
  { 2, 0, 0x22, 0, 0 },                                                         /* ED54 = NEG          -*/
  { 2, 0, 0x2D, 0, 0 },                                                         /* ED55 = RET          -*/
  { 2, 1, 0x14, OPER_BITVALUE | 0x01, 0 },                                      /* ED56 = IM   1        */
  { 2, 2, 0x1D, OPER_REGISTER | 0x00, OPER_REGISTER | 0x07 },                   /* ED57 = LD   A, I     */
  { 2, 2, 0x15, OPER_REGISTER | 0x04, OPER_REGISTER | OPER_BRACKETS | 0x02 },   /* ED58 = IN   E, (C)   */
  { 2, 2, 0x27, OPER_REGISTER | OPER_BRACKETS | 0x02, OPER_REGISTER | 0x04 },   /* ED59 = OUT  (C), E   */
  { 2, 2, 0x00, OPER_REGISTER | 0x0C, OPER_REGISTER | 0x0B },                   /* ED5A = ADC  HL, DE   */
  { 4, 2, 0x1D, OPER_REGISTER | 0x0B, OPER_16BIT | OPER_BRACKETS },             /* ED5B = LD   DE, (nn) */
  { 2, 0, 0x22, 0, 0 },                                                         /* ED5C = NEG          -*/
  { 2, 0, 0x2D, 0, 0 },                                                         /* ED5D = RET          -*/
  { 2, 1, 0x14, OPER_BITVALUE | 0x02, 0 },                                      /* ED5E = IM   2        */
  { 2, 2, 0x1D, OPER_REGISTER | 0x00, OPER_REGISTER | 0x08 },                   /* ED5F = LD   A, R     */
  { 2, 2, 0x15, OPER_REGISTER | 0x05, OPER_REGISTER | OPER_BRACKETS | 0x02 },   /* ED60 = IN   H, (C)   */
  { 2, 2, 0x27, OPER_REGISTER | OPER_BRACKETS | 0x02, OPER_REGISTER | 0x05 },   /* ED61 = OUT  (C), H   */
  { 2, 2, 0x3B, OPER_REGISTER | 0x0C, OPER_REGISTER | 0x0C },                   /* ED62 = SBC  HL, HL   */
  { 4, 2, 0x1D, OPER_16BIT | OPER_BRACKETS, OPER_REGISTER | 0x0C },             /* ED63 = LD   (nn),HL  */
  { 2, 0, 0x22, 0, 0 },                                                         /* ED64 = NEG          -*/
  { 2, 0, 0x2D, 0, 0 },                                                         /* ED65 = RET          -*/
  { 2, 1, 0x14, OPER_BITVALUE | 0x01, 0 },                                      /* ED66 = IM   1       -*/
  { 2, 0, 0x39, 0, 0 },                                                         /* ED67 = RRD           */
  { 2, 2, 0x15, OPER_REGISTER | 0x06, OPER_REGISTER | OPER_BRACKETS | 0x02 },   /* ED68 = IN   L, (C)   */
  { 2, 2, 0x27, OPER_REGISTER | OPER_BRACKETS | 0x02, OPER_REGISTER | 0x06 },   /* ED69 = OUT  (C), L   */
  { 2, 2, 0x00, OPER_REGISTER | 0x0C, OPER_REGISTER | 0x0C },                   /* ED6A = ADC  HL, HL   */
  { 4, 2, 0x1D, OPER_REGISTER | 0x0C, OPER_16BIT | OPER_BRACKETS },             /* ED6B = LD   HL, (nn) */
  { 2, 0, 0x22, 0, 0 },                                                         /* ED6C = NEG          -*/
  { 2, 0, 0x2D, 0, 0 },                                                         /* ED6D = RET          -*/
  { 2, 1, 0x14, OPER_BITVALUE | 0x00, 0 },                                      /* ED6E = IM   0       -*/
  { 2, 0, 0x34, 0, 0 },                                                         /* ED6F = RLD           */
  { 2, 1, 0x15, OPER_REGISTER | OPER_BRACKETS | 0x02 },                         /* ED70 = IN   (C)     -*/
  { 2, 2, 0x27, OPER_REGISTER | OPER_BRACKETS | 0x02, OPER_BITVALUE | 0x00 },   /* ED71 = OUT  (C), 0  -*/
  { 2, 2, 0x3B, OPER_REGISTER | 0x0C, OPER_REGISTER | 0x0D },                   /* ED72 = SBC  HL, SP   */
  { 4, 2, 0x1D, OPER_16BIT | OPER_BRACKETS, OPER_REGISTER | 0x0D },             /* ED73 = LD   (nn), SP */
  { 2, 0, 0x22, 0, 0 },                                                         /* ED74 = NEG          -*/
  { 2, 0, 0x2D, 0, 0 },                                                         /* ED75 = RET          -*/
  { 2, 1, 0x14, OPER_BITVALUE | 0x01, 0 },                                      /* ED76 = IM   1       -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED77 = NOP          -*/
  { 2, 2, 0x15, OPER_REGISTER | 0x00, OPER_REGISTER | OPER_BRACKETS | 0x02 },   /* ED78 = IN   A, (C)   */
  { 2, 2, 0x27, OPER_REGISTER | OPER_BRACKETS | 0x02, OPER_REGISTER | 0x00 },   /* ED79 = OUT  (C), A   */
  { 2, 2, 0x00, OPER_REGISTER | 0x0C, OPER_REGISTER | 0x0D },                   /* ED7A = ADC  HL, SP   */
  { 4, 2, 0x1D, OPER_REGISTER | 0x0D, OPER_16BIT | OPER_BRACKETS },             /* ED7B = LD   SP, (nn) */
  { 2, 0, 0x22, 0, 0 },                                                         /* ED7C = NEG          -*/
  { 2, 0, 0x2D, 0, 0 },                                                         /* ED7D = RET          -*/
  { 2, 1, 0x14, OPER_BITVALUE | 0x02, 0 },                                      /* ED7E = IM   2       -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED7F = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED80 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED81 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED82 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED83 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED84 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED85 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED86 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED87 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED88 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED89 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED8A = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED8B = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED8C = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED8D = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED8E = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED8F = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED90 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED91 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED92 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED93 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED94 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED95 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED96 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED97 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED98 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED99 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED9A = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED9B = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED9C = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED9D = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED9E = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* ED9F = NOP          -*/
  { 2, 0, 0x20, 0, 0 },                                                         /* EDA0 = LDI           */
  { 2, 0, 0x09, 0, 0 },                                                         /* EDA1 = CPI           */
  { 2, 0, 0x19, 0, 0 },                                                         /* EDA2 = INI           */
  { 2, 0, 0x29, 0, 0 },                                                         /* EDA3 = OUTI          */
  { 2, 0, 0x23, 0, 0 },                                                         /* EDA4 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDA5 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDA6 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDA7 = NOP          -*/
  { 2, 0, 0x1E, 0, 0 },                                                         /* EDA8 = LDD           */
  { 2, 0, 0x07, 0, 0 },                                                         /* EDA9 = CPD           */
  { 2, 0, 0x17, 0, 0 },                                                         /* EDAA = IND           */
  { 2, 0, 0x28, 0, 0 },                                                         /* EDAB = OUTD          */
  { 2, 0, 0x23, 0, 0 },                                                         /* EDAC = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDAD = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDAE = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDAF = NOP          -*/
  { 2, 0, 0x21, 0, 0 },                                                         /* EDB0 = LDIR          */
  { 2, 0, 0x0A, 0, 0 },                                                         /* EDB1 = CPIR          */
  { 2, 0, 0x1A, 0, 0 },                                                         /* EDB2 = INIR          */
  { 2, 0, 0x26, 0, 0 },                                                         /* EDB3 = OTIR          */
  { 2, 0, 0x23, 0, 0 },                                                         /* EDB4 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDB5 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDB6 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDB7 = NOP          -*/
  { 2, 0, 0x1F, 0, 0 },                                                         /* EDB8 = LDDR          */
  { 2, 0, 0x08, 0, 0 },                                                         /* EDB9 = CPDR          */
  { 2, 0, 0x18, 0, 0 },                                                         /* EDBA = INDR          */
  { 2, 0, 0x25, 0, 0 },                                                         /* EDBB = OTDR          */
  { 2, 0, 0x23, 0, 0 },                                                         /* EDBC = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDBD = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDBE = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDBF = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDC0 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDC1 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDC2 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDC3 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDC4 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDC5 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDC6 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDC7 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDC8 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDC9 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDCA = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDCB = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDCC = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDCD = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDCE = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDCF = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDD0 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDD1 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDD2 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDD3 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDD4 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDD5 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDD6 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDD7 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDD8 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDD9 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDDA = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDDB = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDDC = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDDD = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDDE = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDDF = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDE0 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDE1 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDE2 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDE3 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDE4 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDE5 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDE6 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDE7 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDE8 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDE9 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDEA = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDEB = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDEC = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDED = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDEE = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDEF = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDF0 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDF1 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDF2 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDF3 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDF4 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDF5 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDF6 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDF7 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDF8 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDF9 = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDFA = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDFB = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDFC = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDFD = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDFE = NOP          -*/
  { 2, 0, 0x23, 0, 0 },                                                         /* EDFF = NOP          -*/
};

/**********************************************************************************************************************************/
/* Define the Spectrum main system variables. The area is [23552,23733] (or [0x5C00,0x5CB5])                                      */
/* These are used to remap 16-bit operands if `_ExpandSystemVariables' is TRUE.                                                   */
/**********************************************************************************************************************************/

#define SYSVARS_START 23552
#define SYSVARS_AFTER 23734

static struct _SystemVariables_s
{
  word                _Address;                                                                               /* Starting address */
  char               *_Name;                                                                                   /* Name as printed */
}             _SystemVariables[] = {
  { 23552, "KSTATE" }, { 23560, "LASTK"  }, { 23561, "REPDEL" }, { 23562, "REPPER" }, { 23563, "DEFADD" }, { 23565, "KDATA"  },
  { 23566, "TVDATA" }, { 23568, "STRMS"  }, { 23606, "CHARS"  }, { 23608, "RASP"   }, { 23609, "PIP"    }, { 23610, "ERRNR"  },
  { 23611, "FLAGS"  }, { 23612, "TVFLAG" }, { 23613, "ERRSP"  }, { 23615, "LISTSP" }, { 23617, "MODE"   }, { 23618, "NEWPPC" },
  { 23620, "NSPPC"  }, { 23621, "PPC"    }, { 23623, "SUBPPC" }, { 23624, "BORDCR" }, { 23625, "EPPC"   }, { 23627, "VARS"   },
  { 23629, "DEST"   }, { 23631, "CHANS"  }, { 23633, "CURCHL" }, { 23635, "PROG"   }, { 23637, "NXTLIN" }, { 23639, "DATADD" },
  { 23641, "ELINE"  }, { 23643, "KCUR"   }, { 23645, "CHADD"  }, { 23647, "XPTR"   }, { 23649, "WORKSP" }, { 23651, "STKBOT" },
  { 23653, "STKEND" }, { 23655, "BREG"   }, { 23656, "MEM"    }, { 23658, "FLAGS2" }, { 23659, "DFSZ"   }, { 23660, "S_TOP"  },
  { 23662, "OLDPPC" }, { 23664, "OSPPC"  }, { 23665, "FLAGX"  }, { 23666, "STRLEN" }, { 23668, "TADDR"  }, { 23670, "SEED"   },
  { 23672, "FRAMES" }, { 23675, "UDG"    }, { 23677, "COORDX" }, { 23678, "COORDY" }, { 23679, "PPOSN"  }, { 23680, "PRCC"   },
  { 23681, "NONAME" }, { 23682, "ECHOE"  }, { 23684, "DFCC"   }, { 23686, "DFCCL"  }, { 23688, "SPOSUC" }, { 23689, "SPOSUL" },
  { 23690, "SPOSLC" }, { 23691, "SPOSLL" }, { 23692, "SCRCT"  }, { 23693, "ATTRP"  }, { 23694, "MASKP"  }, { 23695, "ATTRT"  },
  { 23696, "MASKT"  }, { 23697, "PFLAG"  }, { 23698, "MEMBOT" }, { 23728, "NMIADD" }, { 23730, "RAMTOP" }, { 23732, "PRAMT"  },
  { 23734, "------" }
};
static short _NumSystemVariables = sizeof (_SystemVariables) / sizeof (struct _SystemVariables_s);

/**********************************************************************************************************************************/
/* Define the Spectrum ROM RST literal values (RST 08 and RST 28).                                                                */
/* These are used if `_RemapSpectrumROM' is TRUE.                                                                                 */
/**********************************************************************************************************************************/

static char *_RST08Codes[] = {
  "OK",                       /* 0 00 */
  "NEXT without FOR",         /* 1 01 */
  "Variable not found",       /* 2 02 */
  "Subscript wrong",          /* 3 03 */
  "Out of memory",            /* 4 04 */
  "Out of screen",            /* 5 05 */
  "Number too big",           /* 6 06 */
  "RETURN without GO SUB",    /* 7 07 */
  "<undefined>",              /* 8 08 */
  "STOP statement",           /* 9 09 */
  "Invalid argument",         /* A 10 */
  "Integer out of range",     /* B 11 */
  "Nonsense in BASIC",        /* C 12 */
  "BREAK - CONT repeats",     /* D 13 */
  "Out of data",              /* E 14 */
  "Invalid file name",        /* F 15 */
  "No room for line",         /* G 16 */
  "STOP in INPUT",            /* H 17 */
  "FOR without NEXT",         /* I 18 */
  "Invalid I/O device",       /* J 19 */
  "Invalid Colour",           /* K 20 */
  "BREAK into program",       /* L 21 */
  "RAMTOP no good",           /* M 22 */
  "Statement lost",           /* N 23 */
  "Invalid stream",           /* O 24 */
  "FN without DEF",           /* P 25 */
  "Parameter error",          /* Q 26 */
  "Tape loading error"        /* R 27 */
};
static short _NumRST08Codes = sizeof (_RST08Codes) / sizeof (char *);

static char *_RST28Codes[256] = {
  "jump_true",                /* 00 */
  "exchange",                 /* 01 */
  "delete",                   /* 02 */
  "subtract",                 /* 03 */
  "multiply",                 /* 04 */
  "division",                 /* 05 */
  "to_power",                 /* 06 */
  "or",                       /* 07 */
  "no_&_no",                  /* 08 */
  "no_!_eql",                 /* 09 */
  "no_gr_eq",                 /* 0A */
  "nos_neql",                 /* 0B */
  "no_grtr",                  /* 0C */
  "no_less",                  /* 0D */
  "nos_eql",                  /* 0E */
  "addition",                 /* 0F */
  "str_&_no",                 /* 10 */
  "str_l_eql",                /* 11 */
  "str_gr_eq",                /* 12 */
  "strs_neql",                /* 13 */
  "str_grtr",                 /* 14 */
  "str_less",                 /* 15 */
  "strs_eql",                 /* 16 */
  "strs_add",                 /* 17 */
  "val$",                     /* 18 */
  "usr_$",                    /* 19 */
  "read_in",                  /* 1A */
  "negate",                   /* 1B */
  "code",                     /* 1C */
  "val",                      /* 1D */
  "len",                      /* 1E */
  "sin",                      /* 1F */
  "cos",                      /* 20 */
  "tan",                      /* 21 */
  "asn",                      /* 22 */
  "acs",                      /* 23 */
  "atn",                      /* 24 */
  "ln",                       /* 25 */
  "exp",                      /* 26 */
  "int",                      /* 27 */
  "sqr",                      /* 28 */
  "sgn",                      /* 29 */
  "abs",                      /* 2A */
  "peek",                     /* 2B */
  "in",                       /* 2C */
  "usr_no",                   /* 2D */
  "str$",                     /* 2E */
  "chr$",                     /* 2F */
  "not",                      /* 30 */
  "duplicate",                /* 31 */
  "n_mod_m",                  /* 32 */
  "jump",                     /* 33 */
  "stk_data",                 /* 34 */
  "dec_jr_nz",                /* 35 */
  "less_0",                   /* 36 */
  "greater_0",                /* 37 */
  "end_calc",                 /* 38 */
  "get_argt",                 /* 39 */
  "truncate",                 /* 3A */
  "fp_calc_2",                /* 3B */
  "e_to_fp",                  /* 3C */
  "re_stack",                 /* 3D */
  NULL,                       /* 3E */
  NULL,                       /* 3F */
  NULL,                       /* 40 */
  NULL,                       /* 41 */
  NULL,                       /* 42 */
  NULL,                       /* 43 */
  NULL,                       /* 44 */
  NULL,                       /* 45 */
  NULL,                       /* 46 */
  NULL,                       /* 47 */
  NULL,                       /* 48 */
  NULL,                       /* 49 */
  NULL,                       /* 4A */
  NULL,                       /* 4B */
  NULL,                       /* 4C */
  NULL,                       /* 4D */
  NULL,                       /* 4E */
  NULL,                       /* 4F */
  NULL,                       /* 50 */
  NULL,                       /* 51 */
  NULL,                       /* 52 */
  NULL,                       /* 53 */
  NULL,                       /* 54 */
  NULL,                       /* 55 */
  NULL,                       /* 56 */
  NULL,                       /* 57 */
  NULL,                       /* 58 */
  NULL,                       /* 59 */
  NULL,                       /* 5A */
  NULL,                       /* 5B */
  NULL,                       /* 5C */
  NULL,                       /* 5D */
  NULL,                       /* 5E */
  NULL,                       /* 5F */
  NULL,                       /* 60 */
  NULL,                       /* 61 */
  NULL,                       /* 62 */
  NULL,                       /* 63 */
  NULL,                       /* 64 */
  NULL,                       /* 65 */
  NULL,                       /* 66 */
  NULL,                       /* 67 */
  NULL,                       /* 68 */
  NULL,                       /* 69 */
  NULL,                       /* 6A */
  NULL,                       /* 6B */
  NULL,                       /* 6C */
  NULL,                       /* 6D */
  NULL,                       /* 6E */
  NULL,                       /* 6F */
  NULL,                       /* 70 */
  NULL,                       /* 71 */
  NULL,                       /* 72 */
  NULL,                       /* 73 */
  NULL,                       /* 74 */
  NULL,                       /* 75 */
  NULL,                       /* 76 */
  NULL,                       /* 77 */
  NULL,                       /* 78 */
  NULL,                       /* 79 */
  NULL,                       /* 7A */
  NULL,                       /* 7B */
  NULL,                       /* 7C */
  NULL,                       /* 7D */
  NULL,                       /* 7E */
  NULL,                       /* 7F */
  NULL,                       /* 80 */
  NULL,                       /* 81 */
  NULL,                       /* 82 */
  NULL,                       /* 83 */
  NULL,                       /* 84 */
  NULL,                       /* 85 */
  "series_06",                /* 86 */
  NULL,                       /* 87 */
  "series_08",                /* 88 */
  NULL,                       /* 89 */
  NULL,                       /* 8A */
  NULL,                       /* 8B */
  "series_0C",                /* 8C */
  NULL,                       /* 8D */
  NULL,                       /* 8E */
  NULL,                       /* 8F */
  NULL,                       /* 90 */
  NULL,                       /* 91 */
  NULL,                       /* 92 */
  NULL,                       /* 93 */
  NULL,                       /* 94 */
  NULL,                       /* 95 */
  NULL,                       /* 96 */
  NULL,                       /* 97 */
  NULL,                       /* 98 */
  NULL,                       /* 99 */
  NULL,                       /* 9A */
  NULL,                       /* 9B */
  NULL,                       /* 9C */
  NULL,                       /* 9D */
  NULL,                       /* 9E */
  NULL,                       /* 9F */
  "stk_zero",                 /* A0 */
  "stk_one",                  /* A1 */
  "stk_half",                 /* A2 */
  "stk_pi/2",                 /* A3 */
  "stk_ten",                  /* A4 */
  NULL,                       /* A5 */
  NULL,                       /* A6 */
  NULL,                       /* A7 */
  NULL,                       /* A8 */
  NULL,                       /* A9 */
  NULL,                       /* AA */
  NULL,                       /* AB */
  NULL,                       /* AC */
  NULL,                       /* AD */
  NULL,                       /* AE */
  NULL,                       /* AF */
  NULL,                       /* B0 */
  NULL,                       /* B1 */
  NULL,                       /* B2 */
  NULL,                       /* B3 */
  NULL,                       /* B4 */
  NULL,                       /* B5 */
  NULL,                       /* B6 */
  NULL,                       /* B7 */
  NULL,                       /* B8 */
  NULL,                       /* B9 */
  NULL,                       /* BA */
  NULL,                       /* BB */
  NULL,                       /* BC */
  NULL,                       /* BD */
  NULL,                       /* BE */
  NULL,                       /* BF */
  "st_mem_0",                 /* C0 */
  "st_mem_1",                 /* C1 */
  "st_mem_2",                 /* C2 */
  "st_mem_3",                 /* C3 */
  "st_mem_4",                 /* C4 */
  "st_mem_5",                 /* C5 */
  NULL,                       /* C6 */
  NULL,                       /* C7 */
  NULL,                       /* C8 */
  NULL,                       /* C9 */
  NULL,                       /* CA */
  NULL,                       /* CB */
  NULL,                       /* CC */
  NULL,                       /* CD */
  NULL,                       /* CE */
  NULL,                       /* CF */
  NULL,                       /* D0 */
  NULL,                       /* D1 */
  NULL,                       /* D2 */
  NULL,                       /* D3 */
  NULL,                       /* D4 */
  NULL,                       /* D5 */
  NULL,                       /* D6 */
  NULL,                       /* D7 */
  NULL,                       /* D8 */
  NULL,                       /* D9 */
  NULL,                       /* DA */
  NULL,                       /* DB */
  NULL,                       /* DC */
  NULL,                       /* DD */
  NULL,                       /* DE */
  NULL,                       /* DF */
  "get_mem_0",                /* E0 */
  "get_mem_1",                /* E1 */
  "get_mem_2",                /* E2 */
  "get_mem_3",                /* E3 */
  "get_mem_4",                /* E4 */
  "get_mem_5",                /* E5 */
  NULL,                       /* E6 */
  NULL,                       /* E7 */
  NULL,                       /* E8 */
  NULL,                       /* E9 */
  NULL,                       /* EA */
  NULL,                       /* EB */
  NULL,                       /* EC */
  NULL,                       /* ED */
  NULL,                       /* EE */
  NULL,                       /* EF */
  NULL,                       /* F0 */
  NULL,                       /* F1 */
  NULL,                       /* F2 */
  NULL,                       /* F3 */
  NULL,                       /* F4 */
  NULL,                       /* F5 */
  NULL,                       /* F6 */
  NULL,                       /* F7 */
  NULL,                       /* F8 */
  NULL,                       /* F9 */
  NULL,                       /* FA */
  NULL,                       /* FB */
  NULL,                       /* FC */
  NULL,                       /* FD */
  NULL,                       /* FE */
  NULL                        /* FF */
};

/**********************************************************************************************************************************/
/* These are the variables needed for the Hex Editor.                                                                             */
/**********************************************************************************************************************************/

#define FINDNOMATCH  0xFFFFFFFF

static struct ComboInfo  _EditHWindow;
static struct ComboInfo  _EditAWindow;
static bool              _EditWhichWindow;
static short             _EditCurX = -1;
static short             _EditCurY = -1;
static bool              _EditCurHalf;
static word              _EditCurIndex;
static word              _EditCurAddress;
static bool              _EditChanged;

/**********************************************************************************************************************************/
/* Prototypes of the static functions.                                                                                            */
/**********************************************************************************************************************************/

static double   _FPSpecToDouble     (struct _SpectrumFP_s *FPNumber);
static bool     _MatchLabel         (word Address, bool MatchSpectrumROM, char *Symbol);
static void     _GetDisassemblyLine (struct _InstructionMemory_s *Address, char *Disassembled);
static void     _ScrollBack         (struct _InstructionMemory_s *FirstAddress, struct _InstructionMemory_s *StartAddress,
                                     bool FullPage);
static bool     _GetAddress         (word *Address, bool IsLoadAddress);
static void     _FlushExtLabels     (void);
static void     _ReadLabelFile      (char *FileName, bool SpectrumROM);
static bool     _LoadExtLabels      (bool Merge);
static bool     _SaveExtLabels      (bool SaveAs);
static comboent _AddExtSymbol       (word DefaultAddress);
static bool     _DeleteExtSymbol    (comboent WhichSymbol);
static bool     _SetupSymbolWindow  (void);
static bool     _EditSymbolTable    (word CurrentAddress);
static bool     _GetSaveRange       (word *StartAddress, word *EndAddress);
static dword    _EditFind           (bool ReInit, dword BlockLength);

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE INTERNAL DISASSEMBLER ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

static double _FPSpecToDouble (struct _SpectrumFP_s *FPNumber)

/**********************************************************************************************************************************/
/* Pre   : `FPNumber' points to the Spectrum-style 5-byte floating point number.                                                  */
/* Post  : the `FPNumber' is converted to double and returned.                                                                    */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  double FPValue;
  double Power;
  double Raised;
  double Result;

  if (FPNumber->_Exponent == 0x00)                                                                        /* Small integer form ? */
  {
    FPValue = (double)FPNumber->_Mantissa[1] + 256 * (double)FPNumber->_Mantissa[2];
    if (FPNumber->_Mantissa[0])                                                                                     /* Negative ? */
      FPValue -= (double)65536;                                                                 /* (Maintain bug in Spectrum ROM) */
  }
  else
  {
    Power = pow (2, ((double)FPNumber->_Exponent - 129));
    if (FPNumber->_Mantissa[0] & 0x80)                                                              /* Determine negative numbers */
      Power *= -1;
    Raised = 128;
    Result = (FPNumber->_Mantissa[0] & 0x7F) / Raised;
    Raised *= 256;
    Result += (FPNumber->_Mantissa[1]) / Raised;
    Raised *= 256;
    Result += (FPNumber->_Mantissa[2]) / Raised;
    Raised *= 256;
    Result += (FPNumber->_Mantissa[3]) / Raised;
    FPValue = Power * (1 + Result);
  }
  return (FPValue);
}

static bool _MatchLabel (word Address, bool MatchSpectrumROM, char *Symbol)

/**********************************************************************************************************************************/
/* Pre   : `Address' is the address to match, `MatchSpectrumROM' is TRUE only when inserting the instruction address label.       */
/* Post  : If the address is found in the internal or external label file (in that order), its name is copied in `Symbol' and     */
/*         TRUE is returned.                                                                                                      */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  int register Cnt;
  bool         Done;
  bool         TestSpectrumROM;

  TestSpectrumROM = (Address < 16384 && _ExpandSpectrumROM);
  if (MatchSpectrumROM && !_IsSpectrumROM)                                                /* Doing a instruction address symbol ? */
    TestSpectrumROM = FALSE;
  if (TestSpectrumROM)
  {
    for (Cnt = 0, Done = FALSE ; !Done && Cnt < _NumSpecLabels ; Cnt ++)                    /* Match all 48K Spectrum ROM symbols */
      if (Address == _SpecLabels[Cnt]->_Address)
      {
        strcpy (Symbol, _SpecLabels[Cnt]->_Name);
        return (TRUE);
      }
      else if (Address < _SpecLabels[Cnt]->_Address)
        Done = TRUE;
  }
  if (_HaveLabels)
    for (Cnt = 0, Done = FALSE ; !Done && Cnt < _NumLabels ; Cnt ++)                            /* Match all user defined symbols */
      if (Address == _Labels[Cnt]->_Address)
      {
        strcpy (Symbol, _Labels[Cnt]->_Name);
        return (TRUE);
      }
      else if (Address < _Labels[Cnt]->_Address)
        Done = TRUE;
  return (FALSE);
}

static void _GetDisassemblyLine (struct _InstructionMemory_s *Address, char *Disassembled)

/**********************************************************************************************************************************/
/* Pre   : `Address' holds the index into the object code where the current instruction starts.                                   */
/* Post  : If `_CountBytesOnly' is TRUE, only `Address' is advanced to point to the next instruction.                             */
/*         If `_CountBytesOnly' is FALSE, the current instruction is decoded into `Disassembled' according to the status of the   */
/*         `_NumbersAreHex', `_ExpandSystemVariables' and `_ExpandSpectrumROM' variables, and `Address' is advanced to point to   */
/*         the next instruction.                                                                                                  */
/* Import: _FPSpecToDouble, _MatchLabel.                                                                                          */
/**********************************************************************************************************************************/

{
  struct _SpectrumFP_s    FPNumber;
  struct _Z80Assem_s     *SearchTable = &_SingleCodes;
  struct _Z80Assem_s     *InstructionRecord;
  char                   *DisP;
  word                    RealAdd;
  word                    DisassAdd;                                                                               /* The M0 byte */
  bool                    TreatHLAsIX = FALSE;
  bool                    DoingDDCB   = FALSE;
  byte                    NumBytes;
  int           register  Cnt;
  byte                    OpCnt;
  byte                    Operand;
  byte                    OtherOperand;
  byte                    BValue;
  word                    WValue;
  short                   RelativeValue;
  char                    IndexName[3];                                                                    /* Either 'IX' or 'IY' */
  char                    Symbol[10];
  bool                    Done;

  DisP = Disassembled;
  RealAdd = (word)(Address->_Address + LoadedAddress);
  DisassAdd = Address->_Address;                                                     /* Start at the beginning of the instruction */
  if (!_CountBytesOnly)
  {
    if (_NumbersAreHex)
      sprintf (DisP, " %04X ", RealAdd);
    else
      sprintf (DisP, "%05d ", RealAdd);
    DisP += 6;
  }
  if (Address->_RSTValue != 0)
  {
    NumBytes = 1;
    if (Address->_RSTValue <= 12)                                                    /* (Never set if `_ExpandSpectrumROM' FALSE) */
    {
      NumBytes = (*(_ObjectData + Address->_Address) >> 6) + 2;                                          /* Direct RST 28 numbers */
      if ((*(_ObjectData + Address->_Address) & 0x3F) == 0x00)
        NumBytes ++;
    }
  }
  else
  {
    if (*(_ObjectData + DisassAdd) == 0xCB)                                                                      /* CB Extended ? */
    {
      SearchTable = &_CBCodes;                                                                            /* Set the search table */
      DisassAdd ++;                                                                                          /* Need new M0 cycle */
    }
    else if (*(_ObjectData + DisassAdd) == 0xED)                                                                 /* ED Extended ? */
    {
      SearchTable = &_EDCodes;
      DisassAdd ++;
    }
    else if (*(_ObjectData + DisassAdd) == 0xDD || *(_ObjectData + DisassAdd) == 0xFD)                        /* DD/FD Extended ? */
    {
      TreatHLAsIX = TRUE;
      DisassAdd ++;
      strcpy (IndexName, *(_ObjectData + Address->_Address) == 0xDD ? "IX" : "IY");
      if (*(_ObjectData + DisassAdd) == 0xCB)
      {
        SearchTable = &_CBCodes;
        DisassAdd += 2;                                                                /* Always 4 bytes; instruction in last one */
        DoingDDCB = TRUE;
      }
      else if (*(_ObjectData + DisassAdd) == 0xED || *(_ObjectData + DisassAdd) == 0xDD || *(_ObjectData + DisassAdd) == 0xFD)
      {                                                                                             /* DDED/DDDD/DDFD don't exist */
        if (!_CountBytesOnly)
          sprintf (DisP, "%02X", *(_ObjectData + Address->_Address));                               /* Print and ignore the DD/FD */
        Address->_Address ++;
        return;
      }
    }
    InstructionRecord = SearchTable + *(_ObjectData + DisassAdd);                                 /* Fetch the instruction record */
    NumBytes = InstructionRecord->_NumBytes;
    if (DoingDDCB)                                                                                                      /* DDCB ? */
      NumBytes += 2;
    else if (TreatHLAsIX)
    {
      NumBytes ++;
      if (*(_ObjectData + DisassAdd) != 0xE9)                                          /* Not the (wrong) instruction "JP (HL)" ? */
        if (InstructionRecord->_Operand1 == (OPER_REGISTER | OPER_BRACKETS | 0x0C))                        /* Operand is "(HL)" ? */
          NumBytes ++;
        else if (InstructionRecord->_Operand2 == (OPER_REGISTER | OPER_BRACKETS | 0x0C))
          NumBytes ++;
    }
  }
  if (_CountBytesOnly)
  {
    if (Address->_RSTValue != 0)                                                     /* (Never set if `_ExpandSpectrumROM' FALSE) */
      switch (Address->_RSTValue)
      {
        case 0xCF : Address->_RSTValue = 0; break;                                                /* RST 08 has only 1 byte extra */
        case 0xEF : BValue = *(_ObjectData + Address->_Address);                                     /* (See below for full info) */
                    if (BValue == 0x00 || BValue == 0x33 || BValue == 0x35)                         /* Literals are always 1 byte */
                      Address->_RSTValue = 0x80;
                    else if (BValue == 0x34)
                      Address->_RSTValue = 1;
                    else if (BValue == 0x86 || BValue == 0x88 || BValue == 0x8C)
                      Address->_RSTValue = BValue - 0x80;
                    else if (BValue == 0x38)
                      Address->_RSTValue = 0;
                    break;
        case 0x80 : Address->_RSTValue = 0xEF; break;                               /* RST 28 jump; 1 byte; next is literal again */
        case 0x0C :
        case 0x0B :
        case 0x0A :
        case 0x09 :
        case 0x08 :
        case 0x07 :
        case 0x06 :
        case 0x05 :
        case 0x04 :
        case 0x03 :
        case 0x02 :
        case 0x01 : if (-- Address->_RSTValue == 0)                                        /* Direct RST 28 numbers; 1 less to go */
                      Address->_RSTValue = 0xEF;
                    break;
      }
    else if (_ExpandSpectrumROM && InstructionRecord->_OpCode == 0x3A)                                     /* A RST instruction ? */
    {
      BValue = (InstructionRecord->_Operand1 & 0x0F) * 8;                                                        /* Fetch address */
      if (BValue == 0x08 || BValue == 0x28)                                                        /* One that uses extra bytes ? */
        Address->_RSTValue = BValue + 199;                                                           /* Set a reminder (= opcode) */
    }
    Address->_Address += NumBytes;                                                                              /* Advance the PC */
    return;
  }
  for (Cnt = 0 ; Cnt < NumBytes ; Cnt ++)                                                    /* Print the instruction code in hex */
  {
    sprintf (DisP, "%02X", *(_ObjectData + Address->_Address + Cnt));
    DisP += 2;
  }
  for (Cnt = (5 - NumBytes) * 2 + 1 ; Cnt > 0 ; Cnt --)                                   /* Add spaces to go to the OpCode field */
    *(DisP ++) = ' ';
  if (_IsSpectrumROM || _HaveLabels)                                                             /* Are we using a symbol table ? */
  {
    strcpy (DisP, "          ");
    if (_MatchLabel (RealAdd, TRUE, Symbol))
      strncpy (DisP, Symbol, strlen (Symbol));
    DisP += 10;
  }
  if (Address->_RSTValue != 0)                                                       /* (Never set if `_ExpandSpectrumROM' FALSE) */
  {
    if (Address->_RSTValue == 0xCF)                                                    /* Previous Z80 instruction was a RST 08 ? */
    {
      strcpy (DisP, "DEFB ");
      DisP += 5;
      BValue = *(_ObjectData + Address->_Address);                                                        /* Fetch the error code */
      if (_NumbersAreHex)
        sprintf (DisP, "+%02X", BValue);
      else
        sprintf (DisP, "%-3d", BValue);
      DisP += 3;
      if ((++ BValue) < _NumRST08Codes)                                                           /* Spectrum ROM defined error ? */
      {                                                        /* (Notice that the stored code is one less than the error number) */
        *(DisP ++) = ' ';
        *(DisP ++) = ';';                                                       /* Add the error message description as 'comment' */
        if (BValue < 10)
          *(DisP ++) = BValue + '0';
        else
          *(DisP ++) = BValue - 10 + 'A';
        *(DisP ++) = ' ';
        strcpy (DisP, _RST08Codes[BValue]);
        DisP += strlen (DisP);
      }
      Address->_RSTValue = 0;                                                                                  /* Only 1 databyte */
    }
    else if (Address->_RSTValue == 0xEF)                                               /* Previous Z80 instruction was a RST 08 ? */
    {
      strcpy (DisP, "DEFB ");
      DisP += 5;
      BValue = *(_ObjectData + Address->_Address);                                                  /* Fetch the fp stack literal */
      if (_NumbersAreHex)
        sprintf (DisP, "+%02X", BValue);
      else
        sprintf (DisP, "%-3d", BValue);
      DisP += 3;
      if (_RST28Codes[BValue] != NULL)
      {
        *(DisP ++) = ' ';
        *(DisP ++) = ';';                                                               /* Add the FP stack literals as 'comment' */
        strcpy (DisP, _RST28Codes[BValue]);
        DisP += strlen (DisP);
      }
      if (BValue == 0x00 || BValue == 0x33 || BValue == 0x35)                             /* "jump_true", "jump" or "dec_jr_nz" ? */
        Address->_RSTValue = 0x80;                                                /* Set marker: next literal holds relative jump */
      else if (BValue == 0x34)                                                                                    /* "stk_data" ? */
        Address->_RSTValue = 1;                                                            /* Signal: 1 direct fp value following */
      else if (BValue == 0x86 || BValue == 0x88 || BValue == 0x8C)                   /* "series_06", "series_08" or "series_0C" ? */
        Address->_RSTValue = BValue - 0x80;                              /* Set number of following direct fp values (6, 8 or 12) */
      else if (BValue == 0x38)                                                                                    /* "end_calc" ? */
        Address->_RSTValue = 0;                                           /* Signal: next bytes are normal Z80 instructions again */
    }
    else if (Address->_RSTValue == 0x80)                                                                 /* Doing a RST 28 jump ? */
    {
      strcpy (DisP, "DEFB ");
      DisP += 5;
      BValue = *(_ObjectData + Address->_Address);                                                  /* Fetch the fp stack literal */
      if (_NumbersAreHex)
        sprintf (DisP, "+%02X", BValue);
      else
        sprintf (DisP, "%-3d", BValue);
      DisP += 3;
      *(DisP ++) = ' ';
      *(DisP ++) = ';';
      strcpy (DisP, "TO ");
      DisP += 3;
      RelativeValue = (char)*(_ObjectData + Address->_Address);                                      /* Fetch relative jump value */
      WValue = (word)(RealAdd + RelativeValue);
      if (_NumbersAreHex)
        sprintf (DisP, "%04X", WValue);
      else
        sprintf (DisP, "%d", WValue);
      DisP += strlen (DisP);
      if (_MatchLabel (WValue, FALSE, Symbol))
      {
        *(DisP ++) = ',';
        strcpy (DisP, Symbol);
        DisP += strlen (DisP);
      }
      Address->_RSTValue = 0xEF;                                               /* Set: next byte is a normal RST 28 literal again */
    }
    else if (Address->_RSTValue <= 12)                             /* Direct numbers following ("stk_data" or series generator) ? */
    {
      memset (&FPNumber, 0, sizeof (struct _SpectrumFP_s));                                                   /* Clear the number */
      strcpy (DisP, "DEFB ");
      DisP += 5;
      BValue = (*(_ObjectData + Address->_Address) >> 6) + 1;                                          /* Find length of mantissa */
      if (_NumbersAreHex)
        sprintf (DisP, "+%02X", *(_ObjectData + Address->_Address));
      else
        sprintf (DisP, "%-3d", *(_ObjectData + Address->_Address));
      DisP += 3;
      *(DisP ++) = ',';
      WValue = 1;
      if (*(_ObjectData + Address->_Address) & 0x3F)                                        /* Exponent is in this byte as well ? */
        FPNumber._Exponent = (*(_ObjectData + Address->_Address) & 0x3F) + 0x50;
      else
      {
        FPNumber._Exponent = *(_ObjectData + Address->_Address + 1) + 0x50;                      /* Nope, use next byte unreduced */
        if (_NumbersAreHex)
          sprintf (DisP, "+%02X", *(_ObjectData + Address->_Address + 1));
        else
          sprintf (DisP, "%-3d", *(_ObjectData + Address->_Address + 1));
        DisP += 3;
        *(DisP ++) = ',';
        WValue ++;
      }
      for (Cnt = 1 ; Cnt <= BValue ; Cnt ++, WValue ++)
      {
        FPNumber._Mantissa[Cnt - 1] = *(_ObjectData + Address->_Address + WValue);
        if (_NumbersAreHex)
          sprintf (DisP, "+%02X", FPNumber._Mantissa[Cnt - 1]);
        else
          sprintf (DisP, "%-3d", FPNumber._Mantissa[Cnt - 1]);
        DisP += 3;
        if (Cnt < BValue)
          *(DisP ++) = ',';
      }
      sprintf (DisP, " ;[%lg]", _FPSpecToDouble (&FPNumber));
      DisP += strlen (DisP);
      if (-- Address->_RSTValue == 0)                                                               /* Just did the last number ? */
        Address->_RSTValue = 0xEF;                                             /* Set: next byte is a normal RST 28 literal again */
    }
    else
      Address->_RSTValue = 0;
  }
  else
  {
    sprintf (DisP, "%-5s", _OpCodes[InstructionRecord->_OpCode]);                                 /* Print the OpCode and a space */
    DisP += 5;
    for (OpCnt = 1 ; OpCnt <= InstructionRecord->_NumOperands ; OpCnt ++)                                      /* Do all operands */
    {
      Operand = (OpCnt == 1 ? InstructionRecord->_Operand1 : InstructionRecord->_Operand2);
      OtherOperand = (OpCnt == 1 ? InstructionRecord->_Operand2 : InstructionRecord->_Operand1);
      if (Operand & OPER_BRACKETS)                                                                      /* Operands in brackets ? */
        *(DisP ++) = '(';                                                                            /* Print the opening bracket */
      switch (Operand & 0xE0)                                                                         /* Handle each operand type */
      {
        case OPER_DIRECT   : BValue = (Operand & 0x0F) * 8;                                                /* A RST instruction ! */
                             if (_NumbersAreHex)
                               sprintf (DisP, "00%02X", BValue);
                             else
                               sprintf (DisP, "%d", BValue);
                             DisP += strlen (DisP);
                             if (_ExpandSpectrumROM)
                             {
                               if (BValue == 0x08 || BValue == 0x28)                               /* One that uses extra bytes ? */
                                 Address->_RSTValue = BValue + 199;                                  /* Set a reminder (= opcode) */
                               WValue = BValue;
                               for (Cnt = 0, Done = FALSE ; !Done && Cnt < _NumSpecLabels ; Cnt ++)
                                 if (WValue == _SpecLabels[Cnt]->_Address)                      /* Match Spectrum ROM exclusively */
                                 {
                                   *(DisP ++) = ',';
                                   strcpy (DisP, _SpecLabels[Cnt]->_Name);
                                   DisP += strlen (DisP);
                                   Done = TRUE;
                                 }
                             }
                             break;
        case OPER_REGISTER : if (TreatHLAsIX)
                             {
                               if ((Operand & 0x0F) == 0x0C || DoingDDCB)               /* HL registerpair or autocopy function ? */
                                 if (((Operand & OPER_BRACKETS) &&                                 /* "(HL)" converts to "(IX+d)" */
                                       *(_ObjectData + DisassAdd) != 0xE9) ||         /* Unless the (wrong) instruction "JP (HL)" */
                                      DoingDDCB)
                                 {
                                   Done = FALSE;
                                   if (_ExpandSystemVariables && *(_ObjectData + Address->_Address) == 0xFD)  /* Doing "(IY+d)" ? */
                                   {
                                     WValue = (word)23610 + (char)*(_ObjectData + Address->_Address + 2);    /* (Always 3rd byte) */
                                     if (WValue >= SYSVARS_START && WValue < SYSVARS_AFTER)
                                     {
                                       for (Cnt = 0, Done = FALSE ; !Done && Cnt < _NumSystemVariables ; Cnt ++)
                                         if (WValue == _SystemVariables[Cnt]._Address)
                                         {
                                           strcpy (DisP, _SystemVariables[Cnt]._Name);
                                           Done = TRUE;
                                         }
                                         else if (WValue < _SystemVariables[Cnt]._Address)                 /* Past the variable ? */
                                         {
                                           sprintf (DisP, "%s+%d", _SystemVariables[Cnt - 1]._Name,
                                                    WValue - _SystemVariables[Cnt - 1]._Address);
                                           Done = TRUE;
                                         }
                                     }
                                   }
                                   if (!Done)
                                     if (_NumbersAreHex)
                                       sprintf (DisP, "%s+%02X", IndexName, *(_ObjectData + Address->_Address + 2));
                                     else
                                       if (*(_ObjectData + Address->_Address + 2) & 0x80)                    /* Negative number ? */
                                         sprintf (DisP, "%s%d", IndexName, (char)*(_ObjectData + Address->_Address + 2));
                                       else
                                         sprintf (DisP, "%s+%d", IndexName, *(_ObjectData + Address->_Address + 2));
                                 }
                                 else if (*(_ObjectData + DisassAdd) == 0xE9)     /* EXCEPTION: The (wrong) instruction "JP (HL)" */
                                   strcpy (DisP, IndexName);
                                 else if (*(_ObjectData + Address->_Address) == 0xEB)     /* EXCEPTION: "EX HL,DE" doesn't expand */
                                   strcpy (DisP, _Registers[Operand & 0x0F]);
                                 else
                                   strcpy (DisP, IndexName);
                               else if ((Operand & 0x0F) == 0x05)                                                 /* H register ? */
                                 if (OtherOperand == OPER_REGISTER | OPER_BRACKETS | 0x0C)
                                   strcpy (DisP, _Registers[Operand & 0x0F]);
                                 else
                                   sprintf (DisP, "%sh", IndexName);                                         /* Converts to "IXh" */
                               else if ((Operand & 0x0F) == 0x06)                                                 /* L register ? */
                                 if (OtherOperand == OPER_REGISTER | OPER_BRACKETS | 0x0C)
                                   strcpy (DisP, _Registers[Operand & 0x0F]);
                                 else
                                   sprintf (DisP, "%sl", IndexName);                                         /* Converts to "IXl" */
                               else
                                 strcpy (DisP, _Registers[Operand & 0x0F]);
                               if (DoingDDCB)                                                     /* Possible autocopy function ? */
                                 if ((Operand & 0x0F) != 0x0C &&                       /* Not the "(HL)" register (no autocopy) ? */
                                     InstructionRecord->_OpCode != 0x03)                  /* Not the "BIT" opcode (no autocopy) ? */
                                 {
                                   DisP += strlen (DisP);                               /* Add autocopy register as extra operand */
                                   *(DisP ++) = ',';                                                   /* (E.g. "RES 0,(IX+1),B") */
                                   strcpy (DisP, _Registers[Operand & 0x0F]);
                                 }
                             }
                             else
                               strcpy (DisP, _Registers[Operand & 0x0F]);
                             DisP += strlen (DisP);
                             if (Operand == OtherOperand && Operand == (OPER_REGISTER | 0x09) &&        /* EXCEPTION: "EX AF,AF'" */
                                 OpCnt == 2)                                                            /* Doing second operand ? */
                               *(DisP ++) = '\'';                                                                /* Add the quote */
                             break;
        case OPER_FLAG     : strcpy (DisP, _Flags[Operand & 0x0F]);
                             DisP += strlen (_Flags[Operand & 0x0F]);
                             break;
        case OPER_8BIT     : BValue = *(_ObjectData + Address->_Address + NumBytes - 1);
                             if (_NumbersAreHex)
                               sprintf (DisP, "+%02X", BValue);
                             else
                               sprintf (DisP, "%d", BValue);
                             DisP += strlen (DisP);
                             break;
        case OPER_16BIT    : WValue = *((word *)(_ObjectData + Address->_Address + NumBytes - 2));
                             if (_ExpandSystemVariables && WValue >= SYSVARS_START && WValue < SYSVARS_AFTER)
                             {
                               for (Cnt = 0, Done = FALSE ; !Done && Cnt < _NumSystemVariables ; Cnt ++)
                                 if (WValue == _SystemVariables[Cnt]._Address)
                                 {
                                   strcpy (DisP, _SystemVariables[Cnt]._Name);
                                   Done = TRUE;
                                 }
                                 else if (WValue < _SystemVariables[Cnt]._Address)                         /* Past the variable ? */
                                 {
                                   sprintf (DisP, "%s+%d", _SystemVariables[Cnt - 1]._Name,
                                            WValue - _SystemVariables[Cnt - 1]._Address);
                                   Done = TRUE;
                                 }
                             }
                             else if (_NumbersAreHex)
                               if (Operand & OPER_BRACKETS)
                                 sprintf (DisP, "%04X", WValue);                             /* Addresses don't get a leading '+' */
                               else
                                 sprintf (DisP, "+%04X", WValue);                                             /* While numbers do */
                             else
                               sprintf (DisP, "%d", WValue);
                             DisP += strlen (DisP);
                             break;
        case OPER_ADDRESS  : WValue = *((word *)(_ObjectData + Address->_Address + NumBytes - 2));
                             if (_NumbersAreHex)
                               sprintf (DisP, "%04X", WValue);
                             else
                               sprintf (DisP, "%d", WValue);
                             DisP += strlen (DisP);
                             if (_MatchLabel (WValue, FALSE, Symbol))
                             {
                               *(DisP ++) = ',';
                               strcpy (DisP, Symbol);
                               DisP += strlen (DisP);
                             }
                             break;
        case OPER_RELJUMP  : RelativeValue = (char)*(_ObjectData + Address->_Address + NumBytes - 1);
                             WValue = (word)(RealAdd + NumBytes + RelativeValue);
                             if (_NumbersAreHex)
                               sprintf (DisP, "%04X", WValue);
                             else
                               sprintf (DisP, "%d", WValue);
                             DisP += strlen (DisP);
                             if (_MatchLabel (WValue, FALSE, Symbol))
                             {
                               *(DisP ++) = ',';
                               strcpy (DisP, Symbol);
                               DisP += strlen (DisP);
                             }
                             break;
        case OPER_BITVALUE : *(DisP ++) = (Operand & 0x07) + '0';
                             break;
      }
      if (Operand & OPER_BRACKETS)                                                                      /* Operands in brackets ? */
        *(DisP ++) = ')';                                                                            /* Print the closing bracket */
      if (OpCnt == 1 && InstructionRecord->_NumOperands == 2)                                     /* Instruction has 2 operands ? */
        *(DisP ++) = ',';                                                                          /* Print the operand seperator */
    }
  }
  *DisP = '\0';                                                                                             /* Force an endmarker */
  for (Cnt = strlen (Disassembled) ; Cnt < _DisassemblerWidth ; Cnt ++)
    *(DisP ++) = ' ';                                                                                        /* Clear end of line */
  *DisP = '\0';                                                                                             /* Force an endmarker */
  Address->_Address += NumBytes;                                                                                /* Advance the PC */
}

static void _ScrollBack (struct _InstructionMemory_s *FirstAddress, struct _InstructionMemory_s *StartAddress, bool FullPage)

/**********************************************************************************************************************************/
/* Pre   : `FirstAddress' holds the address to be reached, `FullPage' is TRUE for page scrolling, FALSE for line scrolling.       */
/* Post  : Returned is the address of the instruction one line or page (depending on `FullPage') back in memory.                  */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct _InstructionMemory_s CurrentAddress;
  word                        BytesBack;
  word                        BuffersUsed = 0;
  char                        EmptyString = '\0';

  if (FirstAddress->_Address == 0)                                                              /* Already at the first address ? */
  {
    memset (&StartAddress, 0, sizeof (struct _InstructionMemory_s));
    return;
  }
  BytesBack = (FullPage ? _CurrentHeight * 4 : _CurrentHeight);
  if (FirstAddress->_Address > BytesBack)
    CurrentAddress._Address = FirstAddress->_Address - BytesBack;
  else
    CurrentAddress._Address = 0;
  CurrentAddress._RSTValue = 0;
  _CountBytesOnly = TRUE;
  while (CurrentAddress._Address < FirstAddress->_Address)
  {
    memcpy (&_BackScrollArray[(BuffersUsed ++) % _CurrentHeight], &CurrentAddress, sizeof (struct _InstructionMemory_s));
    _GetDisassemblyLine (&CurrentAddress, EmptyString);
  }
  _CountBytesOnly = FALSE;
  if (!FullPage)                                                                                   /* Only one instruction back ? */
    memcpy (StartAddress, &_BackScrollArray[(BuffersUsed - 1) % _CurrentHeight], sizeof (struct _InstructionMemory_s));
  else if (BuffersUsed >= _CurrentHeight)                                            /* Page back - At least a full buffer done ? */
    memcpy (StartAddress, &_BackScrollArray[BuffersUsed % _CurrentHeight], sizeof (struct _InstructionMemory_s));
  else
    memcpy (StartAddress, &_BackScrollArray[0], sizeof (struct _InstructionMemory_s));
}

static bool _GetAddress (word *Address, bool IsLoadAddress)

/**********************************************************************************************************************************/
/* Pre   : `Address' holds the default value, `IsLoadAddress' is TRUE when asking for the address where the code is to be loaded, */
/*         FALSE when asking for a new start address.                                                                             */
/* Post  : If the user entered a valid address (either hex by starting with a hash character or dec), TRUE is returned and that   */
/*         new address is put into `Address'.                                                                                     */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  char  SAddress[6];
  bool  Accepted = FALSE;
  word  NewAddress;
  short Cnt;

  NewAddress = *Address + (IsLoadAddress ? 0 : LoadedAddress);
  if (_NumbersAreHex && !IsLoadAddress)
    sprintf (SAddress, "$%X", NewAddress);
  else
    sprintf (SAddress, "%d", NewAddress);
  while (!Accepted)
  {
    if (!InputRequester (-1, -1, NULL, NULL, SAddress, 6, TRUE, FALSE,
        IsLoadAddress ? "Get Block Load Address" : "Go To Address",
        IsLoadAddress ? "{CGive the address where the block is loaded}C\n{Cin Spectrum memory}C" : "Enter the new start address"))
      return (FALSE);
    Accepted = TRUE;
    NewAddress = 0;
    Cnt = 0;
    if (SAddress[0] == '$')
    {
      Cnt ++;
      while (SAddress[Cnt] && Accepted)
        if (!isxdigit (SAddress[Cnt]))
          Accepted = FALSE;
        else
        {
          if (toupper (SAddress[Cnt]) > '9')
            NewAddress = NewAddress * 16 + toupper (SAddress[Cnt]) - 'A' + 10;
          else
            NewAddress = NewAddress * 16 + SAddress[Cnt] - '0';
          Cnt ++;
        }
    }
    else
    {
      while (SAddress[Cnt] && Accepted)
        if (!isdigit (SAddress[Cnt]))
          Accepted = FALSE;
        else
        {
          NewAddress = NewAddress * 10 + SAddress[Cnt] - '0';
          Cnt ++;
        }
    }
    if (!Accepted)
      ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - \'%s\' is not a valid address", SAddress);
  }
  *Address = NewAddress;
  return (TRUE);
}

static void _FlushExtLabels (void)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : Any loaded external symbol table file (in `_Labels' array) has been flushed from memory.                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (_Labels != NULL)
  {
    while (-- _NumLabels >= 0)
      free (_Labels[_NumLabels]);
    free (_Labels);
  }
  _Labels = NULL;
  _NumLabels = 0;
  _HaveLabels = FALSE;
}

static void _ReadLabelFile (char *FileName, bool SpectrumROM)

/**********************************************************************************************************************************/
/* Pre   : `FileName' holds the filename of the symbol table file to load, `SpectrumROM' is TRUE if it is the 48K Spectrum ROM    */
/*         symbol table, which is kept in a separate array.                                                                       */
/* Post  : The file has been loaded into the `_Labels' array (or `_SpecLabels' array if `SpectrumROM' was TRUE). Bad lines have   */
/*         been dropped and the first bad line has been reported. The array is sorted in ascending address order.                 */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  static bool         ReportNoSpectFile = TRUE;                            /* Report missing Spectrum symbol table file only once */
  struct _Labels_s ***SymbolTable;
  struct _Labels_s   *SortP;
  int                *NumSymbols;
  FILE               *Fp;
  int                 CurrentUsed = 0;
  int                 LineNo      = 0;
  char                OneLine[81];
  int      register   Cnt;
  int      register   Cnt2;
  bool                Accepted;
  bool                ReportOnce = TRUE;

  if (SpectrumROM)
  {
    if (_SpecLabelsLoaded)
      return;
    _SpecLabelsLoaded = TRUE;
    SymbolTable = &_SpecLabels;
    NumSymbols = &_NumSpecLabels;
  }
  else
  {
    SymbolTable = &_Labels;
    NumSymbols = &_NumLabels;
  }
  if ((Fp = fopen (FileName, "r")) == NULL)
  {
    if (SpectrumROM && ReportNoSpectFile)                                             /* Missing Spectrum ROM symbol table file ? */
    {
      ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Cannot load 48K Spectrum ROM symbol table file \"%s\"\n"
                                                "        %s", FileName, StrError (errno));
      ReportNoSpectFile = FALSE;                                                                            /* (Report only once) */
    }
    return;
  }
  while (fgets (OneLine, 80, Fp) != NULL)
  {
    LineNo ++;
    OneLine[80] = '\0';
    Cnt = strlen (OneLine) - 1;
    while (OneLine[Cnt] == 0x0D || OneLine[Cnt] == 0x0A)
      OneLine[Cnt --] = '\0';
    if ((*NumSymbols) == CurrentUsed)
    {
      CurrentUsed += 1024;
      if (CurrentUsed == 1024)
      {
        if (((*SymbolTable) = (struct _Labels_s **)malloc (CurrentUsed * sizeof (struct _Labels_s))) == NULL)
        {
          ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Out of memory\n        Cannot allocate %d bytes",
                          1024 * sizeof (struct _Labels_s));
          return;
        }
      }
      else
        if (((*SymbolTable) = (struct _Labels_s **)realloc ((*SymbolTable), CurrentUsed * sizeof (struct _Labels_s))) == NULL)
        {
          ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Out of memory\n        Cannot allocate %d bytes",
                          1024 * sizeof (struct _Labels_s));
          return;
        }
    }
    if (((*SymbolTable)[*NumSymbols] = (struct _Labels_s *)malloc (sizeof (struct _Labels_s))) == NULL)
    {
      ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Out of memory\n        Cannot allocate %d bytes",
                      sizeof (struct _Labels_s));
      return;
    }
    Cnt = 0;
    Accepted = TRUE;
    (*SymbolTable)[*NumSymbols]->_Address = 0;
    if (OneLine[0] == '$')
    {
      Cnt ++;
      while (OneLine[Cnt] != '\0' && OneLine[Cnt] != ' ' && OneLine[Cnt] != '\t' && Accepted)
        if (!isxdigit (OneLine[Cnt]))
          Accepted = FALSE;
        else
        {
          if (toupper (OneLine[Cnt]) > '9')
            (*SymbolTable)[*NumSymbols]->_Address = (*SymbolTable)[*NumSymbols]->_Address * 16 + toupper (OneLine[Cnt]) - 'A' + 10;
          else
            (*SymbolTable)[*NumSymbols]->_Address = (*SymbolTable)[*NumSymbols]->_Address * 16 + OneLine[Cnt] - '0';
          Cnt ++;
        }
    }
    else
    {
      while (OneLine[Cnt] != '\0' && OneLine[Cnt] != ' ' && OneLine[Cnt] != '\t' && Accepted)
        if (!isdigit (OneLine[Cnt]))
          Accepted = FALSE;
        else
        {
          (*SymbolTable)[*NumSymbols]->_Address = (*SymbolTable)[*NumSymbols]->_Address * 10 + OneLine[Cnt] - '0';
          Cnt ++;
        }
    }
    while (OneLine[Cnt] != '\0' && (OneLine[Cnt] == ' ' || OneLine[Cnt] == '\t'))
      Cnt ++;
    if (OneLine[Cnt] == '\0')
      Accepted = FALSE;
    if (!Accepted && ReportOnce)
    {
      ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Bad symbol at line %d in \"%s\"\n        \"%s\"\n"
                                                "        (Only the first bad line is reported)", LineNo, FileName, OneLine);
      ReportOnce = FALSE;
    }
    else
    {
      if (strlen (OneLine + Cnt) < 10)
        strcpy ((*SymbolTable)[*NumSymbols]->_Name, OneLine + Cnt);
      else
      {
        strncpy ((*SymbolTable)[*NumSymbols]->_Name, OneLine + Cnt, 9);
        (*SymbolTable)[*NumSymbols]->_Name[9] = '\0';
      }
      (*NumSymbols) ++;
    }
  }
  fclose (Fp);
  Accepted = TRUE;
  for (Cnt = (*NumSymbols) - 1 ; Cnt > 0 && Accepted ; Cnt --)                   /* Now sort the symbols (ascending) - bubblesort */
  {
    Accepted = FALSE;
    for (Cnt2 = 0 ; Cnt2 < Cnt ; Cnt2 ++)
    {
      if ((*SymbolTable)[Cnt2]->_Address > (*SymbolTable)[Cnt2 + 1]->_Address)                             /* Should be swapped ? */
      {
        SortP = (*SymbolTable)[Cnt2];
        (*SymbolTable)[Cnt2] = (*SymbolTable)[Cnt2 + 1];
        (*SymbolTable)[Cnt2 + 1] = SortP;
        Accepted = TRUE;
      }
    }
  }
  if (SpectrumROM)
    _HaveSpecLabels = _NumSpecLabels > 0 ? TRUE : FALSE;
  else
  {
    _HaveLabels = _NumLabels > 0 ? TRUE : FALSE;
    _AllocLabels = CurrentUsed;
  }
}

static bool _LoadExtLabels (bool Merge)

/**********************************************************************************************************************************/
/* Pre   : `Merge' is TRUE if the currently loaded external symbol table should be kept as well.                                  */
/* Post  : The user has selected another symbol table file. This one is either merged or loaded instead.                          */
/*         FALSE is returned if the user cancelled the FileRequester.                                                             */
/* Import: _FlushExtLabels, _ReadLabelFile.                                                                                       */
/**********************************************************************************************************************************/

{
  if (!Merge)
    _FlushExtLabels ();
  if (_Path[0] == '\0')
    strcpy (_Path, CurPath);
  if (FileRequester (180, 50, 48, 29, 20, _Path, _File, _Mask, "Select Symbol Table File", TRUE, FALSE))
  {
    strcpy (_LabelFileName, _File);
    MakeFullPath (_LabelPath, _Path, _File);
    _ReadLabelFile (_LabelPath, FALSE);
    return (TRUE);
  }
  return (FALSE);
}

static bool _SaveExtLabels (bool SaveAs)

/**********************************************************************************************************************************/
/* Pre   : `SaveAs' is TRUE if the user selected Save As rather than Save.                                                        */
/* Post  : The symbol table has been saved, possibly to a new file.                                                               */
/*         FALSE is returned if the user cancelled the FileRequester or the symbols were not changed and the user didn't enforce. */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  int register  Cnt;
  FILE         *Fp;
  bool          DoSave = TRUE;

  if (!_LabelsChanged && !SaveAs)                                                                        /* Symbols not changed ? */
    if (QuestionRequester (-1, -1, FALSE, FALSE, NULL, "@Cancel", NULL, NULL, NULL, "Confirm Unneeded Save",
                           "The symbol table has not been changed!\nSave anyway?"))
    return (FALSE);
  if (_Path[0] == '\0')
    strcpy (_Path, CurPath);
  if (_LabelFileName[0] == '\0')                                                                                 /* No name yet ? */
    SaveAs = TRUE;
  if (SaveAs)
    if (FileRequester (180, 50, 48, 29, 20, _Path, _File, _Mask, "Select New Symbol Table File", FALSE, FALSE))
    {
      strcpy (_LabelFileName, _File);
      MakeFullPath (_LabelPath, _Path, _File);
    }
    else
      DoSave = FALSE;
  if (DoSave)
  {
    if ((Fp = fopen (_LabelPath, "wt")) == NULL)
    {
      ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Cannot create symbol table file \"%s\"\n"
                                                  "        %s", _LabelFileName, StrError (errno));
      return (FALSE);
    }
    for (Cnt = 0 ; Cnt < _NumLabels ; Cnt ++)
      if (_NumbersAreHex)
        fprintf (Fp, "$%04X\t%s\n", _Labels[Cnt]->_Address, _Labels[Cnt]->_Name);
      else
        fprintf (Fp, "%05d\t%s\n", _Labels[Cnt]->_Address, _Labels[Cnt]->_Name);
    fclose (Fp);
    _LabelsChanged = FALSE;
    return (TRUE);
  }
  return (FALSE);
}

static comboent _AddExtSymbol (word DefaultAddress)

/**********************************************************************************************************************************/
/* Pre   : `DefaultAddress' holds the address at the top of the currently disassembled page.                                      */
/* Post  : The user has added a symbol to the external symbol table.                                                              */
/*         The returned value is the location where the symbol is stored in the table (after sorting), or -1 for cancel.          */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct WindowInfo  CWindow = {190, 150, 149, 95, 149, 95, 149, 95, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL,
                                "Add Symbol"};
  struct ButtonInfo  BAddr   = {BT_INPUT, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 80, 20, 5, DBLACK, DWHITE, NULL,
                                0, NULL, NULL};
  struct ButtonInfo  BName   = {BT_INPUT, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 80, 40, 9, DBLACK, DWHITE, NULL,
                                1, NULL, NULL};
  struct ButtonInfo  BCancel = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 10, 70, 8, DBLACK, DWHITE, NULL,
                                2, "@Cancel", NULL};
  struct ButtonInfo  BOk     = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 88, 70, 8, DBLACK, DWHITE, NULL,
                                3, "@Ok", NULL};
  struct _Labels_s  *NewSymbol;
  struct ButtonInfo *ButtonP;
  char               SAddr[6];
  char               SName[10];
  word               AAddr;
  short    register  Cnt;
  bool               Finished       = FALSE;
  bool               Cancelled      = FALSE;
  bool               Accepted       = FALSE;
  bool               BadAddress     = FALSE;

  BAddr.Contents = SAddr;
  BName.Contents = SName;
  if (_NumbersAreHex)
    sprintf (SAddr, "$%X", DefaultAddress + LoadedAddress);
  else
    sprintf (SAddr, "%d", DefaultAddress + LoadedAddress);
  SName[0] = '\0';
  AddButton (&CWindow, &BOk, TRUE);
  AddButton (&CWindow, &BAddr, FALSE);
  AddButton (&CWindow, &BName, FALSE);
  AddButton (&CWindow, &BCancel, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  Wprintf (&CWindow, 2, 1, "{+1Address}N");
  Wprintf (&CWindow, 4, 1, "{+1Name}N");
  while (!Accepted)
  {
    Finished = FALSE;
    while (!Finished)
    {
      switch (HandleWindow (&CWindow, &ButtonP))
      {
        case WINH_CLOSE     : Finished = Cancelled = TRUE; break;
        case WINH_RNOHAND   : switch (ButtonP->ReactNum)
                              {
                                case 2: Finished = Cancelled = TRUE; break;                                           /* (Cancel) */
                                case 3: Finished = TRUE; break;                                                           /* (Ok) */
                              }
                              break;
      }
      MouseStatus ();
      if (GlobalKey == K_ESC)
        Finished = Cancelled = TRUE;
    }
    DestroyWindow (&CWindow);
    if (Cancelled)
      return (-1);
    BadAddress = FALSE;
    Accepted = TRUE;
    if (SName[0] == '\0')                                                                                  /* Empty symbol name ? */
    {
      ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Symbol name is a required field!");
      Accepted = FALSE;
    }
    else if (!strcmp (SAddr, "$"))                                                            /* No address, only the hash mark ? */
      BadAddress = TRUE;
    else
    {
      AAddr = 0;
      Cnt = 0;
      if (SAddr[0] == '$')
      {
        Cnt ++;
        while (SAddr[Cnt] && !BadAddress)
          if (!isxdigit (SAddr[Cnt]))
            BadAddress = TRUE;
          else
          {
            if (toupper (SAddr[Cnt]) > '9')
              AAddr = AAddr * 16 + toupper (SAddr[Cnt]) - 'A' + 10;
            else
              AAddr = AAddr * 16 + SAddr[Cnt] - '0';
            Cnt ++;
          }
      }
      else
      {
        while (SAddr[Cnt] && !BadAddress)
          if (!isdigit (SAddr[Cnt]))
            BadAddress = TRUE;
          else
          {
            AAddr = AAddr * 10 + SAddr[Cnt] - '0';
            Cnt ++;
          }
      }
    }
    if (BadAddress)
    {
      ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - \'%s\' is not a valid address", SAddr);
      Accepted = FALSE;
    }
  }
  if (_NumLabels == _AllocLabels)
  {
    _AllocLabels += 1024;
    if (_AllocLabels == 1024)
    {
      if ((_Labels = (struct _Labels_s **)malloc (_AllocLabels * sizeof (struct _Labels_s))) == NULL)
      {
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Out of memory\n        Cannot allocate %d bytes",
                        1024 * sizeof (struct _Labels_s));
        return (-1);
      }
    }
    else
      if ((_Labels = (struct _Labels_s **)realloc (_Labels, _AllocLabels * sizeof (struct _Labels_s))) == NULL)
      {
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Out of memory\n        Cannot allocate %d bytes",
                        1024 * sizeof (struct _Labels_s));
        return (-1);
      }
  }
  if ((NewSymbol = (struct _Labels_s *)malloc (sizeof (struct _Labels_s))) == NULL)
  {
    ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Out of memory\n        Cannot allocate %d bytes",
                    sizeof (struct _Labels_s));
    return (-1);
  }
  NewSymbol->_Address = AAddr;
  strcpy (NewSymbol->_Name, SName);
  Cnt = 0;
  Finished = FALSE;
  while (Cnt < _NumLabels && !Finished)
    if (_Labels[Cnt]->_Address >= AAddr)                                                               /* Must be inserted here ? */
      Finished = TRUE;
    else
      Cnt ++;
  if (Cnt < _NumLabels)                                                                                               /* Insert ? */
    memmove ((void *)&(_Labels[Cnt + 1]), (void *)&(_Labels[Cnt]), (_NumLabels - Cnt) * sizeof (void *));              /* Move up */
  _Labels[Cnt] = NewSymbol;                                                                         /* Add the newly typed symbol */
  _NumLabels ++;                                                                                            /* And account for it */
  if (!_HaveLabels)
    _HaveLabels = TRUE;
  return (Cnt);
}

static bool _DeleteExtSymbol (comboent WhichSymbol)

/**********************************************************************************************************************************/
/* Pre   : The symbol to be removed must be appointed in the _SymbolWindow.                                                       */
/* Post  : The user has asked confirmation to delete the appointed symbol.                                                        */
/*         TRUE is returned if the user accepted.                                                                                 */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct ComboEntry NEntry;

  if (_SymbolWindow.NumEntries < 0 || WhichSymbol > _SymbolWindow.NumEntries)                                   /* (Sanity check) */
    return (FALSE);
  GetComboEntry (&_SymbolWindow, &NEntry, WhichSymbol);
  if (QuestionRequester (-1, -1, FALSE, FALSE, NULL, "@Cancel", NULL, NULL, NULL, "Confirm Delete",
                         "Are you sure you want to delete the symbol\n\"%s\"?", NEntry.Text))
    return (FALSE);
  free (_Labels[WhichSymbol]);                                                                                      /* Destroy it */
  if (WhichSymbol < _NumLabels - 1)
    memmove ((void *)&(_Labels[WhichSymbol]), (void *)&(_Labels[WhichSymbol + 1]),
             (_NumLabels - WhichSymbol - 1) * sizeof (void *));                                                      /* Move down */
  if (-- _NumLabels == 0)                                                                               /* Removing last symbol ? */
  {
    free (_Labels);
    _HaveLabels = FALSE;
  }
  return (TRUE);
}

static bool _SetupSymbolWindow (void)

/**********************************************************************************************************************************/
/* Pre   : The SymbolWindow must have been created.                                                                               */
/* Post  : All 'external' symbols are inserted into a reinitialized SymbolWindow and the entries have been redrawn.               */
/*         The return value is FALSE if an out-of-memory error occured (which is not reported here).                              */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct ComboEntry NEntry;
  bool              AllOk = TRUE;
  int      register Cnt;

  NEntry.Selected = 0x00;
  ReInitComboWindow (&_SymbolWindow, 0);
  for (Cnt = 0 ; AllOk && Cnt < _NumLabels ; Cnt ++)
  {
    if (_NumbersAreHex)
      sprintf (NEntry.Text, "%04X  ", _Labels[Cnt]->_Address);
    else
      sprintf (NEntry.Text, "%05d ", _Labels[Cnt]->_Address);
    strcat (NEntry.Text, _Labels[Cnt]->_Name);
    NEntry.Identifier = (dword)_Labels[Cnt]->_Address;
    if (!InsertComboEntry (&_SymbolWindow, &NEntry, NULL, -1))
      AllOk = FALSE;
  }
  PrintComboEntries (&_SymbolWindow);
  return (AllOk);
}

static bool _EditSymbolTable (word CurrentAddress)

/**********************************************************************************************************************************/
/* Pre   : `CurrentAddress' is the address of the top-most instruction currently on screen.                                       */
/* Post  : The 'external' symbol table has been adjusted by the user. TRUE is returned if the symbol table was changed.           */
/* Import: _SetupSymbolWindow, _LoadExtLabels, _FlushExtLabels, _SaveExtLabels, _AddExtSymbol, _DeleteExtSymbol.                  */
/**********************************************************************************************************************************/

{
  struct WindowInfo  CWindow = {120, 85, 174, 211, 174, 211, 174, 211, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL,
                                "Edit Symbol Table"};
  struct ButtonInfo  BLoad   = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 115, 6, 8, DBLACK, DWHITE, NULL,
                                1, "@Load", NULL};
  struct ButtonInfo  BMerge  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 115, 26, 8, DBLACK, DWHITE, NULL,
                                2, "@Merge", NULL};
  struct ButtonInfo  BSave   = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 115, 46, 8, DBLACK, DWHITE, NULL,
                                3, "@Save", NULL};
  struct ButtonInfo  BSaveAs = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 115, 66, 8, DBLACK, DWHITE, NULL,
                                4, "Sa@ve As", NULL};
  struct ButtonInfo  BFlush  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 115, 86, 8, DBLACK, DWHITE, NULL,
                                5, "@Flush", NULL};
  struct ButtonInfo  BAdd    = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 115, 116, 8, DBLACK, DWHITE, NULL,
                                6, "@Add", NULL};
  struct ButtonInfo  BDel    = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 115, 136, 8, DBLACK, DWHITE, NULL,
                                7, "@Delete", NULL};
  struct ButtonInfo  BOk     = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 115, 193, 8, DBLACK, DWHITE, NULL,
                                8, "@Ok", NULL};
  struct ButtonInfo *ButtonP;
  comboent           SelectedLine;
  comboent           SelectedSymbol;
  bool               Finished       = FALSE;
  bool               SymbolsChanged = FALSE;
  bool               WasEmpty       = FALSE;
  bool               AdjustDelete   = TRUE;

  AddButton (&CWindow, &BOk, TRUE);
  AddButton (&CWindow, &BLoad, FALSE);
  AddButton (&CWindow, &BMerge, FALSE);
  AddButton (&CWindow, &BSave, FALSE);
  AddButton (&CWindow, &BSaveAs, FALSE);
  AddButton (&CWindow, &BFlush, FALSE);
  AddButton (&CWindow, &BAdd, FALSE);
  AddButton (&CWindow, &BDel, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  MakeComboWindow (&_SymbolWindow, &CWindow, COMT_SELECT, COMF_NOBUFFER | COMF_VSLIDER | COMF_KEYSELECT | COMF_SYNCSLIDER,
                   CWindow.StartX + 4, CWindow.StartY + 4, 0, 0, 16, 15, 20, NULL, FALSE,
                   0, NULL, DBLACK, DWHITE, 0, 0);
  SelectedSymbol = -1;
  _SetupSymbolWindow ();
  SelectEntry (&_SymbolWindow, SelectedSymbol);
  if (_SymbolWindow.NumEntries == -1)                                                                      /* No symbols at all ? */
    WasEmpty = TRUE;
  while (!Finished)
  {
    if (AdjustDelete)
    {
      if (_SymbolWindow.NumEntries == -1)                                                                  /* No symbols at all ? */
        BDel.Flags |= BF_GHOSTED;
      else
        BDel.Flags &= ~BF_GHOSTED;
      UpdateButtonText (&CWindow, &BDel, 0);
      AdjustDelete = FALSE;
    }
    switch (HandleWindow (&CWindow, &ButtonP))
    {
      case WINH_CLOSE     : Finished = TRUE; break;
      case WINH_RNOHAND   : switch (ButtonP->ReactNum)
                            {
                              case 1 : _LoadExtLabels (FALSE);                                                          /* (Load) */
                                       _SetupSymbolWindow ();
                                       _LabelsChanged = FALSE;
                                       SymbolsChanged = TRUE;
                                       SelectedSymbol = SelectedLine = -1;
                                       WasEmpty = FALSE;
                                       AdjustDelete = TRUE;
                                       break;
                              case 2 : _LoadExtLabels (TRUE);                                                          /* (Merge) */
                                       if (WasEmpty && _SetupSymbolWindow ())
                                         _LabelsChanged = TRUE;
                                       SymbolsChanged = TRUE;
                                       SelectedSymbol = SelectedLine = -1;
                                       WasEmpty = FALSE;
                                       AdjustDelete = TRUE;
                                       break;
                              case 3 : _SaveExtLabels (FALSE);                                                          /* (Save) */
                                       break;
                              case 4 : _SaveExtLabels (TRUE);                                                        /* (Save As) */
                                       break;
                              case 5 : _FlushExtLabels ();                                                             /* (Flush) */
                                       SelectedSymbol = SelectedLine = -1;
                                       _SetupSymbolWindow ();
                                       if (!WasEmpty)
                                         _LabelsChanged = SymbolsChanged = TRUE;
                                       WasEmpty = TRUE;
                                       AdjustDelete = TRUE;
                                       break;
                              case 6 : if ((SelectedLine = _AddExtSymbol (CurrentAddress)) >= 0)                         /* (Add) */
                                       {
                                         SelectedSymbol = SelectedLine;
                                         _SetupSymbolWindow ();
                                         SelectEntry (&_SymbolWindow, SelectedSymbol);                /* Appoint the added symbol */
                                         _LabelsChanged = SymbolsChanged = TRUE;
                                         AdjustDelete = TRUE;
                                       }
                                       break;
                              case 7 : if (SelectedSymbol >= 0)                                                       /* (Delete) */
                                         if (_DeleteExtSymbol (SelectedSymbol))
                                         {
                                           DeleteComboEntry (&_SymbolWindow, SelectedSymbol);
                                           if (SelectedSymbol > _SymbolWindow.NumEntries)
                                             _SymbolWindow.LastSelected = -- SelectedSymbol;
                                           PrintComboEntries (&_SymbolWindow);
                                           _LabelsChanged = SymbolsChanged = TRUE;
                                           AdjustDelete = TRUE;
                                         }
                                       break;
                              case 8 : Finished = TRUE; break;                                                            /* (Ok) */
                            }
                            break;
    }
    switch (HandleComboWindow (&_SymbolWindow, &SelectedLine))
    {
      case WINH_LINEFIRST :
      case WINH_LINESEL   : SelectedSymbol = SelectedLine;
                            break;
    }
    MouseStatus ();
    if (GlobalKey == K_ESC)
      Finished = TRUE;
  }
  DestroyComboWindow (&_SymbolWindow, 0);
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  return (SymbolsChanged);
}

static bool _GetSaveRange (word *StartAddress, word *EndAddress)

/**********************************************************************************************************************************/
/* Pre   : `StartAddress' and `EndAddress' hold the default values - the entire data block.                                       */
/* Post  : The user has selected the first and last address of the range.                                                         */
/*         FALSE is returned if the user cancelled the requester.                                                                 */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct WindowInfo  CWindow = {190, 150, 149, 95, 149, 95, 149, 95, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL,
                                "Save Disassembly"};
  struct ButtonInfo  BAddrS  = {BT_INPUT, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 100, 20, 5, DBLACK, DWHITE, NULL,
                                0, NULL, NULL};
  struct ButtonInfo  BAddrE  = {BT_INPUT, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 100, 40, 5, DBLACK, DWHITE, NULL,
                                1, NULL, NULL};
  struct ButtonInfo  BCancel = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 10, 70, 8, DBLACK, DWHITE, NULL,
                                2, "@Cancel", NULL};
  struct ButtonInfo  BOk     = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 88, 70, 8, DBLACK, DWHITE, NULL,
                                3, "@Ok", NULL};
  struct ButtonInfo *ButtonP;
  char               SAddr[2][6];
  word               AAddr[2];
  short    register  Cnt;
  short              CntAddress;
  bool               Finished       = FALSE;
  bool               Cancelled      = FALSE;
  bool               Accepted       = FALSE;
  bool               BadAddress     = FALSE;
  word               MinAddress;
  word               MaxAddress;

  BAddrS.Contents = SAddr[0];
  BAddrE.Contents = SAddr[1];
  MinAddress = *StartAddress;                                                                  /* Save the possible address range */
  MaxAddress = *EndAddress;
  if (_NumbersAreHex)
  {
    sprintf (SAddr[0], "$%X", *StartAddress);
    sprintf (SAddr[1], "$%X", *EndAddress);
  }
  else
  {
    sprintf (SAddr[0], "%d", *StartAddress);
    sprintf (SAddr[1], "%d", *EndAddress);
  }
  AddButton (&CWindow, &BOk, TRUE);
  AddButton (&CWindow, &BAddrS, FALSE);
  AddButton (&CWindow, &BAddrE, FALSE);
  AddButton (&CWindow, &BCancel, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  Wprintf (&CWindow, 2, 1, "{+1First address}N");
  Wprintf (&CWindow, 4, 1, "{+1Last address}N");
  while (!Accepted)
  {
    Finished = FALSE;
    while (!Finished)
    {
      switch (HandleWindow (&CWindow, &ButtonP))
      {
        case WINH_CLOSE     : Finished = Cancelled = TRUE; break;
        case WINH_RNOHAND   : switch (ButtonP->ReactNum)
                              {
                                case 2: Finished = Cancelled = TRUE; break;                                           /* (Cancel) */
                                case 3: Finished = TRUE; break;                                                           /* (Ok) */
                              }
                              break;
      }
      MouseStatus ();
      if (GlobalKey == K_ESC)
        Finished = Cancelled = TRUE;
    }
    if (Cancelled)
    {
      DestroyWindow (&CWindow);
      return (FALSE);
    }
    BadAddress = FALSE;
    Accepted = TRUE;
    for (CntAddress = 0 ; Accepted && CntAddress < 2 ; CntAddress ++)
    {
      if (!strcmp (SAddr[CntAddress], "$"))                                                   /* No address, only the hash mark ? */
        BadAddress = TRUE;
      else
      {
        AAddr[CntAddress] = 0;
        Cnt = 0;
        if (SAddr[CntAddress][0] == '$')
        {
          Cnt ++;
          while (SAddr[CntAddress][Cnt] && !BadAddress)
            if (!isxdigit (SAddr[CntAddress][Cnt]))
              BadAddress = TRUE;
            else
            {
              if (toupper (SAddr[CntAddress][Cnt]) > '9')
                AAddr[CntAddress] = AAddr[CntAddress] * 16 + toupper (SAddr[CntAddress][Cnt]) - 'A' + 10;
              else
                AAddr[CntAddress] = AAddr[CntAddress] * 16 + SAddr[CntAddress][Cnt] - '0';
              Cnt ++;
            }
        }
        else
        {
          while (SAddr[CntAddress][Cnt] && !BadAddress)
            if (!isdigit (SAddr[CntAddress][Cnt]))
              BadAddress = TRUE;
            else
            {
              AAddr[CntAddress] = AAddr[CntAddress] * 10 + SAddr[CntAddress][Cnt] - '0';
              Cnt ++;
            }
        }
      }
      if (BadAddress)
      {
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - \'%s\' is not a valid address", SAddr[CntAddress]);
        Accepted = FALSE;
      }
    }
    if (Accepted && AAddr[0] > AAddr[1])                                               /* Start address higher than end address ? */
    {
      ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - The start address must be lower than the end address!");
      Accepted = FALSE;
    }
    if (Accepted && (AAddr[0] < MinAddress || AAddr[0] > MaxAddress ||
                     AAddr[1] < MinAddress || AAddr[1] > MaxAddress))
    {
      if (_NumbersAreHex)
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - The save range must be within the address range [$%X, $%X]",
                        MinAddress, MaxAddress);
      else
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - The save range must be within the address range [%d, %d]",
                        MinAddress, MaxAddress);
      Accepted = FALSE;
    }
  }
  DestroyWindow (&CWindow);
  *StartAddress = AAddr[0] - LoadedAddress;
  *EndAddress = AAddr[1] - LoadedAddress;
  return (TRUE);
}

static dword _EditFind (bool ReInit, dword BlockLength)

/**********************************************************************************************************************************/
/* Pre   : `ReInit' is TRUE if a the static should be cleared (startup of this block), `BlockLength' is the length of the block.  */
/* Post  : The user has entered a search string, either ASCII or hex and the string has been searched for.                        */
/*         The return value is FINDNOMATCH is no match was found (reported already) or the index address where the match was made.*/
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
#define FINDMAXWIDTH 16
  static struct WindowInfo CWindow = {180, 150, 280, 95, 280, 95, 280, 95, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                      WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL,
                                      "Edit Find"};
  static struct ButtonInfo BStrS   = {BT_INPUT, BC_TEXT, BF_FWIDTH | BF_FITTXT | BF_KEYSELECT, 75, 20, FINDMAXWIDTH,
                                      DBLACK, DWHITE, NULL, 0, NULL, NULL};
  static struct ButtonInfo BStrH   = {BT_INPUT, BC_HEX, BF_FWIDTH | BF_FITTXT | BF_KEYSELECT, 75, 40, FINDMAXWIDTH * 2,
                                      DBLACK, DWHITE, NULL, 1, NULL, NULL};
  static struct ButtonInfo BCancel = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 35, 70, 8, DBLACK, DWHITE, NULL,
                                      2, "@Cancel", NULL};
  static struct ButtonInfo BAgain  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 113, 70, 8, DBLACK, DWHITE, NULL,
                                      3, "@Again", NULL};
  static struct ButtonInfo BFind   = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 191, 70, 8, DBLACK, DWHITE, NULL,
                                      4, "@Start", NULL};
  struct ButtonInfo       *ButtonP;
  static char              SStrS[FINDMAXWIDTH + 1]        = { '\0' };
  static char              SStrSBU[FINDMAXWIDTH + 1]      = { '\0' };
  static char              SStrH[FINDMAXWIDTH * 2 + 1]    = { '\0' };
  static char              SStrHBU[FINDMAXWIDTH * 2 + 1]  = { '\0' };
  static byte              FindValue[FINDMAXWIDTH]        = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
  static byte              FindValueBU[FINDMAXWIDTH]      = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
  static dword             FindResult                     = FINDNOMATCH;
  static dword             FindResultBU                   = FINDNOMATCH;
  short                    FindLength;
  short                    Cnt;
  bool                     Finished                       = FALSE;
  bool                     SearchAgain                    = TRUE;
  bool                     FromStart                      = FALSE;
  bool                     Cancelled                      = FALSE;

  BStrS.Contents = SStrS;
  BStrH.Contents = SStrH;
  if (ReInit)
  {
    FindResult = FindResultBU = FINDNOMATCH;
    SStrS[0] = SStrH[0] = SStrSBU[0] = SStrHBU[0] = '\0';
    for (Cnt = 0 ; Cnt < FINDMAXWIDTH ; Cnt ++)
      FindValue[Cnt] = FindValueBU[Cnt] = 0;
    FromStart = TRUE;
    BAgain.Flags |= BF_GHOSTED;
  }
  else
  {
    FindResult = FindResultBU;
    BAgain.Flags &= ~BF_GHOSTED;
  }
  AddButton (&CWindow, &BCancel, FALSE);
  AddButton (&CWindow, &BFind, TRUE);
  AddButton (&CWindow, &BAgain, FALSE);
  AddButton (&CWindow, &BStrS, FALSE);
  AddButton (&CWindow, &BStrH, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  Wprintf (&CWindow, 2, 1, "{+1ASCII}N");
  Wprintf (&CWindow, 4, 1, "{+1Hex string}N");
  while (!Finished)
  {
    switch (HandleWindow (&CWindow, &ButtonP))
    {
      case WINH_CLOSE     : Finished = Cancelled = TRUE; break;
      case WINH_EDITKEY   : switch (ButtonP->ReactNum)                                           /* An edit button accepted a key */
                            {
                              case 0: for (Cnt = 0 ; Cnt < strlen (SStrS) ; Cnt ++)                                    /* (ASCII) */
                                        FindValue[Cnt] = SStrS[Cnt];
                                      for (Cnt = 0 ; Cnt < strlen (SStrS) ; Cnt ++)
                                      {
                                        if ((SStrS[Cnt] & 0xF0) >= 0xA0)
                                          SStrH[Cnt * 2] = ((SStrS[Cnt] & 0xF0) >> 4) + 'A' - 10;
                                        else
                                          SStrH[Cnt * 2] = ((SStrS[Cnt] & 0xF0) >> 4) + '0';
                                        if ((SStrS[Cnt] & 0x0F) >= 0x0A)
                                          SStrH[Cnt * 2 + 1] = (SStrS[Cnt] & 0x0F) + 'A' - 10;
                                        else
                                          SStrH[Cnt * 2 + 1] = (SStrS[Cnt] & 0x0F) + '0';
                                      }
                                      SStrH[Cnt * 2] = '\0';
                                      UpdateButtonText (&CWindow, &BStrH, BF_FITTXT);
                                      break;
                              case 1: for (Cnt = 0 ; Cnt < strlen (SStrH) ; Cnt += 2)                             /* (Hex string) */
                                      {
                                        if (SStrH[Cnt] >= 'A')
                                          FindValue[Cnt / 2] = (SStrH[Cnt] - 'A' + 10) << 4;
                                        else
                                          FindValue[Cnt / 2] = (SStrH[Cnt] - '0') << 4;
                                        if (Cnt + 1 < strlen (SStrH))                                  /* Even number of digits ? */
                                          if (SStrH[Cnt + 1] >= 'A')
                                            FindValue[Cnt / 2] |= (SStrH[Cnt + 1] - 'A' + 10);
                                          else
                                            FindValue[Cnt / 2] |= (SStrH[Cnt + 1] - '0');
                                      }
                                      for (Cnt = 0 ; Cnt < (strlen (SStrH) / 2) ; Cnt ++)
                                      {
                                        if (FindValue[Cnt] >= 32 && FindValue[Cnt] < 128)
                                          SStrS[Cnt] = FindValue[Cnt];
                                        else
                                          SStrS[Cnt] = '.';
                                      }
                                      SStrS[Cnt] = '\0';
                                      UpdateButtonText (&CWindow, &BStrS, BF_FITTXT);
                                      break;
                            }
      case WINH_EDITEND   : strcpy (SStrSBU, SStrS);
                            strcpy (SStrHBU, SStrH);
                            memcpy (FindValueBU, FindValue, FINDMAXWIDTH);
                            break;
      case WINH_EDITESC   : strcpy (SStrS, SStrSBU);
                            strcpy (SStrH, SStrHBU);
                            memcpy (FindValue, FindValueBU, FINDMAXWIDTH);
                            UpdateButtonText (&CWindow, &BStrS, BF_FITTXT);
                            UpdateButtonText (&CWindow, &BStrH, BF_FITTXT);
                            break;
      case WINH_RNOHAND   : switch (ButtonP->ReactNum)
                            {
                              case 2: Finished = Cancelled = TRUE; break;                                             /* (Cancel) */
                              case 3: Finished = TRUE; break;                                                          /* (Again) */
                              case 4: Finished = TRUE; FromStart = TRUE; break;                                        /* (Start) */
                            }
                            break;
    }
    MouseStatus ();
    if (GlobalKey == K_ESC)
      Finished = Cancelled = TRUE;
  }
  if (Cancelled)
  {
    DestroyWindow (&CWindow);
    return (FINDNOMATCH);
  }
  FindLength = strlen (SStrS);
  SearchAgain = TRUE;
  while (SearchAgain)
  {
    SearchAgain = FALSE;
    if (FromStart)
      FindResult = 0x0000;
    else
      FindResult ++;
    if (FindResult + FindLength + 1 >= BlockLength)
      Finished = FALSE;                                                                                            /* End reached */
    else
    {
      for (Finished = FALSE ; !Finished && FindResult + FindLength < BlockLength ; FindResult ++)
        if (!memcmp (_ObjectData + FindResult, FindValue, FindLength))                                                 /* Got it! */
          Finished = TRUE;
      if (Finished)
        FindResultBU = -- FindResult;                                                                  /* Adjust for the for-loop */
    }
    if (!Finished)                                                                                 /* End reached without a match */
    {
      FindResult = FINDNOMATCH;
      if (QuestionRequester (-1, -1, FALSE, FALSE, "@Cancel", "@Again", NULL, NULL, NULL, "No match found",
                             "No match found. Search again from the start?"))
      {
        SearchAgain = TRUE;
        FromStart = TRUE;
      }
    }
  }
  DestroyWindow (&CWindow);
  return (FindResult);
}

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE EXPORTED DISASSEMBLER ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

bool GetLoadedAddress (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The address where the currently appointed block is loaded in Spectrum memory has been determined.                      */
/*         FALSE is returned if the user cancelled the requester.                                                                 */
/* Import: _GetAddress.                                                                                                           */
/**********************************************************************************************************************************/

{
  struct TapeEntry_s *PrevDirP;
  struct ComboEntry   ComEntry;
  dword               XMSIndex;
  byte                HeaderBuffer[20];

  LoadedAddress = 0x0000;                                              /* Determine default address where the code will be loaded */
  if (TapeWindow[WhichWindow].SelectedLine > 0)
  {
    GetComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &ComEntry, TapeWindow[WhichWindow].SelectedLine - 1);
    if (ComEntry.Identifier != 0)
    {
      PrevDirP = (struct TapeEntry_s *)ComEntry.Identifier;
      if (PrevDirP->Flags.IsHeader)                                                    /* Previous tape block is a header block ? */
      {
        XMSIndex = GetTapeBlockXMSIndex (WhichWindow, PrevDirP);
        CopyFromXMS (20, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, HeaderBuffer);                    /* Fetch that header */
        if (HeaderBuffer[0] == 0x00)                                                                           /* Program block ? */
          LoadedAddress = 23755;                                                            /* Default to normal PROGS area start */
        else if (HeaderBuffer[0] == 0x03)                                                                        /* Bytes block ? */
          LoadedAddress = (word)HeaderBuffer[13] + 256 * (word)HeaderBuffer[14];     /* Get default start address from the header */
      }
    }
  }
  return (_GetAddress (&LoadedAddress, TRUE));                   /* Ask the user where the code will be loaded in Spectrum memory */
}

static void _DrawEditCursor (bool State)

/**********************************************************************************************************************************/
/* Pre   : `State' determines whether the cursor should be written or removed.                                                    */
/* Post  : The edit cursor has been adjusted, but only if it was focussed.                                                        */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  static short      _RealStartX;
  static short      _RealStartY;
  static short      _RealEndX;
  static short      _RealColor;

  if (_EditCurX >= 0)                                                                              /* Flashing cursor is active ? */
    if (_EditCurY >= _EditHWindow.DFirstLine && _EditCurY < (_EditHWindow.DFirstLine + _EditHWindow.DHeight))
    {
      _RealColor = State ? DBLACK : DWHITE;
      if (_EditWhichWindow)                                                                        /* Doing the hex-part window ? */
      {
        _RealStartX = _EditCurX * DefaultFont->Width + 2;
        _RealStartY = (_EditCurY - _EditHWindow.DFirstLine + 1) * (DefaultFont->Height + 2);
        _RealEndX = _RealStartX + DefaultFont->Width - 1;
        WDrawLine (&(_EditHWindow.ComboWindow), _RealColor, _RealStartX, _RealStartY, _RealEndX, _RealStartY);
        WDrawLine (&(_EditHWindow.ComboWindow), _RealColor, _RealStartX, _RealStartY + 1, _RealEndX, _RealStartY + 1);
      }
      else
      {
        _RealStartX = _EditCurX * SpectrumFont->Width + 2;
        _RealStartY = (_EditCurY - _EditAWindow.DFirstLine + 1) * (SpectrumFont->Height + 2);
        _RealEndX = _RealStartX + SpectrumFont->Width - 1;
        WDrawLine (&(_EditAWindow.ComboWindow), _RealColor, _RealStartX, _RealStartY, _RealEndX, _RealStartY);
        WDrawLine (&(_EditAWindow.ComboWindow), _RealColor, _RealStartX, _RealStartY + 1, _RealEndX, _RealStartY + 1);
      }
    }
}

static void _FlashEditCursor (struct WindowInfo *Window)

/**********************************************************************************************************************************/
/* Pre   : None (This is a timed handler, set up in the hex editor).                                                              */
/* Post  : The editor cursor has been flashed if focussed.                                                                        */
/* Import: _DrawEditCursor.                                                                                                       */
/**********************************************************************************************************************************/

{
  static bool EditCursorState = FALSE;

  Window = Window; /* keep compiler happy */
  EditCursorState = !EditCursorState;
  _DrawEditCursor (EditCursorState);
}

bool ViewAsDisassembly (short WhichWindow, dword MaxBlockLength, word *StartAddress)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the tape block to be disassembled.                                                                   */
/* Post  : The currently appointed block has been shown as a block of disassembly.                                                */
/*         The return value is FALSE if `Ok' or window close was pressed, TRUE if the hex editor should be called from the top    */
/*         address.                                                                                                               */
/* Import: _GetAddress, _GetDisassemblyLine, _ScrollBack, _ReadLabelFile, _FlushExtLabels, _GetSaveRange, GetLoadedAddress.       */
/**********************************************************************************************************************************/

{
#define CWOFF        30
#define CWADD        4
#define OWADD        (CWADD + CWOFF + 38)

  struct WindowInfo OWindow  = {125, 90, 0, DEFVWLINES * 10 + OWADD, 0, MINVWLINES * 10 + OWADD, 0, MAXVWLINES * 10 + OWADD,
                                1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Tape Block Disassembly"};
  struct WindowInfo CWindow  = {130, 90 + CWOFF, 0, DEFVWLINES * 10 + CWADD, 0, DEFVWLINES * 10 + CWADD, 0, DEFVWLINES * 10 + CWADD,
                                1, WINB_OUT, DBLACK, DWHITE, NULL, WINF_NOBUFFER, NULL, NULL};
  struct ButtonInfo PRestart = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 50, 0, 8, DBLACK, DWHITE, NULL, 10, "@Go to", NULL};
  struct ButtonInfo PHexDump = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 125, 0, 8, DBLACK, DWHITE, NULL, 14, "@Hexedit", NULL};
  struct ButtonInfo PSymbols = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 200, 9, 8, DBLACK, DWHITE, NULL, 11, "@Symbols", NULL};
  struct ButtonInfo POk      = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 275, 0, 8, DBLACK, DWHITE, NULL, 12, "@Ok", NULL};
  struct ButtonInfo PNumMode = {BT_BOOL, BC_GADGET, BF_FWIDTH | BF_BCLEAR | BF_KEYSELECT, 5, 9, 1, DBLACK, DWHITE, NULL,
                                20, &_NumbersAreHex, NULL};
  struct ButtonInfo PSysMode = {BT_BOOL, BC_GADGET, BF_FWIDTH | BF_BCLEAR | BF_KEYSELECT, 107, 9, 1, DBLACK, DWHITE, NULL,
                                21, &_ExpandSystemVariables, NULL};
  struct ButtonInfo PROMMode = {BT_BOOL, BC_GADGET, BF_FWIDTH | BF_BCLEAR | BF_KEYSELECT, 209, 9, 1, DBLACK, DWHITE, NULL,
                                22, &_ExpandSpectrumROM, NULL};
  struct ButtonInfo PSave    = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 0, 9, 6, DBLACK, DWHITE, NULL, 13, "Sa@ve", NULL};
  struct ButtonInfo PPageUp  = {BT_CPUSH, BC_PIC, BF_FWIDTH | BF_KEYSELECT | BF_NOSURROUND, 0, 0, 1, DBLACK, DWHITE, NULL,
                                1, _GadUpPg, NULL};
  struct ButtonInfo PLineUp  = {BT_CPUSH, BC_PIC, BF_FWIDTH | BF_KEYSELECT | BF_NOSURROUND, 0, 0, 1, DBLACK, DWHITE, NULL,
                                2, _GadUpLn, NULL};
  struct ButtonInfo PByteUp  = {BT_CPUSH, BC_PIC, BF_FWIDTH | BF_KEYSELECT | BF_NOSURROUND, 0, 0, 1, DBLACK, DWHITE, NULL,
                                3, _GadUpBt, NULL};
  struct ButtonInfo PByteDn  = {BT_CPUSH, BC_PIC, BF_FWIDTH | BF_KEYSELECT | BF_NOSURROUND, 0, 0, 1, DBLACK, DWHITE, NULL,
                                4, _GadDnBt, NULL};
  struct ButtonInfo PLineDn  = {BT_CPUSH, BC_PIC, BF_FWIDTH | BF_KEYSELECT | BF_NOSURROUND, 0, 0, 1, DBLACK, DWHITE, NULL,
                                5, _GadDnLn, NULL};
  struct ButtonInfo PPageDn  = {BT_CPUSH, BC_PIC, BF_FWIDTH | BF_KEYSELECT | BF_NOSURROUND, 0, 0, 1, DBLACK, DWHITE, NULL,
                                6, _GadDnPg, NULL};
  struct ButtonInfo      *ButtonP;
  FILE                   *Fp;
  char                    DisassembledLine[81];
  char                    SpecLabelsFullPath[MAXDIR + 1];
  char                    SavePath[MAXDIR + 1];
  char                    SaveFilename[13];
  char                    SaveExt[5];
  static char             SaveMask[13] = "*.DIS";
  struct _InstructionMemory_s FirstAddress = { 0x0000, 0 };                                     /* Address at the top of the page */
  struct _InstructionMemory_s CurrentAddress;                                                                               /* PC */
  struct _InstructionMemory_s NextAddress;                                             /* Address of second instruction in screen */
  struct _InstructionMemory_s SaveFirst;
  struct _InstructionMemory_s SaveLast;
  bool                    Done         = FALSE;
  bool                    Restart      = TRUE;
  bool                    Resize       = TRUE;
  short                   LineCnt;
  short         register  Cnt;
  short                   PrevDisassemblerWidth = DISASSWIDTH;
  short                   UsePagers = -1;                                            /* Combine button and keyboard pager actions */
  word                    SavePreviousAddress;
  bool                    StartHexDump = FALSE;

  if (!_GadgetsConverted)
  {
    for (Cnt = 2 ; Cnt < 8 * 12 + 2 ; Cnt ++)                                                      /* Give the gadgets some color */
      _GadUpBt[Cnt] = (_GadUpBt[Cnt] ? DBLACK : DWHITE);
    for (Cnt = 2 ; Cnt < 8 * 12 + 2 ; Cnt ++)
      _GadUpLn[Cnt] = (_GadUpLn[Cnt] ? DBLACK : DWHITE);
    for (Cnt = 2 ; Cnt < 8 * 12 + 2 ; Cnt ++)
      _GadUpPg[Cnt] = (_GadUpPg[Cnt] ? DBLACK : DWHITE);
    for (Cnt = 2 ; Cnt < 8 * 12 + 2 ; Cnt ++)
      _GadDnBt[Cnt] = (_GadDnBt[Cnt] ? DBLACK : DWHITE);
    for (Cnt = 2 ; Cnt < 8 * 12 + 2 ; Cnt ++)
      _GadDnLn[Cnt] = (_GadDnLn[Cnt] ? DBLACK : DWHITE);
    for (Cnt = 2 ; Cnt < 8 * 12 + 2 ; Cnt ++)
      _GadDnPg[Cnt] = (_GadDnPg[Cnt] ? DBLACK : DWHITE);
    _GadgetsConverted = TRUE;
  }
  strcpy (SpecLabelsFullPath, ProgramDrive);
  strcat (SpecLabelsFullPath, ProgramPath);
  strcat (SpecLabelsFullPath, _SpectrumLabelFile);
  _ReadLabelFile (SpecLabelsFullPath, TRUE);                                           /* We want the Spectrum ROM symbols anyway */
  _IsSpectrumROM = FALSE;
  if (!memcmp (_ObjectData, _SpectrumROMWatermark, 8))                   /* See if the data starts with the 48K Spectrum ROM mark */
    _IsSpectrumROM = TRUE;
  FirstAddress._Address = *StartAddress;                                                    /* We'll start at the desired address */
  Resize = TRUE;
  while (Resize)                                                                 /* Either startup or the window has been resized */
  {
    Resize = FALSE;
    SelectFont (NULL, FONT_NORMAL);
    _CurrentHeight = (short)((OWindow.Height - OWADD) / (CurrentFont->Height + 2));
    _DisassemblerWidth = DISASSWIDTH;
    if (_IsSpectrumROM || _HaveLabels)
    {
      _DisassemblerWidth += 10;                                                                          /* Add space for symbols */
      SelectFont (NULL, FONT_NORMAL);
      if (PrevDisassemblerWidth == DISASSWIDTH)                                                                /* Enlarging now ? */
      {
        OWindow.StartX -= (5 * CurrentFont->Width);                                                   /* Keep the window centered */
        CWindow.StartX -= (5 * CurrentFont->Width);
        if (OWindow.StartX < 0)                                                                      /* And keep it on the screen */
        {
          OWindow.StartX = 0;
          CWindow.StartX = 5;
        }
      }
    }
    PrevDisassemblerWidth = _DisassemblerWidth;
    OWindow.Width = OWindow.MinWidth = OWindow.MaxWidth = _DisassemblerWidth * 6 + 26;
    CWindow.Width = CWindow.MinWidth = CWindow.MaxWidth = _DisassemblerWidth * 6;
    if (OWindow.StartX + OWindow.Width > Screen.Width)                /* Ensure the (possibly enlarged) window fits on the screen */
    {
      OWindow.StartX = Screen.Width - OWindow.Width;
      CWindow.StartX = OWindow.Width + 5;
    }
    OWindow.Height = (CurrentFont->Height + 2) * _CurrentHeight + OWADD;
    CWindow.StartY = OWindow.StartY + CWOFF;
    CWindow.Height = (CurrentFont->Height + 2) * _CurrentHeight + CWADD;
    POk.StartX = 275 + ((_IsSpectrumROM || _HaveLabels) ? 9 * CurrentFont->Width : 0);
    PSymbols.StartX = 200 + ((_IsSpectrumROM || _HaveLabels) ? 6 * CurrentFont->Width : 0);
    PHexDump.StartX = 125 + ((_IsSpectrumROM || _HaveLabels) ? 3 * CurrentFont->Width : 0);
    POk.StartY = PRestart.StartY = PHexDump.StartY = PSymbols.StartY = OWindow.Height - 25;
    PSave.StartX = _DisassemblerWidth * CurrentFont->Width - 38;
    PPageUp.StartX = PLineUp.StartX = PByteUp.StartX = PByteDn.StartX = PLineDn.StartX = PPageDn.StartX =
                     _DisassemblerWidth * CurrentFont->Width + 10;
    PPageUp.StartY = CWOFF + 1;
    PPageDn.StartY = OWindow.Height - OWADD + CWOFF - 11;
    PLineUp.StartY = PPageUp.StartY + 19;
    PLineDn.StartY = PPageDn.StartY - 19;
    PByteUp.StartY = PPageUp.StartY + 38;
    PByteDn.StartY = PPageDn.StartY - 38;
    AddButton (&OWindow, &POk, TRUE);
    AddButton (&OWindow, &PNumMode, FALSE);
    AddButton (&OWindow, &PSysMode, FALSE);
    AddButton (&OWindow, &PROMMode, FALSE);
    AddButton (&OWindow, &PSave, FALSE);
    AddButton (&OWindow, &PPageUp, FALSE);
    AddButton (&OWindow, &PLineUp, FALSE);
    AddButton (&OWindow, &PByteUp, FALSE);
    AddButton (&OWindow, &PByteDn, FALSE);
    AddButton (&OWindow, &PLineDn, FALSE);
    AddButton (&OWindow, &PPageDn, FALSE);
    AddButton (&OWindow, &PRestart, FALSE);
    AddButton (&OWindow, &PHexDump, FALSE);
    AddButton (&OWindow, &PSymbols, FALSE);
    DrawWindow (&OWindow, NULL, WING_CLOSE | WING_RESIZEMAX, TRUE);
    DrawWindow (&CWindow, &OWindow, 0, TRUE);
    Wprintf (&OWindow, 1, 3, "{+1Hex}N");
    Wprintf (&OWindow, 1, 20, "{+1Sysvars}N");
    Wprintf (&OWindow, 1, 37, "{+1Spectrum ROM}N");
    memcpy (&CurrentAddress, &FirstAddress, sizeof (struct _InstructionMemory_s));
    Restart = TRUE;
    while (Restart)                                                                   /* Write new or updated page of disassembly */
    {
      Restart = FALSE;
      memcpy (&FirstAddress, &CurrentAddress, sizeof (struct _InstructionMemory_s));          /* Store address on top of the page */
      for (LineCnt = 0 ; LineCnt < _CurrentHeight && CurrentAddress._Address < MaxBlockLength; LineCnt ++)       /* Disassemble ! */
      {
        _GetDisassemblyLine (&CurrentAddress, DisassembledLine);
        if (LineCnt == 0)
          memcpy (&NextAddress, &CurrentAddress, sizeof (struct _InstructionMemory_s));
        Wprintf (&CWindow, LineCnt, 0, DisassembledLine);
      }
      if (LineCnt < _CurrentHeight)                                                                 /* Clear any remaining bottom */
        for ( ; LineCnt < _CurrentHeight ; LineCnt ++)
          WClearLine (&CWindow, LineCnt);
      Done = FALSE;
      while (!Done)                                                                 /* Let the user decide what is to be done now */
      {
        switch (HandleWindow (&OWindow, &ButtonP))
        {
          case WINH_CLOSE     : Done = TRUE; break;
          case WINH_WINRESIZE : Done = TRUE; Resize = TRUE; break;
          case WINH_RNOHAND   : switch (ButtonP->ReactNum)
                                {
                                  case 10: if (_GetAddress (&FirstAddress._Address, FALSE))                          /* (Restart) */
                                           {
                                             Done = TRUE;
                                             Restart = TRUE;
                                             FirstAddress._Address -= LoadedAddress;
                                             FirstAddress._RSTValue = 0;
                                             memcpy (&CurrentAddress, &FirstAddress, sizeof (struct _InstructionMemory_s));
                                           }
                                           break;
                                  case 11: if (_EditSymbolTable (FirstAddress._Address))                             /* (Symbols) */
                                           {                     /* Redraw the window (possible change size) and the current page */
                                             memcpy (&CurrentAddress, &FirstAddress, sizeof (struct _InstructionMemory_s));
                                             Done = TRUE;
                                             Resize = TRUE;
                                           }
                                           break;
                                  case 12: Done = TRUE; break;                                                            /* (Ok) */
                                  case 13: SaveFirst._Address = LoadedAddress;                                         /* (Save) */
                                           SaveFirst._RSTValue = 0;
                                           SaveLast._Address = LoadedAddress + (word)MaxBlockLength - 1;
                                           if (_GetSaveRange (&(SaveFirst._Address), &(SaveLast._Address)))          /* Get range */
                                           {
                                             _splitpath (TapeWindow[WhichWindow].TapeName, SavePath, SpecLabelsFullPath,
                                                         SaveFilename, SaveExt);
                                             strcat (SavePath, SpecLabelsFullPath);
                                             strcat (SaveFilename, ".DIS");
                                             if (FileRequester (180, 50, 48, 29, 20, SavePath, SaveFilename, SaveMask,
                                                                "Choose Disassembly Save File", FALSE, FALSE))
                                             {
                                               MakeFullPath (SpecLabelsFullPath, SavePath, SaveFilename);
                                               if (!TestFileOverwrite (SpecLabelsFullPath))
                                                 if ((Fp = fopen (SpecLabelsFullPath, "w")) == NULL)
                                                   ErrorRequester (-1, -1, "@Close", NULL, NULL, "Error creating file %s:\n    %s",
                                                                   SpecLabelsFullPath, StrError (errno));
                                                 else
                                                 {
                                                   SavePreviousAddress = SaveFirst._Address;
                                                   while (SaveFirst._Address <= SaveLast._Address)               /* Disassemble ! */
                                                   {
                                                     if (SaveFirst._Address < SavePreviousAddress)           /* (Wrap protection) */
                                                       break;
                                                     _GetDisassemblyLine (&SaveFirst, DisassembledLine);
                                                     Cnt = _DisassemblerWidth;
                                                     while (DisassembledLine[-- Cnt] == ' ')            /* Remove trailing spaces */
                                                       ;
                                                     DisassembledLine[Cnt + 1] = '\0';
                                                     fprintf (Fp, "%s\n", DisassembledLine);
                                                     SavePreviousAddress = SaveFirst._Address;
                                                   }
                                                   fclose (Fp);
                                                 }
                                             }
                                           }
                                           break;
                                  case 14: *StartAddress = FirstAddress._Address;                                    /* (HexDump) */
                                           Done = TRUE;
                                           Restart = FALSE;
                                           StartHexDump = TRUE;
                                           break;
                                }
                                break;
          case WINH_BUTTBOOL  : switch (ButtonP->ReactNum)
                                {
                                  case 20:                                                                           /* (Dec/Hex) */
                                  case 21:                                                                           /* (Sysvars) */
                                  case 22: Done = TRUE;                                                                  /* (ROM) */
                                           Restart = TRUE;
                                           memcpy (&CurrentAddress, &FirstAddress, sizeof (struct _InstructionMemory_s));
                                }
                                break;
          case WINH_PNOHAND   : UsePagers = ButtonP->ReactNum; break;                              /* Merge buttons with keyboard */
        }
        switch (GlobalKey)
        {
          case K_ESC  : Done = TRUE; Restart = FALSE; break;
          case K_PGUP : UsePagers = 1; break;
          case K_CRUP : UsePagers = 2; break;
          case K_CRLT : UsePagers = 3; break;
          case K_CRRT : UsePagers = 4; break;
          case K_CRDN : UsePagers = 5; break;
          case K_PGDN : UsePagers = 6; break;
        }
        if (UsePagers >= 0)                                                                              /* Pager action needed ? */
        {
          switch (UsePagers)
          {
            case 1: if (FirstAddress._Address > 0)                                                                   /* (Page up) */
                    {
                      _ScrollBack (&FirstAddress, &CurrentAddress, TRUE);
                      Done = TRUE;
                      Restart = TRUE;
                    }
                    break;
            case 2: if (FirstAddress._Address > 0)                                                                   /* (Line up) */
                    {
                      _ScrollBack (&FirstAddress, &CurrentAddress, FALSE);
                      Done = TRUE;
                      Restart = TRUE;
                    }
                    break;
            case 3: if (FirstAddress._Address > 0)                                                                   /* (Byte up) */
                    {
                      memcpy (&CurrentAddress, &FirstAddress, sizeof (struct _InstructionMemory_s));
                      CurrentAddress._Address --;
                      Done = TRUE;
                      Restart = TRUE;
                    }
                    break;
            case 4: if (CurrentAddress._Address < MaxBlockLength)                                                  /* (Byte down) */
                    {
                      memcpy (&CurrentAddress, &FirstAddress, sizeof (struct _InstructionMemory_s));
                      CurrentAddress._Address ++;
                      Done = TRUE;
                      Restart = TRUE;
                    }
                    break;
            case 5: if (CurrentAddress._Address < MaxBlockLength)                                                  /* (Line down) */
                    {
                      memcpy (&CurrentAddress, &NextAddress, sizeof (struct _InstructionMemory_s));
                      Done = TRUE;
                      Restart = TRUE;
                    }
                    break;
            case 6: if (CurrentAddress._Address < MaxBlockLength)                                                  /* (Page down) */
                    {
                      Done = TRUE;
                      Restart = TRUE;
                    }
                    break;
          }
          UsePagers = -1;
        }
        MouseStatus ();
      }
    }
    DestroyWindow (&OWindow);
    DestroyWindow (&CWindow);
  }
  GlobalKey = 0x0000;
  return (StartHexDump);
}

void CleanupDisassemblerLabels (void)

/**********************************************************************************************************************************/
/* Pre   : None (called from ReleaseTape).                                                                                        */
/* Post  : The currently loaded tape is being destroyed. If the external symbol table was not saved yet, the user must confirm.   */
/* Import: _SaveExtLabels, _FlushExtLabels.                                                                                       */
/**********************************************************************************************************************************/

{
  if (_LabelsChanged && _NumLabels > 0)                                               /* External symbol table has been changed ? */
  {
    switch (QuestionRequester (-1, -1, TRUE, FALSE, "@Ok", "@Save", "Save @As", NULL, NULL, "Confirm Destroy Changes",
                               "The symbol table has been changed!\nThrow away these changes?"))
    {
      case 1 : _SaveExtLabels (FALSE); break;                                                                           /* (Save) */
      case 2 : _SaveExtLabels (TRUE); break;                                                                         /* (Save As) */
    }
    _FlushExtLabels ();                                                                        /* Ditch any external symbol table */
  }
}

bool ViewSpectrumDump (dword BlockLength, word *StartAddress)

/**********************************************************************************************************************************/
/* Pre   : `BlockLength' holds the length of the block. The block has already been moved to _ObjectData.                          */
/* Post  : The selected block has been shown as a dump in a seperate window, 16 bytes a line, both hex and ASCII.                 */
/*         It may have been edited. The return value is FALSE if `Ok' or window close was pressed, TRUE if the disassemler should */
/*         be called from the edit address.                                                                                       */
/* Import: _DrawEditCursor, _EditFind.                                                                                            */
/**********************************************************************************************************************************/

{
#define CWOFF        30
#define CWADD        4
#define OWADD        (CWADD + CWOFF + 38)

  struct WindowInfo      OWindow  = {72, 90, 496, DEFVWLINES * 10 + OWADD, 496, MINVWLINES * 10 + OWADD, 496,
                                     MAXVWLINES * 10 + OWADD, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                     WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Tape Block Dump"};
  struct ButtonInfo      POk      = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 350, 427, 8, DBLACK, DWHITE, NULL,
                                     1, "@Ok", NULL};
  struct ButtonInfo      PDisass  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 100, 427, 8, DBLACK, DWHITE, NULL,
                                     2, "@Disass", NULL};
  struct ButtonInfo      PFind    = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 225, 427, 8, DBLACK, DWHITE, NULL,
                                     3, "@Find", NULL};
  struct ButtonInfo      PNumMode = {BT_BOOL, BC_GADGET, BF_FWIDTH | BF_BCLEAR | BF_KEYSELECT, 5, 9, 1, DBLACK, DWHITE, NULL,
                                     20, &_NumbersAreHex, NULL};
  struct ButtonInfo      PEdAddr  = {BT_TEXT, BC_TEXT, BF_FWIDTH | BF_FITTXT, 82, 9, 5, DBLACK, DWHITE, NULL, 0, NULL, NULL};
  char                   AEdAddr[6];
  struct MouseShape     *OldShape;
  struct ButtonInfo     *ButtonP;
  struct ComboEntry      Entry;
  comboent               StartingLine;
  comboent               LastKnownPosition;
  bool                   More = TRUE;
  bool                   Resize = TRUE;
  bool                   Restart = TRUE;
  bool                   FirstLine;
  comboent               Dummy;
  dword                  Index;
  dword                  SubIndex;
  short                  LineIndex;
  short                  FirstSkip;
  short                  FirstSkipKeep;
  word                   RealAdd;
  byte                   Bt;
  byte                   BtUpper;
  bool                   StartDisassembler = FALSE;
  short                  CurTmpX;
  short                  CurTmpY;
  bool                   CurTmpHalf;
  word                   CurTmpIndex;
  word                   CurTmpAddress;
  bool                   AcceptClick;
  bool                   SubProcessActive = FALSE;                                  /* TRUE if the cursor flash handler is active */
  bool                   ReInitFind = FALSE;

  AEdAddr[0] = '\0';
  PEdAddr.Contents = AEdAddr;
  LastKnownPosition = -1;
  if (_EditCurX == -1)
    ReInitFind = TRUE;
  while (Resize)
  {
    Resize = FALSE;
    SelectFont (NULL, FONT_NORMAL);
    _CurrentHeight = (short)((OWindow.Height - OWADD) / (CurrentFont->Height + 2));
    POk.StartY = PDisass.StartY = PFind.StartY = OWindow.Height - 25;
    AddButton (&OWindow, &POk, TRUE);
    AddButton (&OWindow, &PNumMode, FALSE);
    AddButton (&OWindow, &PEdAddr, FALSE);
    AddButton (&OWindow, &PDisass, FALSE);
    AddButton (&OWindow, &PFind, FALSE);
    DrawWindow (&OWindow, NULL, WING_CLOSE | WING_RESIZEMAX, TRUE);
    Wprintf (&OWindow, 1, 3, "{+1Hex}N");
    Wprintf (&OWindow, 1, 20, "{+1Address}N");
    MakeComboWindow (&_EditHWindow, &OWindow, COMT_VIEW, COMF_NOBUFFER | COMF_VSLIDER | COMF_SYNCSLIDER,
                     OWindow.StartX + 5, OWindow.StartY + CWOFF, 0, 1024, 60, 56, _CurrentHeight, NULL, FALSE, 0, NULL,
                     SystemColors.ReqAPen, SystemColors.ReqBPen, 0, 0);
    MakeComboWindow (&_EditAWindow, &OWindow, COMT_VIEW, COMF_NOBUFFER,
                     OWindow.StartX + 358, OWindow.StartY + CWOFF, 0, 1024, 32, 16, _CurrentHeight, SpectrumFont, FALSE, 0, NULL,
                     SystemColors.ReqAPen, SystemColors.ReqBPen, 0, 0);
    OldShape = CurrentMouseShape;
    SetMouseShape (DefaultWaitingMouseShape);
    Restart = TRUE;
    while (Restart)
    {
      Restart = FALSE;
      More = TRUE;
      FirstLine = TRUE;
      if ((LoadedAddress & 0xFFF0) == LoadedAddress)
        FirstLine = FALSE;
      Entry.Selected = 0;
      Entry.Identifier = 0;
      FirstSkipKeep = FirstSkip = 0;
      for (Index = 0 ; Index < BlockLength && More ; )
      {
        memset (Entry.Text, ' ', 56);
        if (FirstLine)
        {
          RealAdd = LoadedAddress & 0xFFF0;
          FirstSkipKeep = FirstSkip = LoadedAddress - RealAdd;
        }
        else
          RealAdd = (word)(Index + LoadedAddress);
        if (_NumbersAreHex)
          sprintf (Entry.Text, " %04X - ", RealAdd);
        else
          sprintf (Entry.Text, "%5u - ", RealAdd);
        Entry.Identifier = (dword)RealAdd;
        LineIndex = 8;
        if (FirstSkip > 0)
        {
          Entry.Text[LineIndex] = ' ';
          LineIndex += FirstSkip * 3;
          if (FirstSkip >= 8)
            LineIndex ++;
        }
        for (SubIndex = FirstSkip ; SubIndex < 16 && Index + SubIndex < BlockLength ; )                     /* Build the hex part */
        {
          sprintf (Entry.Text + LineIndex, "%02X ", *((byte *)_ObjectData + Index + SubIndex - FirstSkip));
          LineIndex += 3;
          Entry.Text[LineIndex] = ' ';                                                            /* Don't want string endmarkers */
          if (++ SubIndex == 8)
            LineIndex ++;                                                                              /* Add extra blank at half */
        }
        Entry.Text[-- LineIndex] = '\0';
        if ((More = InsertComboEntry (&_EditHWindow, &Entry, NULL, -1)))                                       /* Insert the line */
        {
          LineIndex = 0;
          if (FirstSkip > 0)
          {
            memset (Entry.Text, ' ', FirstSkip);
            LineIndex += FirstSkip;
          }
          for (SubIndex = FirstSkip ; SubIndex < 16 && Index + SubIndex < BlockLength ; )                 /* Build the ASCII part */
          {
            Bt = *((byte *)_ObjectData + Index + SubIndex - FirstSkip);
            if (Bt >= 32 && Bt < 128)
            {
              if (Bt == '{' || Bt == '}')                                                                           /* Plain text */
                Entry.Text[LineIndex ++] = Bt;
              Entry.Text[LineIndex ++] = Bt;
            }
            else
              Entry.Text[LineIndex ++] = '.';
            SubIndex ++;
          }
          Entry.Text[LineIndex] = '\0';                                                                      /* Add the endmarker */
          More = InsertComboEntry (&_EditAWindow, &Entry, NULL, -1);                                           /* Insert the line */
        }
        FirstLine = FALSE;
        Index += (16 - FirstSkip);
        FirstSkip = 0;
      }
      if (!More)                                                                                /* Listing truncated prematurally */
      {
        if (_EditHWindow.NumEntries == -1)                                                                   /* No entry at all ? */
          FatalError (1, _ErrOutOfMemory, 0);                                                                        /* Then quit */
        GetComboEntry (&_EditHWindow, &Entry, _EditHWindow.NumEntries);
        strcpy (Entry.Text, "{+1>>> TRUNCATED <<<}N");
        PutComboEntry (&_EditHWindow, &Entry, _EditHWindow.NumEntries, FALSE);
        More = TRUE;
      }
      if (*StartAddress > 16)                                                                 /* At least one full line forward ? */
      {
        StartingLine = *StartAddress / 16;                                                             /* Determine starting line */
        if (LastKnownPosition >= 0)                                                             /* (Resizing rather than startup) */
        {
          if (StartingLine >= LastKnownPosition && StartingLine < (LastKnownPosition + _CurrentHeight))
            StartingLine = LastKnownPosition;                            /* Keep at the same line if the cursor is still focussed */
          else
            StartingLine -= (_CurrentHeight / 2);                                          /* Bring the cursor to the middle line */
        }
        else
          StartingLine -= (_CurrentHeight / 2);
        if (StartingLine < 0)                                                               /* Ensure the page fits in the window */
          StartingLine = 0;
        else if (StartingLine + _CurrentHeight > _EditHWindow.NumEntries)
          StartingLine = _EditHWindow.NumEntries - _CurrentHeight + 1;
        _EditAWindow.DFirstLine = _EditHWindow.DFirstLine = StartingLine;                     /* Make the dump start at that line */
        _EditHWindow.VSlider->SliderStart = _EditHWindow.VSlider->StepRate * StartingLine;
      }
      PrintComboEntries (&_EditHWindow);                                                                    /* Print the contents */
      PrintComboEntries (&_EditAWindow);
      if (!SubProcessActive)
      {                                                            /* (Don't move the cursor when the dec/hex button was pressed) */
        _EditWhichWindow = TRUE;
        _EditCurAddress = *StartAddress + LoadedAddress;
        _EditCurIndex = _EditCurAddress % 16;
        _EditCurX = 8 + 3 * _EditCurIndex + (_EditCurIndex >= 8 ? 1 : 0);
        _EditCurHalf = TRUE;
        _EditCurY = (*StartAddress + (LoadedAddress & 0x000F)) / 16;
        if (_NumbersAreHex)
          sprintf (AEdAddr, " %04X", _EditCurAddress);
        else
          sprintf (AEdAddr, "%5u", _EditCurAddress);
        UpdateButtonText (&OWindow, &PEdAddr, BF_FITTXT);
        RegisterTimedHandler (500, _FlashEditCursor, &OWindow);
        SubProcessActive = TRUE;
      }
      SetMouseShape (OldShape);
      LastKnownPosition = _EditHWindow.DFirstLine;
      while (More)
      {
        switch (HandleComboWindow (&_EditHWindow, &Dummy))
        {
          case WINH_VERMOVE : if (_EditHWindow.DFirstLine != LastKnownPosition)
                              {
                                LastKnownPosition = _EditAWindow.DFirstLine = _EditHWindow.DFirstLine;
                                PrintComboEntries (&_EditAWindow);                             /* Update the ASCII window as well */
                              }
                              break;
        }
        switch (HandleWindow (&OWindow, &ButtonP))
        {
          case WINH_CLOSE    : More = FALSE; break;
          case WINH_WINRESIZE: if (_EditCurX < 0)                                                              /* (Window resize) */
                                 *StartAddress = (word)_EditHWindow.DFirstLine * 16;
                               else
                                 *StartAddress = (word)(_EditCurAddress - LoadedAddress);
                               LastKnownPosition = _EditHWindow.DFirstLine;                                  /* Remember top line */
                               More = FALSE;
                               Resize = TRUE;
                               break;
          case WINH_RNOHAND  : switch (ButtonP->ReactNum)
                               {
                                 case  1 : More = FALSE; break;                                                           /* (Ok) */
                                 case  2 : if (_EditCurX < 0)                                                         /* (Disass) */
                                             *StartAddress = (word)_EditHWindow.DFirstLine * 16;
                                           else
                                             *StartAddress = (word)(_EditCurAddress - LoadedAddress);
                                           More = FALSE;
                                           StartDisassembler = TRUE;                        /* Start disassembler at that address */
                                           break;
                                 case  3 : if ((Index = _EditFind (ReInitFind, BlockLength)) != FINDNOMATCH)            /* (Find) */
                                           {                                                                         /* Found it! */
                                             if (SubProcessActive)                                      /* Release current cursor */
                                             {
                                               ReleaseTimedHandler (_FlashEditCursor, &OWindow);
                                               SubProcessActive = FALSE;
                                             }
                                             *StartAddress = (word)Index;
                                             LastKnownPosition = _EditHWindow.DFirstLine;
                                             ReInitFind = FALSE;
                                             More = FALSE;
                                             Restart = TRUE;
                                           }
                                           break;
                               }
                               break;
          case WINH_BUTTBOOL : switch (ButtonP->ReactNum)
                               {
                                 case  20: *StartAddress = (word)_EditHWindow.DFirstLine * 16;                       /* (Dec/Hex) */
                                           if (_EditCurX >= 0)
                                           {
                                             if (_NumbersAreHex)
                                               sprintf (AEdAddr, " %04X", _EditCurAddress);
                                             else
                                               sprintf (AEdAddr, "%5u", _EditCurAddress);
                                             UpdateButtonText (&OWindow, &PEdAddr, BF_FITTXT);
                                           }
                                           More = FALSE;
                                           Restart = TRUE;
                                           break;
                               }
                               break;
        }
        if (GlobalKey == K_ESC)
          More = FALSE;
        else if (Mouse.Left && !Mouse.LeftStillPressed)                                              /* Handle other mouse clicks */
          if (Mouse.Xx > (_EditHWindow.ComboWindow.StartX + 2) &&
              Mouse.Xx < (_EditHWindow.ComboWindow.StartX + _EditHWindow.ComboWindow.Width - 2) &&
              Mouse.Yy > (_EditHWindow.ComboWindow.StartY + 2) &&
              Mouse.Yy < (_EditHWindow.ComboWindow.StartY + _EditHWindow.ComboWindow.Height - 2))
          {                                                                                   /* Click inside the hex-part window */
            CurTmpX = (Mouse.Xx - _EditHWindow.ComboWindow.StartX - 2) / DefaultFont->Width;
            CurTmpY = ((Mouse.Yy - _EditHWindow.ComboWindow.StartY - 2) / (DefaultFont->Height + 2)) + _EditHWindow.DFirstLine;
            switch (CurTmpX)
            {                                                                                         /* Check the position 'map' */
              case  8: CurTmpHalf = TRUE;  CurTmpIndex =  0; AcceptClick = TRUE; break;
              case  9: CurTmpHalf = FALSE; CurTmpIndex =  0; AcceptClick = TRUE; break;
              case 11: CurTmpHalf = TRUE;  CurTmpIndex =  1; AcceptClick = TRUE; break;
              case 12: CurTmpHalf = FALSE; CurTmpIndex =  1; AcceptClick = TRUE; break;
              case 14: CurTmpHalf = TRUE;  CurTmpIndex =  2; AcceptClick = TRUE; break;
              case 15: CurTmpHalf = FALSE; CurTmpIndex =  2; AcceptClick = TRUE; break;
              case 17: CurTmpHalf = TRUE;  CurTmpIndex =  3; AcceptClick = TRUE; break;
              case 18: CurTmpHalf = FALSE; CurTmpIndex =  3; AcceptClick = TRUE; break;
              case 20: CurTmpHalf = TRUE;  CurTmpIndex =  4; AcceptClick = TRUE; break;
              case 21: CurTmpHalf = FALSE; CurTmpIndex =  4; AcceptClick = TRUE; break;
              case 23: CurTmpHalf = TRUE;  CurTmpIndex =  5; AcceptClick = TRUE; break;
              case 24: CurTmpHalf = FALSE; CurTmpIndex =  5; AcceptClick = TRUE; break;
              case 26: CurTmpHalf = TRUE;  CurTmpIndex =  6; AcceptClick = TRUE; break;
              case 27: CurTmpHalf = FALSE; CurTmpIndex =  6; AcceptClick = TRUE; break;
              case 29: CurTmpHalf = TRUE;  CurTmpIndex =  7; AcceptClick = TRUE; break;
              case 30: CurTmpHalf = FALSE; CurTmpIndex =  7; AcceptClick = TRUE; break;
              case 33: CurTmpHalf = TRUE;  CurTmpIndex =  8; AcceptClick = TRUE; break;
              case 34: CurTmpHalf = FALSE; CurTmpIndex =  8; AcceptClick = TRUE; break;
              case 36: CurTmpHalf = TRUE;  CurTmpIndex =  9; AcceptClick = TRUE; break;
              case 37: CurTmpHalf = FALSE; CurTmpIndex =  9; AcceptClick = TRUE; break;
              case 39: CurTmpHalf = TRUE;  CurTmpIndex = 10; AcceptClick = TRUE; break;
              case 40: CurTmpHalf = FALSE; CurTmpIndex = 10; AcceptClick = TRUE; break;
              case 42: CurTmpHalf = TRUE;  CurTmpIndex = 11; AcceptClick = TRUE; break;
              case 43: CurTmpHalf = FALSE; CurTmpIndex = 11; AcceptClick = TRUE; break;
              case 45: CurTmpHalf = TRUE;  CurTmpIndex = 12; AcceptClick = TRUE; break;
              case 46: CurTmpHalf = FALSE; CurTmpIndex = 12; AcceptClick = TRUE; break;
              case 48: CurTmpHalf = TRUE;  CurTmpIndex = 13; AcceptClick = TRUE; break;
              case 49: CurTmpHalf = FALSE; CurTmpIndex = 13; AcceptClick = TRUE; break;
              case 51: CurTmpHalf = TRUE;  CurTmpIndex = 14; AcceptClick = TRUE; break;
              case 52: CurTmpHalf = FALSE; CurTmpIndex = 14; AcceptClick = TRUE; break;
              case 54: CurTmpHalf = TRUE;  CurTmpIndex = 15; AcceptClick = TRUE; break;
              case 55: CurTmpHalf = FALSE; CurTmpIndex = 15; AcceptClick = TRUE; break;
              default: AcceptClick = FALSE; break;
            }
            if (AcceptClick)
            {
              CurTmpAddress = (word)CurTmpY * 16 + CurTmpIndex;
              if ((dword)CurTmpAddress + (LoadedAddress & 0xFFF0) < LoadedAddress)                /* Before incomplete first line */
                AcceptClick = FALSE;
              else if ((dword)CurTmpAddress >= BlockLength + (LoadedAddress & 0x000F))
                AcceptClick = FALSE;                                                                /* After incomplete last line */
            }
            if (AcceptClick)
            {
              _DrawEditCursor (FALSE);                                                                   /* Remove current cursor */
              _EditWhichWindow = TRUE;
              _EditCurX = CurTmpX;
              _EditCurY = CurTmpY;
              _EditCurHalf = CurTmpHalf;
              _EditCurIndex = CurTmpIndex;
              _EditCurAddress = CurTmpAddress + (LoadedAddress & 0xFFF0);
              if (_NumbersAreHex)
                sprintf (AEdAddr, " %04X", _EditCurAddress);
              else
                sprintf (AEdAddr, "%5u", _EditCurAddress);
              UpdateButtonText (&OWindow, &PEdAddr, BF_FITTXT);
            }
          }
          else if (Mouse.Xx > (_EditAWindow.ComboWindow.StartX + 2) &&
                   Mouse.Xx < (_EditAWindow.ComboWindow.StartX + _EditAWindow.ComboWindow.Width - 2) &&
                   Mouse.Yy > (_EditAWindow.ComboWindow.StartY + 2) &&
                   Mouse.Yy < (_EditAWindow.ComboWindow.StartY + _EditAWindow.ComboWindow.Height - 2))
          {                                                                                 /* Click inside the ASCII-part window */
            CurTmpX = (Mouse.Xx - _EditAWindow.ComboWindow.StartX - 2) / SpectrumFont->Width;
            CurTmpY = (Mouse.Yy - _EditAWindow.ComboWindow.StartY - 2) / (SpectrumFont->Height + 2) + _EditAWindow.DFirstLine;
            CurTmpIndex = CurTmpX;
            CurTmpAddress = (word)CurTmpY * 16 + CurTmpIndex;
            AcceptClick = TRUE;
            if ((dword)CurTmpAddress + (LoadedAddress & 0xFFF0) < LoadedAddress)                  /* Before incomplete first line */
              AcceptClick = FALSE;
            else if ((dword)CurTmpAddress >= BlockLength + (LoadedAddress & 0x000F))
              AcceptClick = FALSE;                                                                  /* After incomplete last line */
            if (AcceptClick)
            {
              _DrawEditCursor (FALSE);                                                                   /* Remove current cursor */
              _EditWhichWindow = FALSE;
              _EditCurX = CurTmpX;
              _EditCurY = CurTmpY;
              _EditCurIndex = CurTmpIndex;
              _EditCurAddress = CurTmpAddress + (LoadedAddress & 0xFFF0);
              if (_NumbersAreHex)
                sprintf (AEdAddr, " %04X", _EditCurAddress);
              else
                sprintf (AEdAddr, "%5u", _EditCurAddress);
              UpdateButtonText (&OWindow, &PEdAddr, BF_FITTXT);
            }
          }
        if (GlobalKey != 0x0000)                                                                           /* A key was pressed ? */
          if (GlobalShiftStatus & KS_ACTRL)                                                           /* [CTRL] pressed as well ? */
          {
            switch (GlobalKey)
            {
              case 0x9400 : _DrawEditCursor (FALSE);                                                                     /* [TAB] */
                            if (_EditWhichWindow)
                              _EditCurX = _EditCurIndex;
                            else
                            {
                              _EditCurX = 8 + 3 * _EditCurIndex + (_EditCurIndex >= 8 ? 1 : 0);
                              _EditCurHalf = TRUE;
                            }
                            _EditWhichWindow = !_EditWhichWindow;
                            GlobalKey = 0x0000;                                                                /* Consume the key */
                            break;
            }
          }
          else if (!(GlobalShiftStatus & (KS_ACTRL | KS_AALT)))                                            /* No [CTRL] nor [ALT] */
          {
            switch (GlobalKey)
            {
              case K_PGDN : if (_EditCurY >= _EditHWindow.DFirstLine &&                       /* (Only if the cursor is focussed) */
                                _EditCurY < (_EditHWindow.DFirstLine + _EditHWindow.DHeight))
                            {
                              if (_EditCurY != (_EditHWindow.DFirstLine + _EditHWindow.DHeight - 1))   /* Not at the bottom yet ? */
                              {
                                CurTmpY = _EditHWindow.DFirstLine + _EditHWindow.DHeight - 1;                 /* Go to the bottom */
                                StartingLine = _EditHWindow.DFirstLine;
                              }
                              else
                              {
                                CurTmpY = _EditCurY + _EditHWindow.DHeight;                                  /* Go full page down */
                                if (CurTmpY > _EditHWindow.NumEntries)                                          /* (Out of range) */
                                  CurTmpY = _EditHWindow.NumEntries;
                                StartingLine = CurTmpY - _EditHWindow.DHeight + 1;
                              }
                              CurTmpAddress = (word)CurTmpY * 16 + _EditCurIndex;                            /* Determine address */
                              if ((dword)CurTmpAddress >= BlockLength + (LoadedAddress & 0x000F))       /* (Incomplete last line) */
                              {
                                CurTmpAddress = BlockLength + (LoadedAddress & 0x000F) - 1;
                                CurTmpIndex = (CurTmpAddress & 0x000F);
                              }
                              else
                                CurTmpIndex = _EditCurIndex;
                              CurTmpAddress += (LoadedAddress & 0xFFF0);                                    /* Get 'real' address */
                              if (CurTmpAddress != _EditCurAddress)                         /* A reposition is going to be made ? */
                              {
                                _DrawEditCursor (FALSE);
                                _EditCurAddress = CurTmpAddress;
                                _EditCurY = CurTmpY;
                                _EditCurIndex = CurTmpIndex;
                                if (_EditWhichWindow)
                                {
                                  _EditCurX = 8 + 3 * _EditCurIndex + (_EditCurIndex >= 8 ? 1 : 0);
                                  _EditCurHalf = TRUE;
                                }
                                else
                                  _EditCurX = _EditCurIndex;
                                if (_EditHWindow.DFirstLine != StartingLine)                             /* Need to scroll down ? */
                                {
                                  LastKnownPosition = _EditHWindow.DFirstLine = _EditAWindow.DFirstLine = StartingLine;
                                  _EditHWindow.VSlider->SliderStart = _EditHWindow.VSlider->StepRate * _EditHWindow.DFirstLine;
                                  PrintComboEntries (&_EditHWindow);
                                  PrintComboEntries (&_EditAWindow);
                                }
                                if (_NumbersAreHex)
                                  sprintf (AEdAddr, " %04X", _EditCurAddress);
                                else
                                  sprintf (AEdAddr, "%5u", _EditCurAddress);
                                UpdateButtonText (&OWindow, &PEdAddr, BF_FITTXT);
                              }
                              else
                                SoundErrorBeeper (BEEP_SHORT);
                            }
                            GlobalKey = 0x0000;
                            break;
              case K_PGUP : if (_EditCurY >= _EditHWindow.DFirstLine &&                       /* (Only if the cursor is focussed) */
                                _EditCurY < (_EditHWindow.DFirstLine + _EditHWindow.DHeight))
                            {
                              if (_EditCurY != _EditHWindow.DFirstLine)                                   /* Not at the top yet ? */
                              {
                                CurTmpY = _EditHWindow.DFirstLine;                                               /* Go to the top */
                                StartingLine = _EditHWindow.DFirstLine;
                              }
                              else
                              {
                                CurTmpY = _EditCurY - _EditHWindow.DHeight;                                    /* Go full page up */
                                if (CurTmpY < 0)                                                                /* (Out of range) */
                                  CurTmpY = 0;
                                StartingLine = CurTmpY;
                              }
                              CurTmpAddress = (word)CurTmpY * 16 + _EditCurIndex;                            /* Determine address */
                              if ((dword)CurTmpAddress + (LoadedAddress & 0xFFF0) < LoadedAddress)     /* (Incomplete first line) */
                                CurTmpAddress = CurTmpIndex = (LoadedAddress & 0x000F);
                              else
                                CurTmpIndex = _EditCurIndex;
                              CurTmpAddress += (LoadedAddress & 0xFFF0);                                    /* Get 'real' address */
                              if (CurTmpAddress != _EditCurAddress)                         /* A reposition is going to be made ? */
                              {
                                _DrawEditCursor (FALSE);
                                _EditCurAddress = CurTmpAddress;
                                _EditCurY = CurTmpY;
                                _EditCurIndex = CurTmpIndex;
                                if (_EditWhichWindow)
                                {
                                  _EditCurX = 8 + 3 * _EditCurIndex + (_EditCurIndex >= 8 ? 1 : 0);
                                  _EditCurHalf = TRUE;
                                }
                                else
                                  _EditCurX = _EditCurIndex;
                                if (_EditHWindow.DFirstLine != StartingLine)                             /* Need to scroll down ? */
                                {
                                  LastKnownPosition = _EditHWindow.DFirstLine = _EditAWindow.DFirstLine = StartingLine;
                                  _EditHWindow.VSlider->SliderStart = _EditHWindow.VSlider->StepRate * _EditHWindow.DFirstLine;
                                  PrintComboEntries (&_EditHWindow);
                                  PrintComboEntries (&_EditAWindow);
                                }
                                if (_NumbersAreHex)
                                  sprintf (AEdAddr, " %04X", _EditCurAddress);
                                else
                                  sprintf (AEdAddr, "%5u", _EditCurAddress);
                                UpdateButtonText (&OWindow, &PEdAddr, BF_FITTXT);
                              }
                              else
                                SoundErrorBeeper (BEEP_SHORT);
                            }
                            GlobalKey = 0x0000;
                            break;
              case K_CRDN : if (_EditCurAddress + 16 < LoadedAddress + BlockLength)                           /* Can we go down ? */
                            {
                              _DrawEditCursor (FALSE);
                              _EditCurAddress += 16;
                              _EditCurY ++;
                              if (_EditCurY == (_EditHWindow.DFirstLine + _EditHWindow.DHeight))         /* Need to scroll down ? */
                              {
                                _EditHWindow.DFirstLine ++;
                                _EditAWindow.DFirstLine ++;
                                LastKnownPosition = _EditHWindow.DFirstLine;
                                _EditHWindow.VSlider->SliderStart = _EditHWindow.VSlider->StepRate * _EditHWindow.DFirstLine;
                                PrintComboEntries (&_EditHWindow);
                                PrintComboEntries (&_EditAWindow);
                              }
                              if (_NumbersAreHex)
                                sprintf (AEdAddr, " %04X", _EditCurAddress);
                              else
                                sprintf (AEdAddr, "%5u", _EditCurAddress);
                              UpdateButtonText (&OWindow, &PEdAddr, BF_FITTXT);
                            }
                            else
                              SoundErrorBeeper (BEEP_SHORT);
                            GlobalKey = 0x0000;
                            break;
              case K_CRUP : if (_EditCurAddress - 16 >= LoadedAddress)                                          /* Can we go up ? */
                            {
                              _DrawEditCursor (FALSE);
                              _EditCurAddress -= 16;
                              _EditCurY --;
                              if (_EditCurY == _EditHWindow.DFirstLine - 1)                                /* Need to scroll up ? */
                              {
                                _EditHWindow.DFirstLine --;
                                _EditAWindow.DFirstLine --;
                                LastKnownPosition = _EditHWindow.DFirstLine;
                                _EditHWindow.VSlider->SliderStart = _EditHWindow.VSlider->StepRate * _EditHWindow.DFirstLine;
                                PrintComboEntries (&_EditHWindow);
                                PrintComboEntries (&_EditAWindow);
                              }
                              if (_NumbersAreHex)
                                sprintf (AEdAddr, " %04X", _EditCurAddress);
                              else
                                sprintf (AEdAddr, "%5u", _EditCurAddress);
                              UpdateButtonText (&OWindow, &PEdAddr, BF_FITTXT);
                            }
                            else
                              SoundErrorBeeper (BEEP_SHORT);
                            GlobalKey = 0x0000;
                            break;
              case K_CRRT : AcceptClick = TRUE;
                            if (_EditWhichWindow)                                                                 /* (Hex window) */
                            {
                              if (_EditCurHalf)                                          /* Currently appointing left byte half ? */
                              {
                                _DrawEditCursor (FALSE);
                                _EditCurX ++;
                                _EditCurHalf = FALSE;
                                AcceptClick = FALSE;                                                              /* Signal: done */
                              }
                              else if (_EditCurAddress < LoadedAddress + BlockLength - 1)               /* Is there a next byte ? */
                              {
                                _DrawEditCursor (FALSE);
                                _EditCurAddress ++;
                                _EditCurHalf = TRUE;
                                if (++ _EditCurIndex > 15)                                                       /* End of line ? */
                                {
                                  _EditCurIndex = 0;
                                  _EditCurY ++;
                                  _EditCurX = 8;
                                }
                                else
                                  _EditCurX = 8 + 3 * _EditCurIndex + (_EditCurIndex >= 8 ? 1 : 0);
                              }
                              else
                              {
                                AcceptClick = FALSE;
                                SoundErrorBeeper (BEEP_SHORT);
                              }
                            }
                            else                                                                                /* (ASCII window) */
                              if (_EditCurAddress < LoadedAddress + BlockLength - 1)                    /* Is there a next byte ? */
                              {
                                _DrawEditCursor (FALSE);
                                _EditCurAddress ++;
                                if (++ _EditCurIndex > 15)                                                       /* End of line ? */
                                {
                                  _EditCurX = _EditCurIndex = 0;
                                  _EditCurY ++;
                                }
                                else
                                  _EditCurX = _EditCurIndex;
                              }
                              else
                              {
                                AcceptClick = FALSE;
                                SoundErrorBeeper (BEEP_SHORT);
                              }
                            if (AcceptClick)                                                       /* Need to adjust the screen ? */
                            {
                              if (_EditCurY == (_EditHWindow.DFirstLine + _EditHWindow.DHeight))         /* Need to scroll down ? */
                              {
                                _EditHWindow.DFirstLine ++;
                                _EditAWindow.DFirstLine ++;
                                LastKnownPosition = _EditHWindow.DFirstLine;
                                _EditHWindow.VSlider->SliderStart = _EditHWindow.VSlider->StepRate * _EditHWindow.DFirstLine;
                                PrintComboEntries (&_EditHWindow);
                                PrintComboEntries (&_EditAWindow);
                              }
                              if (_NumbersAreHex)
                                sprintf (AEdAddr, " %04X", _EditCurAddress);
                              else
                                sprintf (AEdAddr, "%5u", _EditCurAddress);
                              UpdateButtonText (&OWindow, &PEdAddr, BF_FITTXT);
                            }
                            GlobalKey = 0x0000;
                            break;
              case K_CRLT : AcceptClick = TRUE;
                            if (_EditWhichWindow)                                                                 /* (Hex window) */
                            {
                              if (!_EditCurHalf)                                        /* Currently appointing right byte half ? */
                              {
                                _DrawEditCursor (FALSE);
                                _EditCurX --;
                                _EditCurHalf = TRUE;
                                AcceptClick = FALSE;                                                              /* Signal: done */
                              }
                              else if (_EditCurAddress > LoadedAddress)                             /* Is there a previous byte ? */
                              {
                                _DrawEditCursor (FALSE);
                                _EditCurAddress --;
                                _EditCurHalf = FALSE;
                                if (_EditCurIndex == 0)                                                        /* Start of line ? */
                                {
                                  _EditCurIndex = 15;
                                  _EditCurY --;
                                }
                                else
                                  _EditCurIndex --;
                                _EditCurX = 9 + 3 * _EditCurIndex + (_EditCurIndex >= 8 ? 1 : 0);
                              }
                              else
                              {
                                AcceptClick = FALSE;
                                SoundErrorBeeper (BEEP_SHORT);
                              }
                            }
                            else                                                                                /* (ASCII window) */
                              if (_EditCurAddress > LoadedAddress)                                  /* Is there a previous byte ? */
                              {
                                _DrawEditCursor (FALSE);
                                _EditCurAddress --;
                                if (_EditCurIndex == 0)                                                        /* Start of line ? */
                                {
                                  _EditCurX = _EditCurIndex = 15;
                                  _EditCurY ++;
                                }
                                else
                                  _EditCurX = -- _EditCurIndex;
                              }
                              else
                              {
                                AcceptClick = FALSE;
                                SoundErrorBeeper (BEEP_SHORT);
                              }
                            if (AcceptClick)                                                       /* Need to adjust the screen ? */
                            {
                              if (_EditCurY == _EditHWindow.DFirstLine - 1)                              /* Need to scroll down ? */
                              {
                                _EditHWindow.DFirstLine --;
                                _EditAWindow.DFirstLine --;
                                LastKnownPosition = _EditHWindow.DFirstLine;
                                _EditHWindow.VSlider->SliderStart = _EditHWindow.VSlider->StepRate * _EditHWindow.DFirstLine;
                                PrintComboEntries (&_EditHWindow);
                                PrintComboEntries (&_EditAWindow);
                              }
                              if (_NumbersAreHex)
                                sprintf (AEdAddr, " %04X", _EditCurAddress);
                              else
                                sprintf (AEdAddr, "%5u", _EditCurAddress);
                              UpdateButtonText (&OWindow, &PEdAddr, BF_FITTXT);
                            }
                            GlobalKey = 0x0000;
                            break;
              case K_HOME : CurTmpAddress = (word)_EditCurY * 16;                                            /* Determine address */
                            if ((dword)CurTmpAddress + (LoadedAddress & 0xFFF0) < LoadedAddress)       /* (Incomplete first line) */
                              CurTmpAddress = CurTmpIndex = (LoadedAddress & 0x000F);
                            else
                              CurTmpIndex = 0;
                            CurTmpAddress += (LoadedAddress & 0xFFF0);                                      /* Get 'real' address */
                            if (CurTmpAddress != _EditCurAddress)                           /* A reposition is going to be made ? */
                            {
                              _DrawEditCursor (FALSE);
                              _EditCurAddress = CurTmpAddress;
                              _EditCurIndex = CurTmpIndex;
                              if (_EditWhichWindow)
                              {
                                _EditCurX = 8 + 3 * _EditCurIndex + (_EditCurIndex >= 8 ? 1 : 0);
                                _EditCurHalf = TRUE;
                              }
                              else
                                _EditCurX = _EditCurIndex;
                              if (_NumbersAreHex)
                                sprintf (AEdAddr, " %04X", _EditCurAddress);
                              else
                                sprintf (AEdAddr, "%5u", _EditCurAddress);
                              UpdateButtonText (&OWindow, &PEdAddr, BF_FITTXT);
                            }
                            GlobalKey = 0x0000;
                            break;
              case K_END  : CurTmpAddress = (word)_EditCurY * 16 + 15;                                       /* Determine address */
                            if ((dword)CurTmpAddress >= BlockLength + (LoadedAddress & 0x000F))         /* (Incomplete last line) */
                            {
                              CurTmpAddress = BlockLength + (LoadedAddress & 0x000F) - 1;
                              CurTmpIndex = (CurTmpAddress & 0x000F);
                            }
                            else
                              CurTmpIndex = 15;
                            CurTmpAddress += (LoadedAddress & 0xFFF0);                                      /* Get 'real' address */
                            if (CurTmpAddress != _EditCurAddress)                           /* A reposition is going to be made ? */
                            {
                              _DrawEditCursor (FALSE);
                              _EditCurAddress = CurTmpAddress;
                              _EditCurIndex = CurTmpIndex;
                              if (_EditWhichWindow)
                              {
                                _EditCurX = 8 + 3 * _EditCurIndex + (_EditCurIndex >= 8 ? 1 : 0);
                                _EditCurHalf = TRUE;
                              }
                              else
                                _EditCurX = _EditCurIndex;
                              if (_NumbersAreHex)
                                sprintf (AEdAddr, " %04X", _EditCurAddress);
                              else
                                sprintf (AEdAddr, "%5u", _EditCurAddress);
                              UpdateButtonText (&OWindow, &PEdAddr, BF_FITTXT);
                            }
                            GlobalKey = 0x0000;
                            break;
              default     : Bt = (byte)(GlobalKey & 0x00FF);
                            AcceptClick = FALSE;                                                /* Assume the key will be dropped */
                            CurTmpX = _EditCurX;
                            CurTmpY = _EditCurY;
                            CurTmpHalf = _EditCurHalf;
                            CurTmpIndex = _EditCurIndex;
                            CurTmpAddress = _EditCurAddress;
                            if (_EditWhichWindow)                                                              /* Hex-part window */
                            {
                              if (isxdigit (Bt))                                                             /* (Only hex digits) */
                              {
                                BtUpper = toupper (Bt);
                                AcceptClick = TRUE;
                                _DrawEditCursor (FALSE);
                                _EditChanged = TRUE;                                                  /* Signal: data was changed */
                                if (CurTmpHalf)                                          /* Currently appointing left byte half ? */
                                {
                                  _DrawEditCursor (FALSE);
                                  if (BtUpper >= 'A')
                                    *(_ObjectData + _EditCurAddress - LoadedAddress) = 
                                           (*(_ObjectData + _EditCurAddress - LoadedAddress) & 0x0F) | ((BtUpper - 'A' + 10) << 4);
                                  else
                                    *(_ObjectData + _EditCurAddress - LoadedAddress) = 
                                           (*(_ObjectData + _EditCurAddress - LoadedAddress) & 0x0F) | ((BtUpper - '0') << 4);
                                  CurTmpX ++;
                                  CurTmpHalf = FALSE;
                                }
                                else
                                {
                                  if (BtUpper >= 'A')
                                    *(_ObjectData + _EditCurAddress - LoadedAddress) = 
                                           (*(_ObjectData + _EditCurAddress - LoadedAddress) & 0xF0) | (BtUpper - 'A' + 10);
                                  else
                                    *(_ObjectData + _EditCurAddress - LoadedAddress) = 
                                           (*(_ObjectData + _EditCurAddress - LoadedAddress) & 0xF0) | (BtUpper - '0');
                                  if (CurTmpAddress < LoadedAddress + BlockLength - 1)                  /* Is there a next byte ? */
                                  {
                                    CurTmpAddress ++;
                                    CurTmpHalf = TRUE;
                                    if (++ CurTmpIndex > 15)                                                     /* End of line ? */
                                    {
                                      CurTmpIndex = 0;
                                      CurTmpY ++;
                                      CurTmpX = 8;
                                    }
                                    else
                                      CurTmpX = 8 + 3 * CurTmpIndex + (CurTmpIndex >= 8 ? 1 : 0);
                                  }
                                }
                              }
                            }
                            else
                            {
                              if (Bt >= 32 && Bt < 128)                                                /* (Only ASCII characters) */
                              {
                                AcceptClick = TRUE;
                                _DrawEditCursor (FALSE);
                                *(_ObjectData + _EditCurAddress - LoadedAddress) = Bt;
                                _EditChanged = TRUE;                                                  /* Signal: data was changed */
                                if (CurTmpAddress < LoadedAddress + BlockLength - 1)                    /* Is there a next byte ? */
                                {
                                  CurTmpAddress ++;
                                  if (++ CurTmpIndex > 15)                                                       /* End of line ? */
                                  {
                                    CurTmpX = CurTmpIndex = 0;
                                    CurTmpY ++;
                                  }
                                  else
                                    CurTmpX = CurTmpIndex;
                                }
                              }
                            }
                            if (AcceptClick)                                                                 /* Move to the right */
                            {
                              Index = (_EditCurAddress - LoadedAddress + FirstSkipKeep) & 0xFFF0;             /* Rebuild the line */
                              GetComboEntry (&_EditHWindow, &Entry, _EditCurY);
                              memset (Entry.Text, ' ', 56);
                              if (_NumbersAreHex)
                                sprintf (Entry.Text, " %04X - ", _EditCurAddress & 0xFFF0);
                              else
                                sprintf (Entry.Text, "%5u - ", _EditCurAddress & 0xFFF0);
                              LineIndex = 8;
                              if (!Index)
                                FirstSkip = FirstSkipKeep;
                              else
                                FirstSkip = 0;
                              if (FirstSkip > 0)
                              {
                                Entry.Text[LineIndex] = ' ';
                                LineIndex += FirstSkip * 3;
                                if (FirstSkip >= 8)
                                  LineIndex ++;
                              }
                              for (SubIndex = FirstSkip ; SubIndex < 16 && Index + SubIndex < BlockLength ; )
                              {
                                sprintf (Entry.Text + LineIndex, "%02X ",
                                         *((byte *)_ObjectData + Index + SubIndex - FirstSkipKeep));
                                LineIndex += 3;
                                Entry.Text[LineIndex] = ' ';
                                if (++ SubIndex == 8)
                                  LineIndex ++;
                              }
                              Entry.Text[-- LineIndex] = '\0';
                              PutComboEntry (&_EditHWindow, &Entry, _EditCurY, TRUE);
                              GetComboEntry (&_EditAWindow, &Entry, _EditCurY);
                              LineIndex = 0;
                              if (FirstSkip > 0)
                              {
                                memset (Entry.Text, ' ', FirstSkip);
                                LineIndex += FirstSkip;
                              }
                              for (SubIndex = FirstSkip ; SubIndex < 16 && Index + SubIndex < BlockLength ; )
                              {
                                Bt = *((byte *)_ObjectData + Index + SubIndex - FirstSkipKeep);
                                if (Bt >= 32 && Bt < 128)
                                {
                                  if (Bt == '{' || Bt == '}')
                                    Entry.Text[LineIndex ++] = Bt;
                                  Entry.Text[LineIndex ++] = Bt;
                                }
                                else
                                  Entry.Text[LineIndex ++] = '.';
                                SubIndex ++;
                              }
                              Entry.Text[LineIndex] = '\0';
                              PutComboEntry (&_EditAWindow, &Entry, _EditCurY, TRUE);
                              _EditCurX = CurTmpX;                                              /* Set the new cursor coordinates */
                              _EditCurY = CurTmpY;
                              _EditCurHalf = CurTmpHalf;
                              _EditCurIndex = CurTmpIndex;
                              _EditCurAddress = CurTmpAddress;
                              if (_EditCurY == (_EditHWindow.DFirstLine + _EditHWindow.DHeight))         /* Need to scroll down ? */
                              {
                                _EditHWindow.DFirstLine ++;
                                _EditAWindow.DFirstLine ++;
                                LastKnownPosition = _EditHWindow.DFirstLine;
                                _EditHWindow.VSlider->SliderStart = _EditHWindow.VSlider->StepRate * _EditHWindow.DFirstLine;
                                PrintComboEntries (&_EditHWindow);
                                PrintComboEntries (&_EditAWindow);
                              }
                              if (_NumbersAreHex)
                                sprintf (AEdAddr, " %04X", _EditCurAddress);
                              else
                                sprintf (AEdAddr, "%5u", _EditCurAddress);
                              UpdateButtonText (&OWindow, &PEdAddr, BF_FITTXT);
                            }
                            break;
            }
          }
        MouseStatus ();
      }
      if (Restart && !Resize)
      {
        ReInitComboWindow (&_EditHWindow, 0);
        ReInitComboWindow (&_EditAWindow, 0);
      }
    }
    if (SubProcessActive)
    {
      ReleaseTimedHandler (_FlashEditCursor, &OWindow);
      SubProcessActive = FALSE;
    }
    DestroyComboWindow (&_EditHWindow, Resize ? COMF_RESIZING : 0);
    DestroyComboWindow (&_EditAWindow, Resize ? COMF_RESIZING : 0);
    DestroyWindow (&OWindow);
  }
  GlobalKey = 0x0000;
  return (StartDisassembler);
}

void RunDisassemblerOrDump (struct MenuEntry *Entry, bool RunDisassembler)

/**********************************************************************************************************************************/
/* Pre   : `Entry' points to the activated menu entry, `RunDisassembler' is TRUE if the disassembler should be used, FALSE if the */
/*         hex editor should be used.                                                                                             */
/* Post  : The disassembler/hex editor has been called, possibly from eachother. The datablock may have been edited (using the    */
/*         hex editor) and has been written back if so. Block encryption has been fully taken care of.                            */
/* Import: ViewAsDisassembly, ViewSpectrumDump.                                                                                   */
/**********************************************************************************************************************************/

{
  struct ComboEntry   ComEntry;
  struct TapeEntry_s *DirP;
  short               WhichWindow;
  dword               MaxBlockLength;
  dword               MaxAllocLength;
  dword               XMSIndex;
  bool                More = TRUE;
  bool                RunWhichOne;
  word                StartAddress = 0;
  word                Size;
  word                Cnt;

  WhichWindow = Entry->ReactNum & 0x01;
  GetComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &ComEntry, TapeWindow[WhichWindow].SelectedLine);
  if (ComEntry.Identifier == 0)                                                                                 /* (Sanity check) */
    return;
  DirP = (struct TapeEntry_s *)ComEntry.Identifier;
  if (!GetLoadedAddress (WhichWindow))                                                          /* Determine load ('ORG') address */
    return;
  XMSIndex = GetTapeBlockXMSIndex (WhichWindow, DirP);
  MaxAllocLength = DirP->AllocLength > (dword)65536 ? (dword)65536 : DirP->AllocLength;
  MaxBlockLength = DirP->BlockLength > (dword)65536 ? (dword)65536 : DirP->BlockLength;
  CopyFromXMS (MaxAllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, _ObjectData);               /* Fetch object code */
  DecryptBlock (WhichWindow, _ObjectData, MaxBlockLength);                                                    /* Decrypt the code */
  _EditCurX = _EditCurY = -1;                                                                          /* Disable any edit cursor */
  _EditChanged = FALSE;                                                                            /* Signal: no changes made yet */
  RunWhichOne = RunDisassembler;
  while (More)
  {
    if (More && RunWhichOne)
    {
      More = ViewAsDisassembly (WhichWindow, MaxBlockLength, &StartAddress);                            /* Start the Disassembler */
      RunWhichOne = FALSE;
    }
    if (More && !RunWhichOne)
    {
      More = ViewSpectrumDump (MaxBlockLength, &StartAddress);                                               /* Start the HexDump */
      RunWhichOne = TRUE;
    }
  }
  if (_EditChanged)                                                                                   /* Changes have been made ? */
    if (!QuestionRequester (-1, -1, FALSE, FALSE, NULL, NULL, NULL, NULL, NULL, "Save Changes",
                            "The data block has been changed!\nDo you want to save the changes?"))
    {
      EncryptBlock (WhichWindow, _ObjectData, MaxBlockLength);                                                /* Encrypt the code */
      CopyToXMS (MaxAllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, _ObjectData, XMSIndex);              /* Save object code */
      if (DirP->Flags.FlagByteUsed)                                                           /* Recalculate parity matching byte */
        DirP->CalculatedParity = DirP->FlagByte;
      else
        DirP->CalculatedParity = 0;
      More = TRUE;
      XMSIndex = 0;
      while (More)
      {
        Size = GetTapeBlock (WhichWindow, &XMSIndex, &More);
        for (Cnt = 0 ; Cnt < Size ; Cnt ++)
          DirP->CalculatedParity ^= *(WorkBuffer + Cnt);
      }
      SignalTapeChanged (WhichWindow, TRUE);
      SetupNewData (WhichWindow);
    }
}
